/***********************************************************************
This software module was originally developed by Khaled MAMOU
(INT - ARTEMIS ) in the course of development of the MPEG-4
Systems (ISO/IEC 14496-1) standard. This software module is an implementation
of a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. The author retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2009
************************************************************************

For more information or to receive an up to date version of this module,
you may contact Khaled MAMOU at Khaled.Mamou@int-edu.eu
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Decoder
FILE:       KMSC3DMCDecoder.cpp
AUTHOR:     Khaled MAMOU (INT)
OVERVIEW:   Implementation of SC3DMC Decoder- 
-------------------------------------------------------------------------*/
#include "StdAfx.h"
#include "KMSC3DMCDecoder.h"
#include ".\SC3DMC\SC3DMC_DecoderLib\h\SC3DMC_SC3DMCDecoder.h"

#include "ODP.h"
#include "../../OpenGL/AFX/AfxStreamProxy.h"
/*-----------------------------------------------------------------
Constructor, destructor
-----------------------------------------------------------------*/
KMSC3DMCDecoder :: KMSC3DMCDecoder()
{
}

KMSC3DMCDecoder :: ~KMSC3DMCDecoder()
{
}
/*-----------------------------------------------------------------
Calculate the output size in bytes
-----------------------------------------------------------------*/
int KMSC3DMCDecoder :: GetOptimalOutputSize ()
{
	// value used for the allocate the composition buffer
	int a= 100000000;
		
	return a;
}

/*-----------------------------------------------------------------
Parse the decoder specific info
-----------------------------------------------------------------*/
	
BOOL KMSC3DMCDecoder :: Setup (AFXExtDescriptor *pSpecific)
{
  return TRUE;
}

/*-------------------------------------------------------------
Decode
-------------------------------------------------------------*/

BOOL KMSC3DMCDecoder :: Decode (LPBYTE pInput, int nInputLength, LPBYTE pOutput, int &rnOutputLength)
{
	SC3DMCDecoder mySC3DMCDecoder;
	SC3DMC_IndexedFaceSet myIFS;
	SC3DMCDecoder myDecoder;
	myDecoder.SetIFS(&myIFS);	
	myDecoder.Decode(pInput, nInputLength);
//	myIFS.SaveIFSVRML2("d:\\temp.wrl");

	unsigned char * buffer = pOutput;
	unsigned int bufferSize = 0;

	float creaseAngle = myIFS.GetCreaseAngle();
	
	unsigned char compressionMask = 0;
	bool isTriangularMesh = true;
	bool otherAttributesPerVertex = true;
	bool markerBit = true;

	if (myIFS.GetCCW()) compressionMask += 1;
	if (myIFS.GetSolid()) compressionMask += 2;
	if (myIFS.GetConvex()) compressionMask += 4;
	if (myIFS.GetColorPerVertex()) compressionMask += 8;
	if (myIFS.GetNormalPerVertex()) compressionMask += 16;
	if (otherAttributesPerVertex) compressionMask += 32;	// otherAttributesPerVertex true
	if (isTriangularMesh) compressionMask += 64;
	if (markerBit) compressionMask += 128;
	unsigned int numberOfCoord = myIFS.GetNCoord();
	unsigned int numberOfNormal = myIFS.GetNNormal();
	unsigned int numberOfColor = myIFS.GetNColor();
	unsigned int numberOfTexCoord = myIFS.GetNTexCoord();
	unsigned int numberOfOtherAttributes = 0;
	unsigned char dOfOtherAttributes = 0;
	
	unsigned int numberOfCoordIndex = myIFS.GetNCoordIndex();
	unsigned int numberOfNormalIndex = myIFS.GetNNormalIndex();
	unsigned int numberOfColorIndex = myIFS.GetNColorIndex();
	unsigned int numberOfTexCoordIndex = myIFS.GetNTexCoordIndex();
	unsigned int numberOfOtherAttributesIndex = 0;

	Write2Stream(buffer, bufferSize, &creaseAngle, sizeof(float));
	Write2Stream(buffer, bufferSize, &compressionMask, sizeof(unsigned char)); 
	
	Write2Stream(buffer, bufferSize, &numberOfCoord, sizeof(unsigned int));
	Write2Stream(buffer, bufferSize, &numberOfNormal, sizeof(unsigned int));
	Write2Stream(buffer, bufferSize, &numberOfColor, sizeof(unsigned int));
	Write2Stream(buffer, bufferSize, &numberOfTexCoord, sizeof(unsigned int));
	Write2Stream(buffer, bufferSize, &numberOfOtherAttributes, sizeof(unsigned int));
	
	Write2Stream(buffer, bufferSize, &dOfOtherAttributes, sizeof(unsigned char));

	Write2Stream(buffer, bufferSize, &numberOfCoordIndex, sizeof(unsigned int));
	Write2Stream(buffer, bufferSize, &numberOfNormalIndex, sizeof(unsigned int));
	Write2Stream(buffer, bufferSize, &numberOfColorIndex, sizeof(unsigned int));
	Write2Stream(buffer, bufferSize, &numberOfTexCoordIndex, sizeof(unsigned int));
	Write2Stream(buffer, bufferSize, &numberOfOtherAttributesIndex, sizeof(unsigned int));

	if (numberOfCoord > 0)  Write2Stream(buffer, bufferSize, myIFS.GetCoord(), sizeof(float) * numberOfCoord * 3);
	if (numberOfNormal > 0) Write2Stream(buffer, bufferSize, myIFS.GetNormal(), sizeof(float) * numberOfNormal * 3);
	if (numberOfColor > 0)  Write2Stream(buffer, bufferSize, myIFS.GetColor(), sizeof(float) * numberOfColor * 3);
	if (numberOfTexCoord > 0)  Write2Stream(buffer, bufferSize, myIFS.GetTexCoord(), sizeof(float) * numberOfTexCoord * 2);

	if (numberOfCoordIndex > 0)  Write2Stream(buffer, bufferSize, myIFS.GetCoordIndex(), sizeof(int) * numberOfCoordIndex * 3);
	if (numberOfNormalIndex > 0) Write2Stream(buffer, bufferSize, myIFS.GetNormalIndex(), sizeof(int) * numberOfNormalIndex * 3);
	if (numberOfColorIndex > 0)  Write2Stream(buffer, bufferSize, myIFS.GetColorIndex(), sizeof(int) * numberOfColorIndex * 3);
	if (numberOfTexCoordIndex > 0)  Write2Stream(buffer, bufferSize, myIFS.GetTexCoordIndex(), sizeof(int) * numberOfTexCoordIndex * 3);

	rnOutputLength = bufferSize;
	return true;
}
