//-------------------------------------------------------------------------
// 
// Copyright ETRI, Hanyang University 2008, 2009
// Licensed Materials, Program Property of ETRI & Hanyang University
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.
// ETRI and Hanyang University retain full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//    AUTHORS:
//    Seung Wook Lee (ETRI)
//    Daiyong Kim (Hanyang University)
//    KyoungSoo Son (Hanyang University)
//
//    http://www.etri.re.kr/
//    http://dmlab.hanyang.ac.kr/
//  
//    Seung Wook Lee     (tajinet@etri.re.kr)  
//    Daiyong Kim        (dykim@dmlab.hanyang.ac.kr)    
//    KyoungSoo Son      (bonokensin@dmlab.hanyang.ac.kr)
//
//-------------------------------------------------------------------------



#include <iostream>
#include <time.h>
#include <string>
#include <iostream>
#include <fstream>
#include <sstream> 
#include <math.h>
#include <vector>
#include <map>
using namespace std;

#include <assert.h>

#include "IndexedFaceSet.h"
#include "QBCR_Decoder.h"
#include "utils.h"

#define LEN_BYTE		8
#define LEN_QP			5
#define COORD_DIVISION	3
#define LEN_BIT			1
#define DEFAULT		0
#define DIFF		1
#define BP			1
#define AC			2
#define LEN_PREFIX	5

#define SC3DMCStartCode 0x00001F1



QBCRDecoder::QBCRDecoder()
{
	readBitstream = NULL;

	unCurrentBitPosition = 7;
	unCurrentBufferPosition = 0;
	codingtype = 0;
	entropytype = 0;
	BPLTable  =NULL;
}

QBCRDecoder::~QBCRDecoder()
{
	if(readBitstream != NULL)
	{
		delete [] readBitstream;
		readBitstream = NULL;
	}
	if(BPLTable != NULL)
	{
		delete [] BPLTable;
	}

}

void QBCRDecoder::SetHeader()
{
		
	numberOfCoord = outIFS_->nCoord_;
	numberOfNormal = outIFS_->nNormal_;
	numberOfColor = outIFS_->nColor_;
	numberOfTexCoord = outIFS_->nTexCoord_;
	numberOfOtherAttributes = m_Header->numberOfOtherAttribute;
	dimensionOfOtherAttributes = m_Header->dimensionOfOtherAttribute;
	numberOfCoordIndex = outIFS_->nCoordIndex_;
	numberOfNormalIndex = outIFS_->nNormalIndex_;
	numberOfColorIndex = outIFS_->nColorIndex_;
	numberOfTexCoordIndex = outIFS_->nTexCoordIndex_;
	numberOfOtherAttributesIndex = m_Header->numberOfOtherAttributeIndex;

	for(int i=0; i< 3; i++)
	{
		quantMinForGeometry[i] = m_Header->quantMinGeometry[i];
		quantMinForNormal[i] = m_Header->quantMinNormal[i];
		quantMinForColor[i] = m_Header->quantMinColor[i];
		quantRangeForGeometry[i] = m_Header->rangeGeometry[i];
		quantRangeForNormal[i] = m_Header->rangeNormal[i];
		quantRangeForColor[i] = m_Header->rangeColor[i];

	}
	for(int i= 0; i < 2; i++)
	{
		quantMinForTexCoord[i] = m_Header->quantMinTexCoord[i] ;
		quantRangeForTexCoord[i] = m_Header->rangeTexCoord[i];

	}
	
	
	QPforGeometry = m_Header->QPforGeometry;
	quantValueforGeometry = pow(2.0,(double)QPforGeometry) - 1;
	QPforNormal =  m_Header->QPforNormal;
	quantValueforNormal = pow(2.0,(double)QPforNormal) - 1;
	QPforTexCoord =  m_Header->QPforTexCoord;
	quantValueforTexCoord = pow(2.0,(double)QPforTexCoord) - 1;
	QPforColor =  m_Header->QPforColor;
	quantValueforColor = pow(2.0,(double)QPforColor) - 1;

}


int QBCRDecoder::DecoderQBCRData(unsigned char * &compressedStream)
{
	m_pReadStream=compressedStream; //ReadData(name);
	SetHeader() ;
	DecoderHeader();
	if(entropytype == BP)
	{
		BPLInit();
	}
	DecoderGeometry();
	DecoderConnectivity();
	DecoderNormal();
	DecoderNormalIndex();
	DecoderTexCoord();
	DecoderTexCoordIndex();
	DecoderColor();
	DecoderColorIndex();
	
	return 1;
}

/*
void QBCRDecoder::DecoderCommonHeader()
{
	unsigned int startCode = 0;
	ReadFromStream( &startCode, sizeof(unsigned int));
	
	if (startCode != SC3DMCStartCode)
	{
		printf("Wrong Started Code\n");
		exit(1);
	}

	m_streamSize = 0;
	ReadFromStream( &m_streamSize, sizeof(unsigned int));

	float creaseAngle = 0.0;
	ReadFromStream( &creaseAngle, sizeof(float));
//	m_ifs->SetCreaseAngle(creaseAngle);

	compressionMask = 0;

	ReadFromStream( &compressionMask, sizeof(unsigned char));
	/*
	if (compressionMask & 1) m_ifs->SetCCW(true);
	else m_ifs->SetCCW(false);
	if (compressionMask & 2) m_ifs->SetSolid(true);
	else m_ifs->SetSolid(false);
	if (compressionMask & 4) m_ifs->SetConvex(true);
	else m_ifs->SetConvex(false);
	if (compressionMask & 8) m_ifs->SetColorPerVertex(true);
	else  m_ifs->SetColorPerVertex(false);
	if (compressionMask & 16) m_ifs->SetNormalPerVertex(true);
	else m_ifs->SetNormalPerVertex(false);
	if (compressionMask & 32) otherAttributesPerVertex = true;
	otherAttributesPerVertex = false;
	if (compressionMask & 64) m_vertexOrderPres = 1;
	if (compressionMask & 128) m_triangleOrderPres = 1;

	m_meshType = 0; // triangular mesh
	ReadFromStream( &m_meshType, sizeof(unsigned char));

	encodingType = 0;
	ReadFromStream( &encodingType, sizeof(unsigned char));

	numberOfCoord = 0;
	numberOfNormal = 0;
	numberOfColor = 0;
	numberOfTexCoord =0;
	numberOfOtherAttributes = 0;
	dimensionOfOtherAttributes = 0;
	
	numberOfCoordIndex = 0;
	numberOfNormalIndex = 0;
	numberOfColorIndex = 0;
	numberOfTexCoordIndex = 0;
	numberOfOtherAttributesIndex = 0;

	ReadFromStream( &numberOfCoord, sizeof(unsigned int));
	ReadFromStream( &numberOfNormal, sizeof(unsigned int));
	ReadFromStream( &numberOfColor, sizeof(unsigned int));
	ReadFromStream( &numberOfTexCoord, sizeof(unsigned int));
	ReadFromStream( &numberOfOtherAttributes, sizeof(unsigned int));

	if (numberOfCoord>0)
	{
		ReadFromStream( &numberOfCoordIndex, sizeof(unsigned int));
	}
	if (numberOfTexCoord>0)
	{
		ReadFromStream( &numberOfTexCoordIndex, sizeof(unsigned int));
	}
	if (numberOfNormal>0)
	{
		ReadFromStream( &numberOfNormalIndex, sizeof(unsigned int));
	}
	if (numberOfColor>0)
	{
		ReadFromStream( &numberOfColorIndex, sizeof(unsigned int));
	}
	if (numberOfOtherAttributes >0)
	{
		ReadFromStream( &dimensionOfOtherAttributes, sizeof(unsigned int));
		ReadFromStream( &numberOfOtherAttributesIndex, sizeof(unsigned int));
	}
/*
	m_ifs->SetNCoord(numberOfCoord);
	m_ifs->SetNNormal(numberOfNormal);
	m_ifs->SetNColor(numberOfColor);
	m_ifs->SetNTexCoord(numberOfTexCoord);	
	m_ifs->SetNCoordIndex(numberOfCoordIndex);
	m_ifs->SetNNormalIndex(numberOfNormalIndex);
	m_ifs->SetNColorIndex(numberOfColorIndex);
	m_ifs->SetNTexCoordIndex(numberOfTexCoordIndex);

	m_ifs->AllocateMem();

}
*/

void QBCRDecoder::DecoderHeader()
{
	ReadFromStream(m_pReadStream, &codingtype, sizeof(unsigned char));
	ReadFromStream(m_pReadStream, &entropytype, sizeof(unsigned char));
	

}

void QBCRDecoder::DecoderConnectivity()
{
	if(entropytype == DEFAULT)
	{
		unsigned int nLoop, lengthCoordIndex;
		lengthCoordIndex = int(log((double)numberOfCoord-1)/log(2.0)) + 1;
		for(nLoop = 0; nLoop < COORD_DIVISION * numberOfCoordIndex; nLoop++)
		{
			outIFS_->coordIndex_[nLoop] = (int)(ReadBitsFromStorage(lengthCoordIndex));
		}
	}
	else if(entropytype == BP)
	{
		unsigned int nLoop, prefixSize, nBPL, nPayload, nSign;
		int previousQuant[3],difValue, i,  nQuantValue;
		nQuantValue=0;
		prefixSize = ReadBitsFromStorage(5);
		for(i = 0; i < 3; i++)
		{
			previousQuant[i] = 0;
		}
		for(nLoop = 0; nLoop < 3 * numberOfCoordIndex; nLoop++)
		{

			nBPL=ReadBitsFromStorage(prefixSize);
			if(nBPL > 2)
			{
				nPayload = ReadBitsFromStorage(nBPL-1);
			}
			else
			{
				nPayload = 0;
			}

			difValue = BPLTable[nBPL] + nPayload;
			if(difValue != 0)
			{
				nSign = ReadBitsFromStorage(1);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}

			nQuantValue = difValue+previousQuant[nLoop%3];
			previousQuant[nLoop % 3] = nQuantValue;
			outIFS_->coordIndex_[nLoop] = nQuantValue;
		}

	}
	else
	{
		unsigned int nLoop, prefixSize, nBPL, nPayload, nSign;
		int previousQuant[3],difValue, i,  nQuantValue;
		nQuantValue=0;
		for(i = 0; i < 3; i++)
		{
			previousQuant[i] = 0;
		}
		unsigned int code_bytes = 0;
		ReadFromStream(m_pReadStream, &code_bytes, sizeof(unsigned int));
		unsigned char * code_buffer = new unsigned char [code_bytes];                                                   
		ReadFromStream(m_pReadStream, code_buffer, sizeof(unsigned char)*code_bytes);
		
		Adaptive_Data_Model mModel(1024);
		Adaptive_Data_Model mSign (2);
		Adaptive_Data_Model mhasnext (2);
		Arithmetic_Codec acd(code_bytes, code_buffer);
		acd.start_decoder();	// initialize decoder

		int data[4];
		data[0] = 1;
		data[1] = 1<<10;
		data[2] = 1<<20;
		data[3] = 1<<30;
		int count = 0;
		for(nLoop = 0; nLoop < 3 * numberOfCoordIndex; nLoop++)
		{

			difValue = 0;

			while(1)
			{
				difValue += data[count] * acd.decode(mModel);
				count++;
				int temp = acd.decode(mhasnext);

				if( temp ==0)
					break;
			}
			count = 0;

			if(difValue != 0)
			{
				nSign = acd.decode(mSign);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}
			nQuantValue = difValue+previousQuant[nLoop%3];
			
			previousQuant[nLoop % 3] = nQuantValue;
			
			//dmlab
			//coord.push_back(InverseQuantization(nQuantValue,(nLoop%3)));
			outIFS_->coordIndex_[nLoop] = nQuantValue;

			 //~dmlab
		}
	}

}

void QBCRDecoder::DumpCompressedVB(char* nameVB)
{
	outIFS_->SaveIFSVRML2(nameVB);
	if(readBitstream != NULL)
	{
		delete [] readBitstream;
		readBitstream = NULL;
	}
	unCurrentBitPosition = 7;
	unCurrentBufferPosition = 0;

}

void QBCRDecoder::DecoderGeometry()
{
	unsigned int nLoop, quantCoord;
	if(entropytype == DEFAULT)
	{
		for(nLoop = 0; nLoop < 3 * numberOfCoord; nLoop++)
		{
			quantCoord = ReadBitsFromStorage(QPforGeometry);
			outIFS_->coord_[nLoop] = InverseQuantization(quantCoord, quantRangeForGeometry[nLoop%3] ,quantMinForGeometry[nLoop%3], quantValueforGeometry);
		}
	}
	else if(entropytype == BP)
	{
		unsigned int nLoop, prefixSize, nBPL, nPayload, nSign;
		int previousQuantCoord[3];
		int difValue, i, nQuantValue;
		prefixSize = ReadBitsFromStorage(5);
		nQuantValue=0;
		for(i = 0; i < 3; i++)
		{
			previousQuantCoord[i] = 0;
		}

		for(nLoop = 0; nLoop < 3 * numberOfCoord; nLoop++)
		{
			nBPL=ReadBitsFromStorage(prefixSize);
			if(nBPL > 2)
			{
				nPayload = ReadBitsFromStorage(nBPL-1);
			}
			else
			{
				nPayload = 0;
			}

			difValue = BPLTable[nBPL] + nPayload;
			if(difValue != 0)
			{
				nSign = ReadBitsFromStorage(1);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}

			nQuantValue = difValue + previousQuantCoord[nLoop%3];
			
			previousQuantCoord[nLoop % 3] = nQuantValue;
			
			//dmlab
			//coord.push_back(InverseQuantization(nQuantValue,(nLoop%3)));
			outIFS_->coord_[nLoop] = InverseQuantization(nQuantValue, quantRangeForGeometry[nLoop%3] ,quantMinForGeometry[nLoop%3], quantValueforGeometry);

			 //~dmlab
		}

	}
	else
	{
		unsigned int nLoop, nSign;
		int previousQuantCoord[3],difValue, i, nQuantValue;
		nQuantValue=0;
		m_pReadStream = m_pReadStream+unCurrentBufferPosition;
		unsigned int code_bytes = 0;
		ReadFromStream(m_pReadStream, &code_bytes, sizeof(unsigned int));

		unsigned char * code_buffer = new unsigned char [code_bytes];                                                   
		ReadFromStream(m_pReadStream, code_buffer, sizeof(unsigned char)*code_bytes);

		
		Adaptive_Data_Model mModel(1<<QPforGeometry);
		Adaptive_Data_Model mSign (2);
		Arithmetic_Codec acd(code_bytes, code_buffer);
		acd.start_decoder();	// initialize decoder

		for(i = 0; i < 3; i++)
		{
			previousQuantCoord[i] = 0;
		}
		

		for(nLoop = 0; nLoop < 3 * numberOfCoord; nLoop++)
		{
			difValue = acd.decode(mModel);
			if(difValue != 0)
			{
				nSign = acd.decode(mSign);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}

			nQuantValue = difValue + previousQuantCoord[nLoop%3];
			
			previousQuantCoord[nLoop % 3] = nQuantValue;
			
			//dmlab
			//coord.push_back(InverseQuantization(nQuantValue,(nLoop%3)));
			outIFS_->coord_[nLoop] = InverseQuantization(nQuantValue, quantRangeForGeometry[nLoop%3] ,quantMinForGeometry[nLoop%3], quantValueforGeometry);

			 //~dmlab
		}

	}

}


void QBCRDecoder::DecoderNormal()
{
	if(numberOfNormal == 0)
		return;
	unsigned int nLoop, quantNormal;
	if(entropytype == DEFAULT)
	{
		for(nLoop = 0; nLoop < 3 * numberOfNormal; nLoop++)
		{
			quantNormal=ReadBitsFromStorage(QPforNormal);
			outIFS_->normal_[nLoop]  = (InverseQuantization(quantNormal,quantRangeForNormal[nLoop%3] ,quantMinForNormal[nLoop%3], quantValueforNormal));
		}
		
	}
	else if(entropytype == BP)
	{
		unsigned int nLoop, prefixSize, nBPL, nPayload, nSign;
		int previousQuant[3];
		int difValue, i, nQuantValue;
		prefixSize = ReadBitsFromStorage(5);
		nQuantValue=0;
		for(i = 0; i < 3; i++)
		{
			previousQuant[i] = 0;
		}

		for(nLoop = 0; nLoop < 3 * numberOfNormal; nLoop++)
		{
			nBPL=ReadBitsFromStorage(prefixSize);
			if(nBPL > 2)
			{
				nPayload = ReadBitsFromStorage(nBPL-1);
			}
			else
			{
				nPayload = 0;
			}

			difValue = BPLTable[nBPL] + nPayload;
			if(difValue != 0)
			{
				nSign = ReadBitsFromStorage(1);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}

			nQuantValue = difValue + previousQuant[nLoop%3];
			
			previousQuant[nLoop % 3] = nQuantValue;
			
			//dmlab
			//coord.push_back(InverseQuantization(nQuantValue,(nLoop%3)));
			outIFS_->normal_[nLoop] = InverseQuantization(nQuantValue, quantRangeForNormal[nLoop%3] ,quantMinForNormal[nLoop%3], quantValueforNormal);

			 //~dmlab
		}

	}
	else
	{
		unsigned int nLoop, nSign;
		int previousQuant[3],difValue, i, nQuantValue;
		nQuantValue=0;
		
		unsigned int code_bytes = 0;
		ReadFromStream(m_pReadStream, &code_bytes, sizeof(unsigned int));

		unsigned char * code_buffer = new unsigned char [code_bytes];                                                   
		ReadFromStream(m_pReadStream, code_buffer, sizeof(unsigned char)*code_bytes);

		for(i = 0; i < 3; i++)
		{
			previousQuant[i] = 0;
		}
		
		Adaptive_Data_Model mModel(1<<QPforNormal);
		Adaptive_Data_Model mSign (2);
		Arithmetic_Codec acd(code_bytes, code_buffer);
		acd.start_decoder();	// initialize decoder
		

		for(nLoop = 0; nLoop < 3 * numberOfNormal; nLoop++)
		{
			difValue = acd.decode(mModel);
			if(difValue != 0)
			{
				nSign = acd.decode(mSign);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}

			nQuantValue = difValue + previousQuant[nLoop%3];
			
			previousQuant[nLoop % 3] = nQuantValue;
			
			//dmlab
			//coord.push_back(InverseQuantization(nQuantValue,(nLoop%3)));
			outIFS_->normal_[nLoop] = InverseQuantization(nQuantValue, quantRangeForNormal[nLoop%3] ,quantMinForNormal[nLoop%3], quantValueforNormal);

			 //~dmlab
		}

	}
	
}


void QBCRDecoder::DecoderColor()
{
	if(numberOfColor ==0)
		return;
	unsigned int nLoop, quantColor;

	if(entropytype == DEFAULT)
	{
		for(nLoop = 0; nLoop < 3 * numberOfColor; nLoop++)
		{
			quantColor=ReadBitsFromStorage(QPforColor);
			outIFS_->color_[nLoop] = (InverseQuantization(quantColor,quantRangeForColor[nLoop%3] ,quantMinForColor[nLoop%3], quantValueforColor));
		}
	}
	else if(entropytype == BP)
	{
		unsigned int nLoop, prefixSize, nBPL, nPayload, nSign;
		int previousQuant[3];
		int difValue, i, nQuantValue;
		prefixSize = ReadBitsFromStorage(5);
		nQuantValue=0;
		for(i = 0; i < 3; i++)
		{
			previousQuant[i] = 0;
		}

		for(nLoop = 0; nLoop < 3 * numberOfColor; nLoop++)
		{
			nBPL=ReadBitsFromStorage(prefixSize);
			if(nBPL > 2)
			{
				nPayload = ReadBitsFromStorage(nBPL-1);
			}
			else
			{
				nPayload = 0;
			}

			difValue = BPLTable[nBPL] + nPayload;
			if(difValue != 0)
			{
				nSign = ReadBitsFromStorage(1);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}

			nQuantValue = difValue + previousQuant[nLoop%3];
			
			previousQuant[nLoop % 3] = nQuantValue;
			
			//dmlab
			//coord.push_back(InverseQuantization(nQuantValue,(nLoop%3)));
			outIFS_->color_[nLoop] = InverseQuantization(nQuantValue, quantRangeForColor[nLoop%3] ,quantMinForColor[nLoop%3], quantValueforColor);

			 //~dmlab
		}

	}
	else
	{
		unsigned int nLoop, nSign;
		int previousQuant[3],difValue, i, nQuantValue;
		nQuantValue=0;
		unsigned int code_bytes = 0;
		ReadFromStream(m_pReadStream, &code_bytes, sizeof(unsigned int));

		unsigned char * code_buffer = new unsigned char [code_bytes];                                                   
		ReadFromStream(m_pReadStream, code_buffer, sizeof(unsigned char)*code_bytes);

		for(i = 0; i < 3; i++)
		{
			previousQuant[i] = 0;
		}
		
		Adaptive_Data_Model mModel(1<<QPforColor);
		Adaptive_Data_Model mSign (2);
		Arithmetic_Codec acd(code_bytes, code_buffer);
		acd.start_decoder();	// initialize decoder
		

		for(nLoop = 0; nLoop < 3 * numberOfColor; nLoop++)
		{
			difValue = acd.decode(mModel);
			if(difValue != 0)
			{
				nSign = acd.decode(mSign);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}

			nQuantValue = difValue + previousQuant[nLoop%3];
			
			previousQuant[nLoop % 3] = nQuantValue;
			
			//dmlab
			//coord.push_back(InverseQuantization(nQuantValue,(nLoop%3)));
			outIFS_->color_[nLoop] = InverseQuantization(nQuantValue, quantRangeForColor[nLoop%3] ,quantMinForColor[nLoop%3], quantValueforColor);

			 //~dmlab
		}

	}
}


void QBCRDecoder::DecoderTexCoord()
{
	if(numberOfTexCoord == 0)
		return;
	unsigned int nLoop, quantTexCoord;
	
	if(entropytype == DEFAULT)
	{
		for(nLoop = 0; nLoop < 2 * numberOfTexCoord; nLoop++)
		{
			quantTexCoord=ReadBitsFromStorage(QPforTexCoord);
			outIFS_->texCoord_[nLoop] = (InverseQuantization(quantTexCoord,quantRangeForTexCoord[nLoop%2] ,quantMinForTexCoord[nLoop%2], quantValueforTexCoord));
		}
	}
	else if(entropytype == BP)
	{
		unsigned int nLoop, prefixSize, nBPL, nPayload, nSign;
		int previousQuant[2];
		int difValue, i, nQuantValue;
		prefixSize = ReadBitsFromStorage(5);
		nQuantValue=0;
		for(i = 0; i < 2; i++)
		{
			previousQuant[i] = 0;
		}

		for(nLoop = 0; nLoop < 2 * numberOfTexCoord; nLoop++)
		{
			nBPL=ReadBitsFromStorage(prefixSize);
			if(nBPL > 2)
			{
				nPayload = ReadBitsFromStorage(nBPL-1);
			}
			else
			{
				nPayload = 0;
			}

			difValue = BPLTable[nBPL] + nPayload;
			if(difValue != 0)
			{
				nSign = ReadBitsFromStorage(1);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}

			nQuantValue = difValue + previousQuant[nLoop%2];
			
			previousQuant[nLoop % 2] = nQuantValue;
			
			//dmlab
			//coord.push_back(InverseQuantization(nQuantValue,(nLoop%3)));
			outIFS_->texCoord_[nLoop] = InverseQuantization(nQuantValue,quantRangeForTexCoord[nLoop%2] ,quantMinForTexCoord[nLoop%2], quantValueforTexCoord);

			 //~dmlab
		}

	}
	else
	{
		unsigned int nLoop, nSign;
		int previousQuant[3],difValue, i, nQuantValue;
		nQuantValue=0;
		unsigned int code_bytes = 0;
		ReadFromStream(m_pReadStream, &code_bytes, sizeof(unsigned int));

		unsigned char * code_buffer = new unsigned char [code_bytes];                                                   
		ReadFromStream(m_pReadStream, code_buffer, sizeof(unsigned char)*code_bytes);

		for(i = 0; i < 2; i++)
		{
			previousQuant[i] = 0;
		}
		
		Adaptive_Data_Model mModel(1<<QPforTexCoord);
		Adaptive_Data_Model mSign (2);
		Arithmetic_Codec acd(code_bytes, code_buffer);
		acd.start_decoder();	// initialize decoder
		

		for(nLoop = 0; nLoop < 2 * numberOfTexCoord; nLoop++)
		{
			difValue = acd.decode(mModel);
			if(difValue != 0)
			{
				nSign = acd.decode(mSign);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}

			nQuantValue = difValue + previousQuant[nLoop%2];
			
			previousQuant[nLoop % 2] = nQuantValue;
			
			//dmlab
			//coord.push_back(InverseQuantization(nQuantValue,(nLoop%3)));
			outIFS_->texCoord_[nLoop] = InverseQuantization(nQuantValue,quantRangeForTexCoord[nLoop%2] ,quantMinForTexCoord[nLoop%2], quantValueforTexCoord);
			 //~dmlab
		}

	}
}


void QBCRDecoder::DecoderNormalIndex()
{

	if(numberOfNormalIndex == 0)
		return;
	if(entropytype == DEFAULT)
	{
		unsigned int nLoop, lengthNormalIndex;
		lengthNormalIndex = int(log((double)numberOfNormal-1)/log(2.0)) + 1;
		for(nLoop = 0; nLoop < COORD_DIVISION * numberOfNormalIndex; nLoop++)
		{
			outIFS_->normalIndex_[nLoop] = (int)(ReadBitsFromStorage(lengthNormalIndex));
		}
	}
	else if(entropytype == BP)
	{
		unsigned int nLoop, prefixSize, nBPL, nPayload, nSign;
		int previousQuant[3],difValue, i,  nQuantValue;
		nQuantValue=0;
		prefixSize = ReadBitsFromStorage(5);
		for(i = 0; i < 3; i++)
		{
			previousQuant[i] = 0;
		}
		for(nLoop = 0; nLoop < 3 * numberOfNormalIndex; nLoop++)
		{

			nBPL=ReadBitsFromStorage(prefixSize);
			if(nBPL > 2)
			{
				nPayload = ReadBitsFromStorage(nBPL-1);
			}
			else
			{
				nPayload = 0;
			}

			difValue = BPLTable[nBPL] + nPayload;
			if(difValue != 0)
			{
				nSign = ReadBitsFromStorage(1);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}

			nQuantValue = difValue+previousQuant[nLoop%3];
			previousQuant[nLoop % 3] = nQuantValue;
			outIFS_->normalIndex_[nLoop] = nQuantValue;
		}

	}
	else
	{
		unsigned int nLoop, prefixSize, nBPL, nPayload, nSign;
		int previousQuant[3],difValue, i,  nQuantValue;
		nQuantValue=0;
		for(i = 0; i < 3; i++)
		{
			previousQuant[i] = 0;
		}
		unsigned int code_bytes = 0;
		ReadFromStream(m_pReadStream, &code_bytes, sizeof(unsigned int));
		unsigned char * code_buffer = new unsigned char [code_bytes];                                                   
		ReadFromStream(m_pReadStream, code_buffer, sizeof(unsigned char)*code_bytes);
		
		Adaptive_Data_Model mModel(1024);
		Adaptive_Data_Model mSign (2);
		Adaptive_Data_Model mhasnext (2);
		Arithmetic_Codec acd(code_bytes, code_buffer);
		acd.start_decoder();	// initialize decoder

		int data[4];
		data[0] = 1;
		data[1] = 1<<10;
		data[2] = 1<<20;
		data[3] = 1<<30;
		int count = 0;
		for(nLoop = 0; nLoop < 3 * numberOfNormalIndex; nLoop++)
		{

			difValue = 0;

			while(1)
			{
				difValue += data[count] * acd.decode(mModel);
				count++;
				int temp = acd.decode(mhasnext);

				if( temp ==0)
					break;
			}
			count = 0;

			if(difValue != 0)
			{
				nSign = acd.decode(mSign);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}
			nQuantValue = difValue+previousQuant[nLoop%3];
			
			previousQuant[nLoop % 3] = nQuantValue;
			
			//dmlab
			//coord.push_back(InverseQuantization(nQuantValue,(nLoop%3)));
			outIFS_->normalIndex_[nLoop] = nQuantValue;

			 //~dmlab
		}
	}

}


void QBCRDecoder::DecoderColorIndex()
{
	if(numberOfColor == 0)
		return;
	if(entropytype == DEFAULT)
	{
		unsigned int nLoop, lengthColorIndex;
		lengthColorIndex = int(log((double)numberOfColor-1)/log(2.0)) + 1;
		for(nLoop = 0; nLoop < COORD_DIVISION * numberOfColorIndex; nLoop++)
		{
			outIFS_->colorIndex_[nLoop] = (int)(ReadBitsFromStorage(lengthColorIndex));
		}
	}
	else if(entropytype == BP)
	{
		unsigned int nLoop, prefixSize, nBPL, nPayload, nSign;
		int previousQuant[3],difValue, i,  nQuantValue;
		nQuantValue=0;
		prefixSize = ReadBitsFromStorage(5);
		for(i = 0; i < 3; i++)
		{
			previousQuant[i] = 0;
		}
		for(nLoop = 0; nLoop < 3 * numberOfColorIndex; nLoop++)
		{

			nBPL=ReadBitsFromStorage(prefixSize);
			if(nBPL > 2)
			{
				nPayload = ReadBitsFromStorage(nBPL-1);
			}
			else
			{
				nPayload = 0;
			}

			difValue = BPLTable[nBPL] + nPayload;
			if(difValue != 0)
			{
				nSign = ReadBitsFromStorage(1);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}

			nQuantValue = difValue+previousQuant[nLoop%3];
			previousQuant[nLoop % 3] = nQuantValue;
			outIFS_->colorIndex_[nLoop] = nQuantValue;
		}

	}
	else
	{
		unsigned int nLoop, prefixSize, nBPL, nPayload, nSign;
		int previousQuant[3],difValue, i,  nQuantValue;
		nQuantValue=0;
		for(i = 0; i < 3; i++)
		{
			previousQuant[i] = 0;
		}
		unsigned int code_bytes = 0;
		ReadFromStream(m_pReadStream, &code_bytes, sizeof(unsigned int));
		unsigned char * code_buffer = new unsigned char [code_bytes];                                                   
		ReadFromStream(m_pReadStream, code_buffer, sizeof(unsigned char)*code_bytes);
		
		Adaptive_Data_Model mModel(1024);
		Adaptive_Data_Model mSign (2);
		Adaptive_Data_Model mhasnext (2);
		Arithmetic_Codec acd(code_bytes, code_buffer);
		acd.start_decoder();	// initialize decoder

		int data[4];
		data[0] = 1;
		data[1] = 1<<10;
		data[2] = 1<<20;
		data[3] = 1<<30;
		int count = 0;
		for(nLoop = 0; nLoop < 3 * numberOfColorIndex; nLoop++)
		{

			difValue = 0;

			while(1)
			{
				difValue += data[count] * acd.decode(mModel);
				count++;
				int temp = acd.decode(mhasnext);

				if( temp ==0)
					break;
			}
			count = 0;

			if(difValue != 0)
			{
				nSign = acd.decode(mSign);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}
			nQuantValue = difValue+previousQuant[nLoop%3];
			
			previousQuant[nLoop % 3] = nQuantValue;
			
			//dmlab
			//coord.push_back(InverseQuantization(nQuantValue,(nLoop%3)));
			outIFS_->colorIndex_[nLoop] = nQuantValue;

			 //~dmlab
		}
	}

}

void QBCRDecoder::DecoderTexCoordIndex()
{

	if(numberOfTexCoord == 0)
		return;

	if(entropytype == DEFAULT)
	{
		unsigned int nLoop, lengthTexCoordIndex;
		lengthTexCoordIndex = int(log((double)numberOfTexCoord-1)/log(2.0)) + 1;
		for(nLoop = 0; nLoop < 3 * numberOfTexCoordIndex; nLoop++)
		{
			outIFS_->texCoordIndex_[nLoop] = (int)(ReadBitsFromStorage(lengthTexCoordIndex));
		}
	}
	else if(entropytype == BP)
	{
		unsigned int nLoop, prefixSize, nBPL, nPayload, nSign;
		int previousQuant[3],difValue, i,  nQuantValue;
		nQuantValue=0;
		prefixSize = ReadBitsFromStorage(5);
		for(i = 0; i < 3; i++)
		{
			previousQuant[i] = 0;
		}
		for(nLoop = 0; nLoop < 3 * numberOfTexCoordIndex; nLoop++)
		{

			nBPL=ReadBitsFromStorage(prefixSize);
			if(nBPL > 2)
			{
				nPayload = ReadBitsFromStorage(nBPL-1);
			}
			else
			{
				nPayload = 0;
			}

			difValue = BPLTable[nBPL] + nPayload;
			if(difValue != 0)
			{
				nSign = ReadBitsFromStorage(1);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}

			nQuantValue = difValue+previousQuant[nLoop%3];
			previousQuant[nLoop % 3] = nQuantValue;
			outIFS_->texCoordIndex_[nLoop] = nQuantValue;
		}

	}
	else
	{
		unsigned int nLoop, prefixSize, nBPL, nPayload, nSign;
		int previousQuant[3],difValue, i,  nQuantValue;
		nQuantValue=0;
		for(i = 0; i < 3; i++)
		{
			previousQuant[i] = 0;
		}
		unsigned int code_bytes = 0;
		ReadFromStream(m_pReadStream, &code_bytes, sizeof(unsigned int));
		unsigned char * code_buffer = new unsigned char [code_bytes];                                                   
		ReadFromStream(m_pReadStream, code_buffer, sizeof(unsigned char)*code_bytes);
		
		Adaptive_Data_Model mModel(1024);
		Adaptive_Data_Model mSign (2);
		Adaptive_Data_Model mhasnext (2);
		Arithmetic_Codec acd(code_bytes, code_buffer);
		acd.start_decoder();	// initialize decoder

		int data[4];
		data[0] = 1;
		data[1] = 1<<10;
		data[2] = 1<<20;
		data[3] = 1<<30;
		int count = 0;
		for(nLoop = 0; nLoop < 3 * numberOfTexCoordIndex; nLoop++)
		{

			difValue = 0;

			while(1)
			{
				difValue += data[count] * acd.decode(mModel);
				count++;
				int temp = acd.decode(mhasnext);

				if( temp ==0)
					break;
			}
			count = 0;

			if(difValue != 0)
			{
				nSign = acd.decode(mSign);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}
			nQuantValue = difValue+previousQuant[nLoop%2];
			
			previousQuant[nLoop % 2] = nQuantValue;
			
			//dmlab
			//coord.push_back(InverseQuantization(nQuantValue,(nLoop%3)));
			outIFS_->texCoordIndex_[nLoop] = nQuantValue;

			 //~dmlab
		}
	}
}



void QBCRDecoder::ComposeIndexedFaceSet(IndexedFaceSet &_myIFS)
{
/*

	unsigned int v, p, h ;
	unsigned int k=0;

	_myIFS.nCoord_ = numberOfCoord;
	_myIFS.coord_ = new float [3 * numberOfCoord];
	
	for (unsigned int v = 0; v < 3 * numberOfCoord; v++) {
		_myIFS.coord_[v] = coord[v];
	}
	
	_myIFS.nCoordIndex_ = numberOfCoordIndex;
	_myIFS.coordIndex_ = new int [numberOfCoordIndex*3];
	

	for (unsigned int p = 0; p < numberOfCoordIndex; p++) {
		for (unsigned int h = 0; h <3; h++) {
			_myIFS.coordIndex_[p*3+h] = coordIndex[p*3+h];
		}
	}

	if(numberOfNormal != 0)
	{
		_myIFS.nNormal_ = numberOfNormal;
		_myIFS.normal_ = new float [3 * numberOfNormal];
		
		for (unsigned int v = 0; v < 3 * numberOfNormal; v++) {
			_myIFS.normal_[v] = normal[v];
		}
	}
	
	if(numberOfNormalIndex != 0)
	{
		_myIFS.nNormalIndex_ = numberOfNormalIndex;
		_myIFS.normalIndex_ = new int [numberOfNormalIndex*3];
		

		for (unsigned int p = 0; p < numberOfNormalIndex; p++) {
			for (unsigned int h = 0; h <3; h++) {
				_myIFS.normalIndex_[p*3+h] = normalIndex[p*3+h];
			}
		}
	}

	if(numberOfColor != 0)
	{
		_myIFS.nColor_ = numberOfColor;
		_myIFS.color_ = new float [3 * numberOfColor];
		
		for (unsigned int v = 0; v < 3 * numberOfColor; v++) {
			_myIFS.color_[v] = color[v];
		}
	}

	if(numberOfColorIndex != 0)
	{
		_myIFS.nColorIndex_ = numberOfColorIndex;
		_myIFS.colorIndex_ = new int [numberOfColorIndex*3];
		

		for (unsigned int p = 0; p < numberOfColorIndex; p++) {
			for (unsigned int h = 0; h <3; h++) {
				_myIFS.colorIndex_[p*3+h] = colorIndex[p*3+h];
			}
		}
	}

	if(numberOfTexCoord != 0)
	{
		_myIFS.nTexCoord_ = numberOfTexCoord;
		_myIFS.texCoord_ = new float [3 * numberOfTexCoord];
		
		for (unsigned int v = 0; v < 3 * numberOfTexCoord; v++) {
			_myIFS.texCoord_[v] = texCoord[v];
		}
	}


	if(numberOfTexCoordIndex != 0)
	{
		_myIFS.nTexCoordIndex_ = numberOfTexCoordIndex;
		_myIFS.texCoordIndex_ = new int [numberOfTexCoordIndex*3];
		

		for (unsigned int p = 0; p < numberOfTexCoordIndex; p++) {
			for (unsigned int h = 0; h <3; h++) {
				_myIFS.texCoordIndex_[p*3+h] = texCoordIndex[p*3+h];
			}
		}
	}
	coord.clear();
	coordIndex.clear();
	normal.clear();
	normalIndex.clear();
	texCoord.clear();
	texCoordIndex.clear();
	color.clear();
	colorIndex.clear();
	if(readBitstream != NULL)
	{
		delete [] readBitstream;
		readBitstream = NULL;
	}
	unCurrentBitPosition = BS_BIT_MAX_VALUE;
	unCurrentBufferPosition = 0;
*/

}

float QBCRDecoder::InverseQuantization(unsigned quantCoord, float quantRange, float quantMin, float quantValue)
{
	float data;
	data = ((float)quantCoord/quantValue)*quantRange+quantMin;
	return data;
	
}

int QBCRDecoder::ReadData(char *filename)
{
	FILE* m_inputFile;			
	m_inputFile = fopen(filename,"rb");


	fseek(m_inputFile,0L, SEEK_END);
	unsigned int filesize = ftell(m_inputFile);
	readBitstream = new unsigned char [filesize];
	rewind(m_inputFile);
	fread(readBitstream, sizeof(unsigned char),filesize,m_inputFile );
	m_pReadStream=readBitstream;

	fclose(m_inputFile);

	return 1;
}

unsigned int QBCRDecoder::ReadBitsFromStorage(unsigned int unReadLength)
{
	unsigned int n = unReadLength;
	int c=(n + 14 - unCurrentBitPosition) >> 3;

	union { 
		long long l;
		unsigned char c[8];
	} buf;
	buf.l = 0;

	int i=0;
	while(n) {
		i++;
		buf.c[c - i] = *(m_pReadStream + unCurrentBufferPosition);
		if(unCurrentBitPosition >= n) {
			unCurrentBitPosition -= n;
			break;
		} else {
			n -= unCurrentBitPosition + 1;
			unCurrentBitPosition = 7;
			unCurrentBufferPosition++;
		}
	}

	return (buf.l >> ((unCurrentBitPosition + 1) % 8)) & ~(~static_cast<unsigned int>(0) << unReadLength);
}


void QBCRDecoder::BPLInit()
{
	int i;
	BPLTable = new int [32];
	BPLTable[0] = 0;
	BPLTable[1] = 1;
	BPLTable[2] = 2;

	for(i = 3; i <32; i++)
	{
		BPLTable[i] = pow(2.0,(double)(i-1))-1;
	}

}
