//-------------------------------------------------------------------------
// 
// Copyright ETRI, Hanyang University 2008, 2009
// Licensed Materials, Program Property of ETRI & Hanyang University
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.
// ETRI and Hanyang University retain full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//    AUTHORS:
//    Seung Wook Lee (ETRI)
//    Daiyong Kim (Hanyang University)
//    KyoungSoo Son (Hanyang University)
//
//    http://www.etri.re.kr/
//    http://dmlab.hanyang.ac.kr/
//  
//    Seung Wook Lee     (tajinet@etri.re.kr)  
//    Daiyong Kim        (dykim@dmlab.hanyang.ac.kr)    
//    KyoungSoo Son      (bonokensin@dmlab.hanyang.ac.kr)
//
//-------------------------------------------------------------------------



#include <iostream>
#include <time.h>
#include <string>
#include <iostream>
#include <fstream>
#include <sstream> 
#include <math.h>
#include <vector>
#include <map>
using namespace std;

#include <assert.h>

#include "IndexedFaceSet.h"
#include "QBCR_Encoder.h"
#include "utils.h"
#include "arithmetic_codec.h"

#define MAX_VLAUE 1000000
#define MIN_VLAUE -1000000
#define INIT_UCHAR 0x00
#define LEN_BYTE	8
#define LEN_QP		5
#define LEN_BIT		1
#define DEFAULT		0
#define DIFF		1
#define BP			1
#define AC			2
#define LEN_PREFIX	5


#define SC3DMCStartCode 0x00001F1


QBCREncoder::QBCREncoder()
{
	dumpBit =INIT_UCHAR;
	bitPostion = 0;
	MaxDifferentialForGeometry = 0;
	MaxDifferentialForNormal = 0;
	MaxDifferentialForColor = 0;
	MaxDifferentialForTexCoord = 0;
	BPLTable = NULL;
	payloadTable =NULL;

}

QBCREncoder::~QBCREncoder()
{
	if(BPLTable != NULL)
	{

		delete [ ] BPLTable;
		delete [ ] payloadTable;
	}

}


void QBCREncoder::EncodeDataBuffer(unsigned char *compressedStream, unsigned int &compressedStreamSize)
{
	Initialization();

	QuantizationGeometry(QPforGeomerty);
	if(nNormal != 0)
	{
		Normalization();
		QuantizationNormal(QPforNormal);		
	}
	if(nColor != 0)
		QuantizationColor(QPforColor);		
	if(nTexCoord != 0)
		QuantizationTexCoord(QPforTexCoord);

	m_compressedStream = compressedStream; 
	m_compressedStreamSize = compressedStreamSize;

	if(entropytype == BP)
	{
		BPCInit();
	}

	Encode_Header();
	Encode_Geometry();
	Encode_Connectivity();
	Encode_Normal();						
	Encode_NormalIndex();					
	Encode_Texture();						
	Encode_TextureIndex();					
	Encode_Color();						
	Encode_ColorIndex();
	Flush();

	compressedStreamSize = m_compressedStreamSize;

}

void QBCREncoder::VectorToBuffer(unsigned char *compressedStream, unsigned int &compressedStreamSize)
{
	for(int i=0; i< dumpBitstream.size(); i++)
	{
		compressedStream[compressedStreamSize++] = dumpBitstream[i];

	}
}

void QBCREncoder::Initialization()
{

	codingtype = DIFF;
	entropytype = AC;
	//entropytype = AC;
	
	
	//codingtype = DEFAULT;
	//entropytype = DEFAULT;

	nOtherAtrribute = 0;
	nOtherAtrributeIndex = 0;

	
	nCoord = m_ifs->GetNCoord();
	nCoordIndex = m_ifs->GetNCoordIndex();
	nNormal = m_ifs->GetNNormal();
	nNormalIndex = m_ifs->GetNNormalIndex();
	nColor = m_ifs->GetNColor();
	nColorIndex = m_ifs->GetNColorIndex();
	nTexCoord = m_ifs->GetNTexCoord();
	nTexCoordIndex = m_ifs->GetNTexCoordIndex();

	Coord = m_ifs->GetCoord();
	CoordIndex = m_ifs->GetCoordIndex();
	Normal = m_ifs->GetNormal();								
	NormalIndex = m_ifs->GetNormalIndex();
	TexCoord = m_ifs->GetTexCoord();					
	TexCoordIndex = m_ifs->GetTexCoordIndex();					
	Color = m_ifs->GetColor();
	ColorIndex = m_ifs->GetColorIndex();

	
	for(int i = 0; i < 3; i++)
	{
		
		quantMinGeometry[i] = m_ifs->minCoord[i];
		quantMinNormal[i] = m_ifs->minNormal[i];
		quantMinColor[i] = m_ifs->minColor[i];	
		quantMinOtherAttribute[i]= 0.0;
		quantRangeGeometry[i] = m_ifs->rangeCoord[i];
		quantRangeNormal[i] = m_ifs->rangeNormal[i];
		quantRangeColor[i] = m_ifs->rangeColor[i];
	
	}
	for(int i = 0; i < 2; i++)
	{
		quantMinTexCoord[i] = m_ifs->minTexCoord[i];
		quantRangeTexCoord[i] = m_ifs->rangeTexCoord[i];	
	}


					
	quantRangeOtherAttribute = 0;		

	QPforGeomerty = m_params.m_nQCoordBits;
	QPforNormal = m_params.m_nQNormalBits;
	QPforColor = m_params.m_nQColorBits;
	QPforTexCoord = m_params.m_nQTexCoordBits;

}

void QBCREncoder::ParsingConfigure(char *config)
{
	char* tgbuf = NULL;
	int count = 0;
	int tempQP[4] = {0};
	while(config != NULL)
	{
		tgbuf = strtok(config, " ");
		tempQP[count++] = atoi(tgbuf);
		if(count >= 4)
			break;
	}
	QPforGeomerty = tempQP[0];
	QPforNormal = tempQP[1];
	QPforColor = tempQP[2];
	QPforTexCoord = tempQP[3];
	
}


void QBCREncoder:: Encode_CommonHeader(IndexedFaceSet &m_ifs, EncodingParams &m_params)
{

}

int QBCREncoder::Encode_QBCR(IndexedFaceSet &myIFS, EncodingParams &m_params)
{

	Initialization();
	//ParsingConfigure(config);
	// Quantization Part;
	QuantizationGeometry(QPforGeomerty);
	if(nNormal != 0)
	{
		Normalization();
		QuantizationNormal(QPforNormal);		
	}
	if(nColor != 0)
		QuantizationColor(QPforColor);		
	if(nTexCoord != 0)
		QuantizationTexCoord(QPforTexCoord);	

	// Encoding Part
	Encode_CommonHeader(myIFS,m_params);
	Encode_Header();
	Encode_Geometry();
	Encode_Connectivity();
	Encode_Normal();						
	Encode_NormalIndex();					
	Encode_Texture();						
	Encode_TextureIndex();					
	Encode_Color();						
	Encode_ColorIndex();
	Flush();


	return 1;
}

void QBCREncoder::Encode_Header()
{
	Write2Stream(m_compressedStream, m_compressedStreamSize, &codingtype, sizeof(unsigned char));
	Write2Stream(m_compressedStream, m_compressedStreamSize, &entropytype, sizeof(unsigned char));
}

void QBCREncoder::Encode_Geometry()
{
	int symbol = 0;
	unsigned int attr_num;
	if(entropytype == DEFAULT)
	{
		for(attr_num = 0 ; attr_num < nCoord*3; attr_num++)
		{
			symbol = quantCoord[attr_num];
			Insert_Data(symbol,QPforGeomerty);
		}
	}
	else if(entropytype == BP)
	{
		unsigned int nMaxCheckBPL = int(log((float)(MaxDifferentialForGeometry+1)) / log((float)2)) +1;
		unsigned int prefix_size = int(log((float)(nMaxCheckBPL+1)) / log((float)2)) +1;	

		Insert_Data(prefix_size, LEN_PREFIX);

		int nSymbol, nSign;
		unsigned int i;
		for(i = 0 ; i < quantCoord.size(); i++)
		{

			nSymbol = quantCoord[i];
		
			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;


			Insert_Data(BPLTable[nSymbol],prefix_size);

			if(payloadTable[nSymbol] != -1)
				Insert_Data(payloadTable[nSymbol], (BPLTable[nSymbol]-1));
			if(nSign != 2)
			{
				Insert_Data(nSign, 1);
			}
		
		}
	}
	else
	{
		int nSymbol, nSign;
		unsigned int i;
		unsigned int NMAX = (int) quantCoord.size() * 10 + 100;
		Adaptive_Data_Model mModel(1<<QPforGeomerty);
		Adaptive_Data_Model mSign (2);
		Arithmetic_Codec ace(NMAX, NULL);
		ace.start_encoder();
		for(i = 0 ; i < quantCoord.size(); i++)
		{

			nSymbol = quantCoord[i];
		
			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;
			ace.encode(nSymbol, mModel);
			if(nSign != 2)
				ace.encode(nSign, mSign);

		}
		ace.write_to_mem(m_compressedStream, m_compressedStreamSize);

	}
}

void QBCREncoder::BPCInit()
{

	int maxValue = 0;
	if(nCoord > 0)
	{
		maxValue = (maxValue > (1<<QPforGeomerty) ) ? maxValue : (1<<QPforGeomerty);
	}
	if(nNormal > 0)
	{
		maxValue = (maxValue > (1<<QPforNormal) ) ? maxValue : (1<<QPforNormal);
	}
	if(nColor > 0)
	{
		maxValue = (maxValue > (1<<QPforColor) ) ? maxValue : (1<<QPforColor);
	}
	if(nTexCoord > 0)
	{
		maxValue = (maxValue > (1<<QPforTexCoord) ) ? maxValue : (1<<QPforTexCoord);
	}
	if(nCoordIndex > 0)
	{
		maxValue = (maxValue > (nCoord<<1) ) ? maxValue : (nCoord<<1);
	}
	if(nNormalIndex > 0)
	{
		maxValue = (maxValue > (nNormal<<1) ) ? maxValue : (nNormal<<1);
	}
	if(nColorIndex > 0)
	{
		maxValue = (maxValue > (nColor<<1) ) ? maxValue : (nColor<<1);
	}
	if(nTexCoordIndex > 0)
	{
		maxValue = (maxValue > (nTexCoord<<1) ) ? maxValue : (nTexCoord<<1);
	}
	BPLTable = new int [maxValue];
	payloadTable = new int [maxValue];
	int i;
	for(i = 0 ; i < 3; i++)
	{
		BPLTable[i] = i;
		payloadTable[i] = -1;
	}
	int payloadLength = 4;
	int minValueInPresentBPL = 3;
	int presentBPL = 3;
	for( ; i<maxValue; i++)
	{
		if(i == payloadLength+minValueInPresentBPL)
		{
			minValueInPresentBPL += payloadLength;
			payloadLength *= 2;
			presentBPL++;
		}
		BPLTable[i] = presentBPL;
		payloadTable[i] = i - minValueInPresentBPL;
	}

		

}

void QBCREncoder::Encode_Connectivity()
{
	int symbol = 0;
	unsigned int attr_num;

	if(entropytype == DEFAULT)
	{
		unsigned int lengthCoordIndex;
		lengthCoordIndex = int(log((double)nCoord-1)/log(2.0)) + 1;
		for(attr_num = 0 ; attr_num < nCoordIndex*3; attr_num++)
		{
			Insert_Data(CoordIndex[attr_num],lengthCoordIndex);
		}
	}
	else if(entropytype == BP)
	{
		unsigned int nMaxCheckBPL = int(log((float)(nCoord-1)) / log((float)2)) +1;
		unsigned int prefix_size = int(log((float)(nMaxCheckBPL+1)) / log((float)2)) +1;	
		int nSymbol, nSign;
		unsigned int i;

		Insert_Data(prefix_size, LEN_PREFIX);
		int previousQuantValue[3];
		for(attr_num = 0; attr_num < 3; attr_num++)
		{
			previousQuantValue[attr_num] = 0;
		}
		for(i = 0 ; i < nCoordIndex*3; i++)
		{

			nSymbol = CoordIndex[i]-previousQuantValue[i%3];
			previousQuantValue[i%3] = CoordIndex[i];
		
			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;


			Insert_Data(BPLTable[nSymbol],prefix_size);

			if(payloadTable[nSymbol] != -1)
				Insert_Data(payloadTable[nSymbol], (BPLTable[nSymbol]-1));
			if(nSign != 2)
			{
				Insert_Data(nSign, 1);
			}
		
		}

	}
	else
	{
		unsigned int attr_num;
		int previousQuantValue[3];
		for(attr_num = 0; attr_num < 3; attr_num++)
		{
			previousQuantValue[attr_num] = 0;
		}

		int nSymbol, nSign, input;
		unsigned int i;
		int lengthCoordIndex = int(log((double)nCoord-1)/log(2.0)) + 1;
		unsigned int NMAX = (int) nCoordIndex*3 * lengthCoordIndex + 100;
		Adaptive_Data_Model mModel(1024);
		Adaptive_Data_Model mhasnext (2);
		Adaptive_Data_Model mSign (2);
		Arithmetic_Codec ace(NMAX, NULL);
		ace.start_encoder();

		for(attr_num = 0 ; attr_num < nCoordIndex*3; attr_num++)
		{
			nSymbol = CoordIndex[attr_num] - previousQuantValue[attr_num%3];
			previousQuantValue[attr_num%3] =CoordIndex[attr_num];

			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;
			while(1)
			{
				ace.encode(nSymbol%1024, mModel);
				nSymbol/=1024;
				if(nSymbol)
					ace.encode(1, mhasnext);
				else
					break;
			}
			ace.encode(0, mhasnext);
			if(nSign != 2)
				ace.encode(nSign, mSign);
			
		}
		ace.write_to_mem(m_compressedStream, m_compressedStreamSize);

	}

}

void QBCREncoder::Encode_Normal()
{
	if(nNormal == 0)
		return;

	int symbol = 0;
	unsigned int attr_num;
	if(entropytype == DEFAULT)
	{
		for(attr_num = 0 ; attr_num < nNormal*3; attr_num++)
		{
			symbol = quantNormal[attr_num];
			Insert_Data(symbol,QPforNormal);
		}
	}
	else if(entropytype == BP)
	{
		unsigned int nMaxCheckBPL = int(log((float)(MaxDifferentialForNormal+1)) / log((float)2)) +1;
		unsigned int prefix_size = int(log((float)(nMaxCheckBPL+1)) / log((float)2)) +1;	

		Insert_Data(prefix_size, LEN_PREFIX);

		int nSymbol, nSign;
		unsigned int i;
		for(i = 0 ; i < nNormal*3; i++)
		{

			nSymbol = quantNormal[i];
		
			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;


			Insert_Data(BPLTable[nSymbol],prefix_size);

			if(payloadTable[nSymbol] != -1)
				Insert_Data(payloadTable[nSymbol], (BPLTable[nSymbol]-1));
			if(nSign != 2)
			{
				Insert_Data(nSign, 1);
			}
		
		}
	}
	else
	{
		int nSymbol, nSign;
		unsigned int i;
		unsigned int NMAX = (int) quantNormal.size() * 10 + 100;
		Adaptive_Data_Model mModel(1<<QPforNormal);
		Adaptive_Data_Model mSign (2);
		Arithmetic_Codec ace(NMAX, NULL);
		ace.start_encoder();
		for(i = 0 ; i < nNormal*3; i++)
		{

			nSymbol = quantNormal[i];
		
			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;
			ace.encode(nSymbol, mModel);
			if(nSign != 2)
				ace.encode(nSign, mSign);

		}
		ace.write_to_mem(m_compressedStream, m_compressedStreamSize);

	}

}
	
void QBCREncoder::Encode_NormalIndex()
{
	if(nNormalIndex == 0)
		return;
	int symbol = 0;
	unsigned int attr_num;


	if(entropytype == DEFAULT)
	{
		unsigned int lengthNormalIndex;
		lengthNormalIndex = int(log((double)nNormal-1)/log(2.0)) + 1;
		for(attr_num = 0 ; attr_num < nNormalIndex*3; attr_num++)
		{
			Insert_Data(NormalIndex[attr_num],lengthNormalIndex);
		}
	}
	else if(entropytype == BP)
	{
		unsigned int nMaxCheckBPL = int(log((float)(nNormal-1)) / log((float)2)) +1;
		unsigned int prefix_size = int(log((float)(nMaxCheckBPL+1)) / log((float)2)) +1;	
		int nSymbol, nSign;
		unsigned int i;

		Insert_Data(prefix_size, LEN_PREFIX);
		int previousQuantValue[3];
		for(attr_num = 0; attr_num < 3; attr_num++)
		{
			previousQuantValue[attr_num] = 0;
		}
		for(i = 0 ; i < nNormalIndex*3; i++)
		{

			nSymbol = NormalIndex[i]-previousQuantValue[i%3];
			previousQuantValue[i%3] = NormalIndex[i];
		
			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;


			Insert_Data(BPLTable[nSymbol],prefix_size);

			if(payloadTable[nSymbol] != -1)
				Insert_Data(payloadTable[nSymbol], (BPLTable[nSymbol]-1));
			if(nSign != 2)
			{
				Insert_Data(nSign, 1);
			}
		
		}

	}
	else
	{
		unsigned int attr_num;
		int previousQuantValue[3];
		for(attr_num = 0; attr_num < 3; attr_num++)
		{
			previousQuantValue[attr_num] = 0;
		}

		int nSymbol, nSign, input;
		unsigned int i;
		int lengthNormalIndex = int(log((double)nNormal-1)/log(2.0)) + 1;
		unsigned int NMAX = (int) nNormalIndex*3 * lengthNormalIndex + 100;
		Adaptive_Data_Model mModel(1024);
		Adaptive_Data_Model mhasnext (2);
		Adaptive_Data_Model mSign (2);
		Arithmetic_Codec ace(NMAX, NULL);
		ace.start_encoder();

		for(attr_num = 0 ; attr_num < nNormalIndex*3; attr_num++)
		{
			nSymbol = NormalIndex[attr_num] - previousQuantValue[attr_num%3];
			previousQuantValue[attr_num%3] =NormalIndex[attr_num];

			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;
			while(1)
			{
				ace.encode(nSymbol%1024, mModel);
				nSymbol/=1024;
				if(nSymbol)
					ace.encode(1, mhasnext);
				else
					break;
			}
			ace.encode(0, mhasnext);
			if(nSign != 2)
				ace.encode(nSign, mSign);
			
		}
		ace.write_to_mem(m_compressedStream, m_compressedStreamSize);

	}
	
}

void QBCREncoder::Encode_Texture()
{
	if(nTexCoord == 0)
		return;
	int symbol = 0;
	unsigned int attr_num;

	if(entropytype == DEFAULT)
	{
		for(attr_num = 0 ; attr_num < nTexCoord*2; attr_num++)
		{
			symbol = quantTexCoord[attr_num];
			Insert_Data(symbol,QPforTexCoord);
		}
	}
	else if(entropytype == BP)
	{
		unsigned int nMaxCheckBPL = int(log((float)(MaxDifferentialForTexCoord+1)) / log((float)2)) +1;
		unsigned int prefix_size = int(log((float)(nMaxCheckBPL+1)) / log((float)2)) +1;	

		Insert_Data(prefix_size, LEN_PREFIX);

		int nSymbol, nSign;
		unsigned int i;
		for(i = 0 ; i < nTexCoord*2; i++)
		{

			nSymbol = quantTexCoord[i];
		
			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;


			Insert_Data(BPLTable[nSymbol],prefix_size);

			if(payloadTable[nSymbol] != -1)
				Insert_Data(payloadTable[nSymbol], (BPLTable[nSymbol]-1));
			if(nSign != 2)
			{
				Insert_Data(nSign, 1);
			}
		
		}
	}
	else
	{
		int nSymbol, nSign;
		unsigned int i;
		unsigned int NMAX = (int) quantTexCoord.size() * 10 + 100;
		Adaptive_Data_Model mModel(1<<QPforTexCoord);
		Adaptive_Data_Model mSign (2);
		Arithmetic_Codec ace(NMAX, NULL);
		ace.start_encoder();
		for(i = 0 ; i < nTexCoord*2; i++)
		{

			nSymbol = quantTexCoord[i];
		
			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;
			ace.encode(nSymbol, mModel);
			if(nSign != 2)
				ace.encode(nSign, mSign);

		}
		ace.write_to_mem(m_compressedStream, m_compressedStreamSize);

	}
}

void QBCREncoder::Encode_TextureIndex()
{
	if(nTexCoordIndex == 0)
		return;
	int symbol = 0;
	unsigned int attr_num;

	
	if(entropytype == DEFAULT)
	{
		unsigned int lengthTexCoordIndex;
		lengthTexCoordIndex = int(log((double)nTexCoord-1)/log(2.0)) + 1;

		for(attr_num = 0 ; attr_num < nTexCoordIndex*3.; attr_num++)
		{
			Insert_Data(TexCoordIndex[attr_num],lengthTexCoordIndex);
		}
	}
	else if(entropytype == BP)
	{
		unsigned int nMaxCheckBPL = int(log((float)(nTexCoord-1)) / log((float)2)) +1;
		unsigned int prefix_size = int(log((float)(nMaxCheckBPL+1)) / log((float)2)) +1;	
		int nSymbol, nSign;
		unsigned int i;

		Insert_Data(prefix_size, LEN_PREFIX);
		int previousQuantValue[3];
		for(attr_num = 0; attr_num < 3; attr_num++)
		{
			previousQuantValue[attr_num] = 0;
		}
		for(i = 0 ; i < nTexCoordIndex*3; i++)
		{

			nSymbol = TexCoordIndex[i]-previousQuantValue[i%3];
			previousQuantValue[i%3] = TexCoordIndex[i];
		
			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;


			Insert_Data(BPLTable[nSymbol],prefix_size);

			if(payloadTable[nSymbol] != -1)
				Insert_Data(payloadTable[nSymbol], (BPLTable[nSymbol]-1));
			if(nSign != 2)
			{
				Insert_Data(nSign, 1);
			}
		
		}

	}
	else
	{
		unsigned int attr_num;
		int previousQuantValue[3];
		for(attr_num = 0; attr_num < 3; attr_num++)
		{
			previousQuantValue[attr_num] = 0;
		}

		int nSymbol, nSign, input;
		unsigned int i;
		int lengthTexCoordIndex = int(log((double)nTexCoord-1)/log(2.0)) + 1;
		unsigned int NMAX = (int) nTexCoordIndex*2 * lengthTexCoordIndex + 100;
		Adaptive_Data_Model mModel(1024);
		Adaptive_Data_Model mhasnext (2);
		Adaptive_Data_Model mSign (2);
		Arithmetic_Codec ace(NMAX, NULL);
		ace.start_encoder();

		for(attr_num = 0 ; attr_num < nTexCoordIndex*3; attr_num++)
		{
			nSymbol = TexCoordIndex[attr_num] - previousQuantValue[attr_num%3];
			previousQuantValue[attr_num%3] =TexCoordIndex[attr_num];

			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;
			while(1)
			{
				ace.encode(nSymbol%1024, mModel);
				nSymbol/=1024;
				if(nSymbol)
					ace.encode(1, mhasnext);
				else
					break;
			}
			ace.encode(0, mhasnext);
			if(nSign != 2)
				ace.encode(nSign, mSign);
			
		}
		ace.write_to_mem(m_compressedStream, m_compressedStreamSize);

	}

}

void QBCREncoder::Encode_Color()
{
	if(nColor == 0)
		return;
	int symbol = 0;
	unsigned int attr_num;

	if(entropytype == DEFAULT)
	{
		for(attr_num = 0 ; attr_num < nColor*3; attr_num++)
		{
			symbol = quantColor[attr_num];
			Insert_Data(symbol,QPforColor);
		}
	}
	else if(entropytype == BP)
	{
		unsigned int nMaxCheckBPL = int(log((float)(MaxDifferentialForColor+1)) / log((float)2)) +1;
		unsigned int prefix_size = int(log((float)(nMaxCheckBPL+1)) / log((float)2)) +1;	

		Insert_Data(prefix_size, LEN_PREFIX);

		int nSymbol, nSign;
		unsigned int i;
		for(i = 0 ; i < nColor*3; i++)
		{

			nSymbol =  quantColor[i];
		
			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;


			Insert_Data(BPLTable[nSymbol],prefix_size);

			if(payloadTable[nSymbol] != -1)
				Insert_Data(payloadTable[nSymbol], (BPLTable[nSymbol]-1));
			if(nSign != 2)
			{
				Insert_Data(nSign, 1);
			}
		
		}
	}
	else
	{
		int nSymbol, nSign;
		unsigned int i;
		unsigned int NMAX = (int) quantColor.size() * 10 + 100;
		Adaptive_Data_Model mModel(1<<QPforColor);
		Adaptive_Data_Model mSign (2);
		Arithmetic_Codec ace(NMAX, NULL);
		ace.start_encoder();
		for(i = 0 ; i < nColor*3; i++)
		{

			nSymbol = quantColor[i];
		
			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;
			ace.encode(nSymbol, mModel);
			if(nSign != 2)
				ace.encode(nSign, mSign);

		}
		ace.write_to_mem(m_compressedStream, m_compressedStreamSize);

	}
}

void QBCREncoder::Encode_ColorIndex()
{
	if(nColorIndex == 0)
		return;
	int symbol = 0;
	unsigned int attr_num;
	
	if(entropytype == DEFAULT)
	{
		unsigned int lengthColorIndex;
		lengthColorIndex = int(log((double)nColor-1)/log(2.0)) + 1;

		for(attr_num = 0 ; attr_num < nColorIndex*3.; attr_num++)
		{
			Insert_Data(ColorIndex[attr_num],lengthColorIndex);
		}
	}
	else if(entropytype == BP)
	{
		unsigned int nMaxCheckBPL = int(log((float)(nColor-1)) / log((float)2)) +1;
		unsigned int prefix_size = int(log((float)(nMaxCheckBPL+1)) / log((float)2)) +1;	
		int nSymbol, nSign;
		unsigned int i;

		Insert_Data(prefix_size, LEN_PREFIX);
		int previousQuantValue[3];
		for(attr_num = 0; attr_num < 3; attr_num++)
		{
			previousQuantValue[attr_num] = 0;
		}
		for(i = 0 ; i <  nColorIndex*3; i++)
		{

			nSymbol = ColorIndex[i]-previousQuantValue[i%3];
			previousQuantValue[i%3] = ColorIndex[i];
		
			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;


			Insert_Data(BPLTable[nSymbol],prefix_size);

			if(payloadTable[nSymbol] != -1)
				Insert_Data(payloadTable[nSymbol], (BPLTable[nSymbol]-1));
			if(nSign != 2)
			{
				Insert_Data(nSign, 1);
			}
		
		}

	}
	else
	{
		unsigned int attr_num;
		int previousQuantValue[3];
		for(attr_num = 0; attr_num < 3; attr_num++)
		{
			previousQuantValue[attr_num] = 0;
		}

		int nSymbol, nSign, input;
		unsigned int i;
		int lengthTexCoordIndex = int(log((double)nColor-1)/log(2.0)) + 1;
		unsigned int NMAX = (int) nTexCoordIndex*2 * lengthTexCoordIndex + 100;
		Adaptive_Data_Model mModel(1024);
		Adaptive_Data_Model mhasnext (2);
		Adaptive_Data_Model mSign (2);
		Arithmetic_Codec ace(NMAX, NULL);
		ace.start_encoder();

		for(attr_num = 0 ; attr_num < nColorIndex*3; attr_num++)
		{
			nSymbol = ColorIndex[attr_num] - previousQuantValue[attr_num%3];
			previousQuantValue[attr_num%3] =ColorIndex[attr_num];

			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;
			while(1)
			{
				ace.encode(nSymbol%1024, mModel);
				nSymbol/=1024;
				if(nSymbol)
					ace.encode(1, mhasnext);
				else
					break;
			}
			ace.encode(0, mhasnext);
			if(nSign != 2)
				ace.encode(nSign, mSign);
			
		}
		ace.write_to_mem(m_compressedStream, m_compressedStreamSize);

	}

}

int QBCREncoder::QuantizationGeometry(int QP)
{
	unsigned int i,j,k;
	float quantized_Value = (float)((1 << QP) - 1) ;
	int previousQuantValue[3];
	for(i = 0; i < 3; i++)
	{
		previousQuantValue[i] = 0;
	}
	// quantization
	int m_result, input;
	if(codingtype == DEFAULT)
	{
		for(i = 0; i < nCoord * 3; i++)
		{
			m_result = (int)(((Coord[i]-quantMinGeometry[i%3])/quantRangeGeometry[i%3])*quantized_Value + 0.5);
			quantCoord.push_back(m_result);
		}	
	}
	else
	{
		for(i = 0; i < nCoord * 3; i++)
		{
			m_result = (int)(((Coord[i]-quantMinGeometry[i%3])/quantRangeGeometry[i%3])*quantized_Value + 0.5);
			input = m_result - previousQuantValue[i%3];
			previousQuantValue[i%3]= m_result;
			quantCoord.push_back(input);
			if(MaxDifferentialForGeometry < abs(input))
			{
				MaxDifferentialForGeometry = abs(input);
			}
		}

	}

	return 1;

}

void QBCREncoder::Normalization()
{
	float distance;
	int i, k;
	for(i = 0 ; i< nNormal ; i++)
	{
		distance = 0.0;
		for(k = 0; k < 3; k++)
		{
			distance += (Normal[3*i+k]*Normal[3*i+k]);
		}
		distance = sqrt(distance);
		for(k = 0; k < 3; k++)
		{
			normalized.push_back(Normal[3*i+k]/distance);;
		}

		
	}

}

int QBCREncoder::QuantizationNormal(int QP)
{
	unsigned int i,j,k;
	float quantized_Value = (float)((1 << QP) - 1) ;
	int previousQuantValue[3];
	for(i = 0; i < 3; i++)
	{
		previousQuantValue[i] = 0;
	}
	// quantization
	int m_result, input;
	if(codingtype == DEFAULT)
	{
		for(i = 0; i < nNormal * 3; i++)
		{
			m_result = (int)(((normalized[i]-quantMinNormal[i%3])/quantRangeNormal[i%3])*quantized_Value + 0.5);
			quantNormal.push_back(m_result);
		}	
	}
	else
	{
		for(i = 0; i < nNormal * 3; i++)
		{
			m_result = (int)(((normalized[i]-quantMinNormal[i%3])/quantRangeNormal[i%3])*quantized_Value + 0.5);
			input = m_result - previousQuantValue[i%3];
			previousQuantValue[i%3]= m_result;
			quantNormal.push_back(input);
			if(MaxDifferentialForNormal < abs(input))
			{
				MaxDifferentialForNormal = abs(input);
			}
		}

	}


	return 1;

}

int QBCREncoder::QuantizationColor(int QP)
{
	unsigned int i,j,k;
	float quantized_Value = (float)((1 << QP) - 1) ;
	int previousQuantValue[3];
	for(i = 0; i < 3; i++)
	{
		previousQuantValue[i] = 0;
	}
	// quantization
	int m_result, input;
	if(codingtype == DEFAULT)
	{
		for(i = 0; i < nColor * 3; i++)
		{
			m_result = (int)(((Color[i]-quantMinColor[i%3])/quantRangeColor[i%3])*quantized_Value + 0.5);
			quantColor.push_back(m_result);
		}	
	}
	else
	{
		for(i = 0; i < nColor * 3; i++)
		{
			m_result = (int)(((Color[i]-quantMinColor[i%3])/quantRangeColor[i%3])*quantized_Value + 0.5);
			input = m_result - previousQuantValue[i%3];
			previousQuantValue[i%3]= m_result;
			quantColor.push_back(input);
			if(MaxDifferentialForColor < abs(input))
			{
				MaxDifferentialForColor = abs(input);
			}
		}

	}

	return 1;

}


int QBCREncoder::QuantizationTexCoord(int QP)
{
	unsigned int i,j,k;
	float quantized_Value = (float)((1 << QP) - 1) ;
	int previousQuantValue[2];
	for(i = 0; i < 2; i++)
	{
		previousQuantValue[i] = 0;
	}
	// quantization
	int m_result, input;
	if(codingtype == DEFAULT)
	{
		for(i = 0; i < nTexCoord * 2; i++)
		{
			m_result = (int)(((TexCoord[i]-quantMinTexCoord[i%2])/quantRangeTexCoord[i%2])*quantized_Value + 0.5);
			quantTexCoord.push_back(m_result);
		}	
	}
	else
	{
		for(i = 0; i < nTexCoord * 2; i++)
		{
			m_result = (int)(((TexCoord[i]-quantMinTexCoord[i%2])/quantRangeTexCoord[i%2])*quantized_Value + 0.5);
			input = m_result - previousQuantValue[i%2];
			previousQuantValue[i%2]= m_result;
			quantTexCoord.push_back(input);
			if(MaxDifferentialForTexCoord < abs(input))
			{
				MaxDifferentialForTexCoord = abs(input);
			}
		}

	}

	return 1;

}


int QBCREncoder::DumpBitstream(char* filename)
{
	FILE* m_fp;
	m_fp = fopen(filename,"wb");
	unsigned char data;
	for(int i=0;i<dumpBitstream.size();i++)
	{
		data = dumpBitstream[i];
		fwrite(&data,sizeof(unsigned char),1,m_fp);
		
	}	
	
	int filesize =  ftell(m_fp);
	dumpBitstream.clear();
	quantCoord.clear();
	quantNormal.clear();
	normalized.clear();
	quantColor.clear();
	quantTexCoord.clear();
	fclose(m_fp);
	return filesize;
}

void QBCREncoder::Insert_Data(int symbol, int length)
{
	int i;
	int comparison = 1 << length;
	for(i=0; i < length; i++)
	{
		dumpBit <<= 1;
		comparison >>= 1;
		bitPostion++;
		dumpBit += (symbol & comparison) ?  1:0;
		if(bitPostion == LEN_BYTE)
		{
			bitPostion = 0;
			Write2Stream(m_compressedStream, m_compressedStreamSize, &dumpBit, sizeof(unsigned char));
			dumpBit = INIT_UCHAR;
		}
	}

}

void QBCREncoder::Flush()
{
	if(bitPostion == 0)
		return;
	int shift = LEN_BYTE - bitPostion;
	dumpBit <<= shift;
	Write2Stream(m_compressedStream, m_compressedStreamSize, &dumpBit, sizeof(unsigned char));
}

