//-------------------------------------------------------------------------
// 
// Copyright Institut Telecom, ETRI and Hanyang University 2008, 2009
// Licensed Materials, Program Property of Institut Telecom, ETRI and Hanyang University
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  Institut Telecom 
// ETRI and Hanyang University retain full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//    AUTHORS:
//    Khaled Mammou
//    Marius Preda
//    Seung Wook Lee
//    Daiyong Kim
//    KyoungSoo Son
//
//
//    http://www-artemis.it-sudparis.eu/
//    http://www.etri.re.kr/
//    http://dmlab.hanyang.ac.kr/
//    
//  
//    Marius Preda       (marius.preda@int-evry.fr)
//    Khaled Mammou      (khaled_mamou@yahoo.fr)
//    Seung Wook Lee     (tajinet@etri.re.kr)  
//    Daiyong Kim        (dykim@dmlab.hanyang.ac.kr)    
//    KyoungSoo Son      (bonokensin@dmlab.hanyang.ac.kr)
//
//-------------------------------------------------------------------------


#include "stdafx.h"
#include "SC3DMCEncoder.h"
#include "utils.h"


#define INIT_UCHAR 0x00
#define LEN_BYTE	8
#define LEN_BIT		1
#define LEN_QP		5
#define MIN_VLAUE -1000000

//dmlab
void SC3DMCEncoder::EncodeHeader(unsigned char * compressedStream, unsigned int & compressedStreamSize)
{
	int nLoop, j;
	unsigned int startCode = SC3DMCStartCode;
	Write2Stream(compressedStream, compressedStreamSize, &startCode, sizeof(unsigned int));

	// This information will be filled after the encoding process
	unsigned int streamSize = 0;
	Write2Stream(compressedStream, compressedStreamSize, &streamSize, sizeof(unsigned int));

	unsigned char encodingType = (unsigned char) m_params.m_encodingType; // 3bit -> 8bit 
	Write2Stream(compressedStream, compressedStreamSize, &encodingType, sizeof(unsigned char));
	
	float creaseAngle = m_ifs->GetCreaseAngle();
	Write2Stream(compressedStream, compressedStreamSize, &creaseAngle, sizeof(unsigned float));
	
	unsigned int compressionMask = 0;
	bool isTriangulaMesh = false;
	bool otherAttributesPerVertex = true;

	if (m_ifs->GetCCW()) compressionMask += 1;
	if (m_ifs->GetSolid()) compressionMask += 2;
	if (m_ifs->GetConvex()) compressionMask += 4;
	if (m_ifs->GetColorPerVertex()) compressionMask += 8;
	if (m_ifs->GetNormalPerVertex()) compressionMask += 16;
	if (otherAttributesPerVertex) compressionMask += 32;	// otherAttributesPerVertex true
	if (isTriangulaMesh) compressionMask += 64;
	if (m_params.m_vertexOrderPres == 1) compressionMask += 128;
	if (m_params.m_triangleOrderPres == 1) compressionMask += 256;

	Write2Stream(compressedStream, compressedStreamSize, &compressionMask, sizeof(unsigned char)*2); // 10bit -> 16bit


	int numberOfCoord = m_ifs->GetNCoord();
	int numberOfNormal = m_ifs->GetNNormal();
	int numberOfColor = m_ifs->GetNColor();
	int numberOfTexCoord = m_ifs->GetNTexCoord();
	int numberOfOtherAttributes = 0;
	int dimensionOfOtherAttributes = 0;
	
	int numberOfCoordIndex = m_ifs->GetNCoordIndex();
	int numberOfNormalIndex = m_ifs->GetNNormalIndex();
	int numberOfColorIndex = m_ifs->GetNColorIndex();
	int numberOfTexCoordIndex = m_ifs->GetNTexCoordIndex();
	int numberOfOtherAttributesIndex = 0;

	Write2Stream(compressedStream, compressedStreamSize, &numberOfCoord, sizeof(unsigned int));
	Write2Stream(compressedStream, compressedStreamSize, &numberOfNormal, sizeof(unsigned int));
	Write2Stream(compressedStream, compressedStreamSize, &numberOfColor, sizeof(unsigned int));
	Write2Stream(compressedStream, compressedStreamSize, &numberOfTexCoord, sizeof(unsigned int));
	Write2Stream(compressedStream, compressedStreamSize, &numberOfOtherAttributes, sizeof(unsigned int));

	unsigned char* m_QP;

	if (numberOfOtherAttributes >0)
	{
		Write2Stream(compressedStream, compressedStreamSize, &dimensionOfOtherAttributes, sizeof(unsigned char));// 2bit -> 8 bit

	}

	if (numberOfCoord>0)
	{
		Write2Stream(compressedStream, compressedStreamSize, &numberOfCoordIndex, sizeof(unsigned int));
		Write2Stream(compressedStream, compressedStreamSize, &m_params.m_nQCoordBits, sizeof(unsigned char));  // 5bit -> 8bit
	}
	if (numberOfNormal>0)
	{
		Write2Stream(compressedStream, compressedStreamSize, &numberOfNormalIndex, sizeof(unsigned int));
		Write2Stream(compressedStream, compressedStreamSize, & m_params.m_nQNormalBits, sizeof(unsigned char));  // 5bit -> 8bit
	
	}
	if (numberOfColor>0)
	{
		Write2Stream(compressedStream, compressedStreamSize, &numberOfColorIndex, sizeof(unsigned int));
		Write2Stream(compressedStream, compressedStreamSize, & m_params.m_nQColorBits, sizeof(unsigned char));  // 5bit -> 8bit
	
	}
	if (numberOfTexCoord>0)
	{
		Write2Stream(compressedStream, compressedStreamSize, &numberOfTexCoordIndex, sizeof(unsigned int));
		Write2Stream(compressedStream, compressedStreamSize, & m_params.m_nQTexCoordBits, sizeof(unsigned char));  // 5bit -> 8bit
	
	}	

	if (numberOfOtherAttributes >0)
	{
		Write2Stream(compressedStream, compressedStreamSize, &numberOfOtherAttributesIndex, sizeof(unsigned int));
		Write2Stream(compressedStream, compressedStreamSize, & m_params.m_nQOtherAttributesBits, sizeof(unsigned char));  // 5bit -> 8bit
	
	}


	if (numberOfCoord>0)
	{
		m_ifs->SetMinMaxCoord();
	
		
		for(j=0;j<3;j++)
		{
			Write2Stream(compressedStream, compressedStreamSize, & m_ifs->minCoord[j], sizeof(unsigned float)); 
		}
		
		for(j=0;j<3;j++)
		{
			Write2Stream(compressedStream, compressedStreamSize, &(m_ifs->rangeCoord[j]), sizeof(unsigned float));  
		}
	}
	if(numberOfNormal > 0)
	{
		m_ifs->SetMinMaxNormal();


		for(j=0;j<3;j++)
		{
			Write2Stream(compressedStream, compressedStreamSize, & m_ifs->minNormal[j], sizeof(unsigned float));  
		}
		for(j=0;j<3;j++)
		{
			Write2Stream(compressedStream, compressedStreamSize, &(m_ifs->rangeNormal[j]), sizeof(unsigned float));  
		}
	}
	
	
	if(numberOfColor > 0)
	{

		m_ifs->SetMinMaxColor();
		
		for(j=0;j<3;j++)
		{
			Write2Stream(compressedStream, compressedStreamSize, & m_ifs->minColor[j], sizeof(unsigned float));  
		}
		for(j=0;j<3;j++)
		{
			Write2Stream(compressedStream, compressedStreamSize, &m_ifs->rangeColor[j], sizeof(unsigned float));  
		}

	}

	if(numberOfTexCoord > 0)
	{
		m_ifs->SetMinMaxTexCoord();
		
		for(j=0;j<2;j++)
		{
			Write2Stream(compressedStream, compressedStreamSize, & m_ifs->minTexCoord[j], sizeof(unsigned float));  
		}
		for(j=0;j<2;j++)
		{
			Write2Stream(compressedStream, compressedStreamSize, &m_ifs->rangeTexCoord[j], sizeof(unsigned float));  
		}
	}

	if(dimensionOfOtherAttributes > 0 )
	{
		float* minOtherAtrribute;
		minOtherAtrribute = new float [dimensionOfOtherAttributes ];
		int rangeOtherAtrribute = 0;
		for(j=0;j<dimensionOfOtherAttributes;j++)
		{
			minOtherAtrribute[j] = 0.0;

		}

		if(numberOfOtherAttributes > 0)
		{
			for(j=0;j<dimensionOfOtherAttributes;j++)
			{
				Write2Stream(compressedStream, compressedStreamSize, & minOtherAtrribute[j], sizeof(unsigned float));  
			}
			Write2Stream(compressedStream, compressedStreamSize, & rangeOtherAtrribute, sizeof(unsigned float));

		}
		delete [] minOtherAtrribute;
	}

}
//~dmlab



void SC3DMCEncoder::EncodeDataBuffer(unsigned char * compressedStream, unsigned int & compressedStreamSize)
{

	if (m_params.m_encodingType == 2) // TFAN
	{
		Flush(compressedStream, compressedStreamSize);
		TFANEncoder tfenc;
		tfenc.SetIFS(m_ifs);
		tfenc.SetParams(m_params);
		tfenc.EncodeDataBuffer(compressedStream+compressedStreamSize, compressedStreamSize);
	}

	else if (m_params.m_encodingType == 1) 
	{
		//dmlab
		SVAEncoder svaenc;
		svaenc.SetIFS(m_ifs);
		svaenc.SetParams(m_params);
		svaenc.EncodeDataBuffer(compressedStream+compressedStreamSize, compressedStreamSize);
		//~dmlab
	
	}
	else if (m_params.m_encodingType == 0) 
	{
		//dmlab
		QBCREncoder qbcrenc;
		qbcrenc.SetIFS(m_ifs);
		qbcrenc.SetParams(m_params);
		qbcrenc.EncodeDataBuffer(compressedStream+compressedStreamSize, compressedStreamSize);
		//~dmlab
	}
}

//dmlab
void SC3DMCEncoder::Insert_Data(unsigned char * compressedStream, unsigned int & compressedStreamSize, int symbol, int length)
{
	int i;
	int comparison = 1 << length;
	for(i=0; i < length; i++)
	{
		dumpBit <<= 1;
		comparison >>= 1;
		bitPostion++;
		dumpBit += (symbol & comparison) ?  1:0;
		if(bitPostion == LEN_BYTE)
		{
			bitPostion = 0;
			compressedStream[compressedStreamSize++]=dumpBit;
			dumpBit = INIT_UCHAR;
		}
	}
}
//~dmlab

//dmlab
void SC3DMCEncoder::Flush(unsigned char * compressedStream, unsigned int & compressedStreamSize)
{
	if(bitPostion == 0)
		return;
	int shift = 8 - bitPostion;
	dumpBit <<= shift;
	compressedStream[compressedStreamSize++] = dumpBit;
	
}
//~dmlab
