//-------------------------------------------------------------------------
// 
// Copyright Institut Telecom, ETRI and Hanyang University 2008, 2009
// Licensed Materials, Program Property of Institut Telecom, ETRI and Hanyang University
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  Institut Telecom 
// ETRI and Hanyang University retain full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//    AUTHORS:
//    Khaled Mammou
//    Marius Preda
//    Seung Wook Lee
//    Daiyong Kim
//    KyoungSoo Son
//
//
//    http://www-artemis.it-sudparis.eu/
//    http://www.etri.re.kr/
//    http://dmlab.hanyang.ac.kr/
//    
//  
//    Marius Preda       (marius.preda@int-evry.fr)
//    Khaled Mammou      (khaled_mamou@yahoo.fr)
//    Seung Wook Lee     (tajinet@etri.re.kr)  
//    Daiyong Kim        (dykim@dmlab.hanyang.ac.kr)    
//    KyoungSoo Son      (bonokensin@dmlab.hanyang.ac.kr)
//
//-------------------------------------------------------------------------


#include "stdafx.h"
#include "SC3DMCDecoder.h"


//dmlab
bool SC3DMCDecoder::DecodeHeader(unsigned char * & compressedStream)
{
	unsigned char * compressedStreamO = compressedStream;
	unsigned char tHeader[4];
	int loop, index;
	unsigned int startCode = 0;
	
	ReadFromStream(compressedStream, &startCode, sizeof(unsigned int));

	if (startCode != SC3DMCStartCode) return false;

	m_streamSize = 0;
	ReadFromStream(compressedStream, &m_streamSize, sizeof(unsigned int));

	ReadFromStream(compressedStream, &m_encodingType, sizeof(unsigned char));

	float creaseAngle = 0.0;
	ReadFromStream(compressedStream, &creaseAngle, sizeof(unsigned float));
	m_ifs->SetCreaseAngle(creaseAngle);

	unsigned int compressionMask = 0;
	ReadFromStream(compressedStream, &compressionMask, sizeof(unsigned char)*2);


	if (compressionMask & 1) m_ifs->SetCCW(true);
	else m_ifs->SetCCW(false);
	if (compressionMask & 2) m_ifs->SetSolid(true);
	else m_ifs->SetSolid(false);
	if (compressionMask & 4) m_ifs->SetConvex(true);
	else m_ifs->SetConvex(false);
	if (compressionMask & 8) m_ifs->SetColorPerVertex(true);
	else  m_ifs->SetColorPerVertex(false);
	if (compressionMask & 16) m_ifs->SetNormalPerVertex(true);
	else m_ifs->SetNormalPerVertex(false);
	if (compressionMask & 32) m_Header.otherAttributesPerVertex = true;
	m_Header.otherAttributesPerVertex = false;
	if (compressionMask & 64) m_Header.isTriangulaMesh = true;
	m_Header.isTriangulaMesh = false;
	if (compressionMask & 128) m_vertexOrderPres = 1;
	if (compressionMask & 256) m_triangleOrderPres = 1;

	int numberOfCoord = 0;
	int numberOfNormal = 0;
	int numberOfColor = 0;
	int numberOfTexCoord =0;
	int numberOfOtherAttributes = 0;
	int dimensionOfOtherAttributes = 0;
	
	int numberOfCoordIndex = 0;
	int numberOfNormalIndex = 0;
	int numberOfColorIndex = 0;
	int numberOfTexCoordIndex = 0;
	int numberOfOtherAttributesIndex = 0;

	ReadFromStream(compressedStream, &numberOfCoord, sizeof(unsigned int));
	ReadFromStream(compressedStream, &numberOfNormal, sizeof(unsigned int));
	ReadFromStream(compressedStream, &numberOfColor, sizeof(unsigned int));
	ReadFromStream(compressedStream, &numberOfTexCoord, sizeof(unsigned int));
	ReadFromStream(compressedStream, &dimensionOfOtherAttributes, sizeof(unsigned int));



	if (numberOfOtherAttributes >0)
	{
		ReadFromStream(compressedStream, &dimensionOfOtherAttributes, sizeof(unsigned char));
		m_Header.dimensionOfOtherAttribute = dimensionOfOtherAttributes;
	}

	int m_QP=0;
	if (numberOfCoord>0)
	{
		ReadFromStream(compressedStream, &numberOfCoordIndex, sizeof(unsigned int));		
		
		ReadFromStream(compressedStream, &m_QP, sizeof(unsigned char));
		m_Header.QPforGeometry = m_QP;
	}

	if (numberOfNormal>0)
	{
		ReadFromStream(compressedStream, &numberOfNormalIndex, sizeof(unsigned int));		
		ReadFromStream(compressedStream, &m_QP, sizeof(unsigned char));
		m_Header.QPforNormal = m_QP;
	}

	
	if (numberOfColor>0)
	{
		ReadFromStream(compressedStream, &numberOfColorIndex, sizeof(unsigned int));		
		ReadFromStream(compressedStream, &m_QP, sizeof(unsigned char));;
		m_Header.QPforColor = ReadBitsFromStorage(5,compressedStream);
	}

	if (numberOfTexCoord>0)
	{
		ReadFromStream(compressedStream, &numberOfTexCoordIndex, sizeof(unsigned int));		
		ReadFromStream(compressedStream, &m_QP, sizeof(unsigned char));;
		m_Header.QPforTexCoord = m_QP;
	}

	if (numberOfOtherAttributes >0)
	{
		ReadFromStream(compressedStream, &numberOfOtherAttributes, sizeof(unsigned int));		
		ReadFromStream(compressedStream, &m_QP, sizeof(unsigned char));;
		m_Header.QPforOtherAttribute = m_QP;
	}

	m_ifs->SetNCoord(numberOfCoord);
	m_ifs->SetNNormal(numberOfNormal);
	m_ifs->SetNColor(numberOfColor);
	m_ifs->SetNTexCoord(numberOfTexCoord);	
	m_ifs->SetNCoordIndex(numberOfCoordIndex);
	m_ifs->SetNNormalIndex(numberOfNormalIndex);
	m_ifs->SetNColorIndex(numberOfColorIndex);
	m_ifs->SetNTexCoordIndex(numberOfTexCoordIndex);
	m_ifs->AllocateMem();


	m_Header.quantmodeGeometry = ReadBitsFromStorage(1,compressedStream);
	float fValue=0.0;
	int quantmode=0;
	if (numberOfCoord>0)
	{
	
		for(index = 0 ; index < 3; index++)
		{
			ReadFromStream(compressedStream, &fValue, sizeof(unsigned float));		
			m_Header.quantMinGeometry[index] = fValue;
		}
		for(index = 0 ; index < 3; index++)
		{
			ReadFromStream(compressedStream, &fValue, sizeof(unsigned float));
			m_Header.rangeGeometry[index ] = fValue;
		}
	}
	
	
	if(numberOfNormal > 0)
	{
		for(index = 0 ; index < 3; index++)
		{
			ReadFromStream(compressedStream, &fValue, sizeof(unsigned float));	
			m_Header.quantMinNormal[index] = fValue;
		}
		for(index = 0 ; index < 3; index++)
		{
			ReadFromStream(compressedStream, &fValue, sizeof(unsigned float));
			m_Header.rangeNormal[index ] = fValue;
		}
	}
	if (numberOfColor>0)
	{
		for(index = 0 ; index < 3; index++)
		{
			ReadFromStream(compressedStream, &fValue, sizeof(unsigned float));	
			m_Header.quantMinColor[index] = fValue;
		}
		for(index = 0 ; index < 3; index++)
		{
			ReadFromStream(compressedStream, &fValue, sizeof(unsigned float));
			m_Header.rangeColor[index ] = fValue;
		}
		
	}
	
	if (numberOfTexCoord>0)
	{
		for(index = 0 ; index < 2; index++)
		{
			ReadFromStream(compressedStream, &fValue, sizeof(unsigned float));	
			m_Header.quantMinTexCoord[index] = fValue;
		}
		for(index = 0 ; index < 2; index++)
		{
			ReadFromStream(compressedStream, &fValue, sizeof(unsigned float));
			m_Header.rangeTexCoord[index ] = fValue;
		}
		
		
	}
	if(m_Header.dimensionOfOtherAttribute > 0)
	{
		m_Header.quantMinOtherAttribute = new float [m_Header.dimensionOfOtherAttribute];
		if (numberOfOtherAttributes >0)
		{
			ReadFromStream(compressedStream, &quantmode, sizeof(unsigned char));	
			m_Header.quantmodeOtherAttribute = quantmode;
			for(index = 0 ; index < m_Header.dimensionOfOtherAttribute; index++)
			{
				ReadFromStream(compressedStream, &fValue, sizeof(unsigned float));	
				m_Header.quantMinOtherAttribute[index] = fValue;
			}
			if(m_Header.quantmodeTexCoord == 0)
			{
				ReadFromStream(compressedStream, &fValue, sizeof(unsigned float));
				m_Header.quantRangeOtherAttribute = fValue;
			}
			else
			{
				for(index = 0 ; index < m_Header.dimensionOfOtherAttribute; index++)
				{
					ReadFromStream(compressedStream, &fValue, sizeof(unsigned float));
					m_Header.rangeOtherAttribute[index ] = fValue;
				}
			}
			
		}
	}

	return true;
}

//~dmlab

void SC3DMCDecoder::DecodeDataBuffer(unsigned char * &compressedStream)
{
	
	if (m_encodingType == 2) // TFAN
	{	
		compressedStream++;
		TFANDecoder tfandec(m_vertexOrderPres, m_triangleOrderPres);
		tfandec.SetIFS(m_ifs);
		tfandec.SetHeader(&m_Header);
		tfandec.DecodeTFANData(compressedStream);
	}
	else if (m_encodingType == 1) // SVA 
	{
		SVADecoder svadec;
		//dmlab
		svadec.SetIFS(m_ifs);
		svadec.SetSC3DMCHeader(&m_Header);
		//~dmlab
		svadec.DecoderSVAData(compressedStream); 

	}
	else if (m_encodingType == 0) // QBCR
	{
		
		QBCRDecoder qbcrdec;
		//dmlab
		qbcrdec.SetIFS(m_ifs);
		qbcrdec.SetSC3DMCHeader(&m_Header);
		//~dmlab
		qbcrdec.DecoderQBCRData(compressedStream); 
	}
}

//dmlab
unsigned int SC3DMCDecoder::ReadBitsFromStorage(unsigned int unReadLength,unsigned char * &compressedStream)
{
	unsigned int n = unReadLength;
	int c=(n + 14 - unCurrentBufferPosition) >> 3;

	union { 
		long long l;
		char c[8];
	} buf;
	buf.l = 0;

	int i=0;
	while(n) {
		i++;
		buf.c[c - i] = *(compressedStream + unCurrentBufferPosition + i - 1);
		if(unCurrentBitPosition >= n) {
			unCurrentBitPosition -= n;
			break;
		} else {
			n -= unCurrentBitPosition + 1;
			unCurrentBitPosition = 7;
			unCurrentBufferPosition++;
		}
	}

	return (buf.l >> ((unCurrentBitPosition + 1) % 8)) & ~(~static_cast<unsigned int>(0) << unReadLength);
}

void SC3DMCDecoder::CheckFlush(unsigned char * &compressedStream)
{
	if(unCurrentBitPosition != 0)
		return;
	compressedStream++;
}

//~dmlab