//-------------------------------------------------------------------------
// 
// Copyright ETRI, Hanyang University 2008, 2009
// Licensed Materials, Program Property of ETRI & Hanyang University
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.
// ETRI and Hanyang University retain full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//    AUTHORS:
//    Seung Wook Lee (ETRI)
//    Daiyong Kim (Hanyang University)
//    KyoungSoo Son (Hanyang University)
//
//    http://www.etri.re.kr/
//    http://dmlab.hanyang.ac.kr/
//  
//    Seung Wook Lee     (tajinet@etri.re.kr)  
//    Daiyong Kim        (dykim@dmlab.hanyang.ac.kr)    
//    KyoungSoo Son      (bonokensin@dmlab.hanyang.ac.kr)
//
//-------------------------------------------------------------------------



#include <iostream>
#include <time.h>
#include <string>
#include <iostream>
#include <fstream>
#include <sstream> 
#include <math.h>
#include <vector>
#include <map>
using namespace std;

#include <assert.h>
#include "IndexedFaceSet.h"
#include "SVA_Decoder.h"
#include "utils.h"

#define LEN_BPLMAX	32
#define LEN_BYTE	8
#define LEN_QP		5
#define LEN_PREFIX	5
#define COORD_DIVISION	3
#define LEN_BIT		1
#define DEFAULT		0
#define DIFF		1
#define BP			1
#define AC			2

#define SC3DMCStartCode 0x00001F1


SVADecoder::SVADecoder()
{
	readBitstream = NULL;
	unCurrentBitPosition = 7;
	unCurrentBufferPosition = 0;
	previousQuant[0] = 0;
	previousQuant[1] = 0;
	previousQuant[2] = 0;
	codingtype = 0;
	entropytype = 0;
	FDmode= -1;
	FDvalue = -1;
	BPLTable = NULL;
	
}

SVADecoder::~SVADecoder()
{
	if(readBitstream != NULL)
	{
		delete [] readBitstream;
		readBitstream = NULL;
	}
	if(BPLTable != NULL)
	{
		delete [] BPLTable;
		BPLTable = NULL;
	}
}

int SVADecoder::ReadData(char * filename)
{
	FILE* m_inputFile;			
	m_inputFile = fopen(filename,"rb");


	fseek(m_inputFile,0L, SEEK_END);
	unsigned int filesize = ftell(m_inputFile);
	readBitstream = new unsigned char [filesize];
	rewind(m_inputFile);
	fread(readBitstream, sizeof(unsigned char),filesize,m_inputFile );
	m_pReadStream=readBitstream;
	fclose(m_inputFile);

	return 1;
}


int SVADecoder::DecoderSVA(char* name)
{
	ReadData(name);
	BPLInit();
	DecoderCommonHeader();
	DecoderHeader();
	DecoderGeometry();
	DecoderConnectivity();
	DecoderNormal();
	DecoderNormalIndex();
	DecoderTexCoord();
	DecoderTexCoordIndex();
	DecoderColor();
	DecoderColorIndex();

	return 1;
}



void SVADecoder::DecoderCommonHeader()
{
	

}

void SVADecoder::DumpCompressedVB(char* nameVB)
{
	outIFS_->SaveIFSVRML2(nameVB);

	unCurrentBitPosition = 7;
	unCurrentBufferPosition = 0;
	previousQuant[0] = 0;
	previousQuant[1] = 0;
	previousQuant[2] = 0;
	if(readBitstream != NULL)
	{
		delete [] readBitstream;
		readBitstream = NULL;
	}

}
void SVADecoder::DecoderHeader()
{
	ReadFromStream(m_pReadStream, &codingtype, sizeof(unsigned char));
	ReadFromStream(m_pReadStream, &entropytype, sizeof(unsigned char));
}

void SVADecoder::DecoderGeometry()
{
	if(entropytype == BP)
	{
		
		unsigned int nLoop, prefixSize, nBPL, nPayload, nSign, nQuantValue;
		int difValue;
		prefixSize = ReadBitsFromStorage(LEN_PREFIX);

		for(nLoop = 0; nLoop < 3 * numberOfCoord; nLoop++)
		{
			nBPL=ReadBitsFromStorage(prefixSize);
			if(nBPL > 2)
			{
				nPayload = ReadBitsFromStorage(nBPL-1);
			}
			else
			{
				nPayload = 0;
			}

			difValue = BPLTable[nBPL] + nPayload;
			if(difValue != 0)
			{
				nSign = ReadBitsFromStorage(1);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}
			nQuantValue = InverseCircularDifferenceForGeometry(previousQuant[nLoop % 3],difValue);
			
			previousQuant[nLoop % 3] = nQuantValue;
			
			//dmlab
			//coord.push_back(InverseQuantization(nQuantValue,(nLoop%3)));
			outIFS_->coord_[nLoop] =  InverseQuantization(nQuantValue, quantRangeForGeometry[nLoop%3] ,quantMinForGeometry[nLoop%3], quantValueforGeometry);
			

			 //~dmlab
		}
	}
	else
	{
		unsigned int nLoop, nSign;
		int previousQuant[3],difValue, i, nQuantValue;
		nQuantValue=0;
//		m_pReadStream = m_pReadStream+unCurrentBufferPosition;
		unsigned int code_bytes = 0;
		ReadFromStream(m_pReadStream, &code_bytes, sizeof(unsigned int));

		unsigned char * code_buffer = new unsigned char [code_bytes];                                                   
		ReadFromStream(m_pReadStream, code_buffer, sizeof(unsigned char)*code_bytes);

		
		Adaptive_Data_Model mModel(1<<QPforGeometry);
		Adaptive_Data_Model mSign (2);
		Arithmetic_Codec acd(code_bytes, code_buffer);
		acd.start_decoder();	// initialize decoder

		for(i = 0; i < 3; i++)
		{
			previousQuant[i] = 0;
		}
		

		for(nLoop = 0; nLoop < 3 * numberOfCoord; nLoop++)
		{
			difValue = acd.decode(mModel);
			if(difValue != 0)
			{
				nSign = acd.decode(mSign);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}

			nQuantValue = InverseCircularDifferenceForGeometry(previousQuant[nLoop % 3],difValue);
			
			previousQuant[nLoop % 3] = nQuantValue;
			
			//dmlab
			//coord.push_back(InverseQuantization(nQuantValue,(nLoop%3)));
			outIFS_->coord_[nLoop] = InverseQuantization(nQuantValue, quantRangeForGeometry[nLoop%3] ,quantMinForGeometry[nLoop%3], quantValueforGeometry);
			 //~dmlab
		}

	}

}

float SVADecoder::InverseQuantization(unsigned quantCoord, float quantRange, float quantMin, float quantValue)
{
	float data;
	data = ((float)quantCoord/quantValue)*quantRange+quantMin;
	return data;
	
}


unsigned int SVADecoder::InverseCircularDifferenceForConnectivity(unsigned int previous, int differential)
{
	int result = (int)previous+differential;
	if(result >= (int)numberOfCoord)
	{
		result = result - numberOfCoord;
	}
	else if(result < 0 )
	{
		result = numberOfCoord + result;
	}


	return result;

}

unsigned int SVADecoder::InverseCircularDifferenceForNormalIndex(unsigned int previous, int differential)
{
	int result = (int)previous+differential;
	if(result >= (int)numberOfNormal)
	{
		result = result - numberOfCoord;
	}
	else if(result < 0 )
	{
		result = numberOfCoord + result;
	}


	return result;

}

unsigned int SVADecoder::InverseCircularDifferenceForColorIndex(unsigned int previous, int differential)
{
	int result = (int)previous+differential;
	if(result >= (int)numberOfColor)
	{
		result = result - numberOfCoord;
	}
	else if(result < 0 )
	{
		result = numberOfCoord + result;
	}


	return result;

}

unsigned int SVADecoder::InverseCircularDifferenceForTexCoordIndex(unsigned int previous, int differential)
{
	int result = (int)previous+differential;
	if(result >= (int)numberOfTexCoord)
	{
		result = result - numberOfCoord;
	}
	else if(result < 0 )
	{
		result = numberOfCoord + result;
	}


	return result;

}

unsigned int SVADecoder::InverseCircularDifferenceForGeometry(unsigned int previous, int differential)
{
	int result = previous+differential;
	if(result >=  ((int)quantValueforGeometry+1))
	{
		result = result -  ((int)quantValueforGeometry+1);
	}
	else if(result < 0 )
	{
		result = ((int)quantValueforGeometry+1)+result;
	}


	return result;

}

unsigned int SVADecoder::InverseCircularDifferenceForColor(unsigned int previous, int differential)
{
	int result = previous+differential;
	if(result >=  ((int)quantValueforColor+1))
	{
		result = result -  ((int)quantValueforColor+1);
	}
	else if(result < 0 )
	{
		result = ((int)quantValueforColor+1)+result;
	}


	return result;

}

unsigned int SVADecoder::InverseCircularDifferenceForTexCoord(unsigned int previous, int differential)
{
	int result = previous+differential;
	if(result >=  ((int)quantValueforTexCoord+1))
	{
		result = result -  ((int)quantValueforTexCoord+1);
	}
	else if(result < 0 )
	{
		result = ((int)quantValueforTexCoord+1)+result;
	}


	return result;

}
void SVADecoder::SetHeader()
{
		
	numberOfCoord = outIFS_->nCoord_;
	numberOfNormal = outIFS_->nNormal_;
	numberOfColor = outIFS_->nColor_;
	numberOfTexCoord = outIFS_->nTexCoord_;
	numberOfOtherAttributes = m_Header->numberOfOtherAttribute;
	dimensionOfOtherAttributes = m_Header->dimensionOfOtherAttribute;
	numberOfCoordIndex = outIFS_->nCoordIndex_;
	numberOfNormalIndex = outIFS_->nNormalIndex_;
	numberOfColorIndex = outIFS_->nColorIndex_;
	numberOfTexCoordIndex = outIFS_->nTexCoordIndex_;
	numberOfOtherAttributesIndex = m_Header->numberOfOtherAttributeIndex;

	for(int i=0; i< 3; i++)
	{
		quantMinForGeometry[i] = m_Header->quantMinGeometry[i];
		quantMinForNormal[i] = m_Header->quantMinNormal[i];
		quantMinForColor[i] = m_Header->quantMinColor[i];
		quantRangeForGeometry[i] = m_Header->rangeGeometry[i];
		quantRangeForNormal[i] = m_Header->rangeNormal[i];
		quantRangeForColor[i] = m_Header->rangeColor[i];

	}
	for(int i= 0; i < 2; i++)
	{
		quantMinForTexCoord[i] = m_Header->quantMinTexCoord[i] ;
		quantRangeForTexCoord[i] = m_Header->rangeTexCoord[i];

	}
	
	
	QPforGeometry = m_Header->QPforGeometry;
	quantValueforGeometry = pow(2.0,(double)QPforGeometry) - 1;
	QPforNormal =  m_Header->QPforNormal;
	quantValueforNormal = pow(2.0,(double)QPforNormal) - 1;
	QPforTexCoord =  m_Header->QPforTexCoord;
	quantValueforTexCoord = pow(2.0,(double)QPforTexCoord) - 1;
	QPforColor =  m_Header->QPforColor;
	quantValueforColor = pow(2.0,(double)QPforColor) - 1;

}

int SVADecoder::DecoderSVAData(unsigned char * &compressedStream)
{
	m_pReadStream = compressedStream;
	SetHeader() ;
	DecoderHeader();
	if(entropytype == BP)
	{
		BPLInit();
	}
	
	DecoderGeometry();
	
	DecoderConnectivity();

	DecoderNormal();
	DecoderNormalIndex();
	DecoderTexCoord();
	DecoderTexCoordIndex();
	DecoderColor();
	DecoderColorIndex();

	return 1;

}

void SVADecoder::_code2Normal(const unsigned long code, float *normal, unsigned int subdivision)
{
  unsigned long mask =  (1 << (2 * subdivision)) - 1;
  unsigned long tricode = code & mask;
  unsigned long tcode = tricode;

  // Find y coordinate by solving 2nd degree equation
  // long factor = 1 << (numLevels + 1);
  // long y = (factor - sqrt(factor * factor - 4 * tricode)) / 2.0;
  unsigned long factor = 1 << subdivision;
  unsigned long y = (unsigned long)(factor-sqrt((double)factor*factor-tricode));
  tricode += y * ( y - 2 * factor);
  unsigned long x = tricode / 2;
  short upsideDown = tricode % 2;

  // Calculate coordinates for all vertices in triangle
  float v1x = float(x + upsideDown);
  float v1y = float(y + upsideDown);
  float v2x = float(x + 1);
  float v2y = float(y);
  float v3x = float(x);
  float v3y = float(y + 1);

  // Calculate coordinates of barycenter
  float invMaxCoord = 1.0 / factor;
  normal[0] = (v1x + v2x + v3x) * invMaxCoord;
  normal[1] = (v1y + v2y + v3y) * invMaxCoord;
  normal[2] = 3.0 - normal[0] - normal[1];

  // Flip component signs if necessary
  // unsigned long octantCode = (code & ~mask) >> 2*_nSubdivisionLevels;
  unsigned long octantCode = (code >> 2*subdivision)&0x7;

  if (octantCode & 0x4) normal[0] = -normal[0];
  if (octantCode & 0x2) normal[1] = -normal[1];
  if (octantCode & 0x1) normal[2] = -normal[2];

  //Normalize vector
  float invNorm = 1.0 / sqrt(normal[0]*normal[0] +
			     normal[1]*normal[1] +
			     normal[2]*normal[2]);
  normal[0] *= invNorm; normal[1] *= invNorm; normal[2] *= invNorm;
}

void SVADecoder::DecoderNormal()
{
	if(numberOfNormal == 0)
		return;
	if(entropytype == BP)
	{
		int i;
		int BPLTableForNormal[32];
		for(i = 0 ; i<= 4; i++)
		{
			BPLTableForNormal[i] = i;

		}
		int count=0;
		for( ; i< 32; i++)
		{
			BPLTableForNormal[i] = BPLTableForNormal[i-1] + (1 << (i-3));

		}
		int codeForNormalO = 0;
		int codeForNormalS = 0;
		int code = 0;
		int previousO = 0;
		int previousS = 0;
		int prefix = ReadBitsFromStorage(5);
		int nBPL, nSign;
		float dNormal[3] = {0};
		int subdivision = (QPforNormal - 3) / 2;

		
		if(codingtype == DIFF)
		{
			for(i = 0; i < numberOfNormal; i++)
			{
				for(count = 0; count < 6; count++)
				{
					if(ReadBitsFromStorage(1) == 1)
					{
						codeForNormalO = count;
						break;
					}
				}
				if(count == 6)
				{
					if(ReadBitsFromStorage(1) == 1)
					{
						codeForNormalO = 6;
					}
					else
					{
						codeForNormalO = 7;
					}
				}
				if(codeForNormalO!=0)
				{
					nSign = ReadBitsFromStorage(1);
					if(nSign == 1)
						codeForNormalO = -codeForNormalO;
				}
				nBPL=ReadBitsFromStorage(prefix);
				codeForNormalS=BPLTableForNormal[nBPL];
				if(codeForNormalS >= 4)
				{
					codeForNormalS += ReadBitsFromStorage(nBPL-2);
				}
				if(codeForNormalS != 0)
				{
					nSign = ReadBitsFromStorage(1);
					if(nSign == 1)
						codeForNormalS = -codeForNormalS;
				}
				codeForNormalO = codeForNormalO+previousO;
				codeForNormalS = codeForNormalS+previousS;
				code = (codeForNormalO<<(2*subdivision)) | codeForNormalS;
				previousO = codeForNormalO;
				previousS = codeForNormalS;
					
				_code2Normal(code,dNormal,subdivision);
			
				

				outIFS_->normal_[i] = dNormal[0];
				outIFS_->normal_[i+1] = dNormal[1];
				outIFS_->normal_[i+2] = dNormal[2];


			}


		}
		else
		{
			for(i = 0; i < numberOfNormal; i++)
			{
				for(count = 0; count < 6; count++)
				{
					if(ReadBitsFromStorage(1) == 1)
					{
						codeForNormalO = count;
						break;
					}
				}
				if(count == 6)
				{
					if(ReadBitsFromStorage(1) == 1)
					{
						codeForNormalO = 6;
					}
					else
					{
						codeForNormalO = 7;
					}
				}
				
				nBPL=ReadBitsFromStorage(prefix);
				codeForNormalS=BPLTableForNormal[nBPL];
				if(codeForNormalS >= 4)
				{
					codeForNormalS += ReadBitsFromStorage(nBPL-2);
				}
				codeForNormalO = codeForNormalO^previousO;
				codeForNormalS = codeForNormalS^previousS;
				code = (codeForNormalO<<(2*subdivision)) | codeForNormalS;
				previousO = codeForNormalO;
				previousS = codeForNormalS;
					
				_code2Normal(code,dNormal,subdivision);

				outIFS_->normal_[i] = dNormal[0];
				outIFS_->normal_[i+1] = dNormal[1];
				outIFS_->normal_[i+2] = dNormal[2];


			}


		}


	}
	else
	{
		int i;
		
		int codeForNormalO = 0;
		int codeForNormalS = 0;
		int code = 0;
		int previousO = 0;
		int previousS = 0;
		int nSign;
		float dNormal[3] = {0};
		int subdivision = (QPforNormal - 3) / 2;

		unsigned int nLoop;
		


		if(codingtype == DIFF)
		{
			unsigned int code_bytes = 0;
			ReadFromStream(m_pReadStream, &code_bytes, sizeof(unsigned int));

			unsigned char * code_buffer = new unsigned char [code_bytes];                                                   
			ReadFromStream(m_pReadStream, code_buffer, sizeof(unsigned char)*code_bytes);

			
			Adaptive_Data_Model mModelO(1<<3);
			Adaptive_Data_Model mModelS(1<<(2*subdivision));
			Adaptive_Data_Model mSign (2);
			Arithmetic_Codec acd(code_bytes, code_buffer);
			acd.start_decoder();	// initialize decoder


			for(i = 0; i < numberOfNormal; i++)
			{
				codeForNormalO = acd.decode(mModelO);
				if(codeForNormalO!=0)
				{
					nSign = acd.decode(mSign);
					if(nSign == 1)
						codeForNormalO = -codeForNormalO;
				}
				codeForNormalS = acd.decode(mModelS);
				if(codeForNormalS != 0)
				{
					nSign = acd.decode(mSign);
					if(nSign == 1)
						codeForNormalS = -codeForNormalS;
				}
				codeForNormalO = codeForNormalO+previousO;
				codeForNormalS = codeForNormalS+previousS;
				code = (codeForNormalO<<(2*subdivision)) | codeForNormalS;
				previousO = codeForNormalO;
				previousS = codeForNormalS;
					
				_code2Normal(code,dNormal,subdivision);

				outIFS_->normal_[i] = dNormal[0];
				outIFS_->normal_[i+1] = dNormal[1];
				outIFS_->normal_[i+2] = dNormal[2];


			}

		}
		else
		{
			unsigned int code_bytes = 0;
			ReadFromStream(m_pReadStream, &code_bytes, sizeof(unsigned int));

			unsigned char * code_buffer = new unsigned char [code_bytes];                                                   
			ReadFromStream(m_pReadStream, code_buffer, sizeof(unsigned char)*code_bytes);

			
			Adaptive_Data_Model mModelO(1<<3);
			Adaptive_Data_Model mModelS(1<<(2*subdivision));
			Adaptive_Data_Model mSign (2);
			Arithmetic_Codec acd(code_bytes, code_buffer);
			acd.start_decoder();	// initialize decoder


			for(i = 0; i < numberOfNormal; i++)
			{
				codeForNormalO = acd.decode(mModelO);
				codeForNormalS = acd.decode(mModelS);
					codeForNormalO = codeForNormalO^previousO;
				codeForNormalS = codeForNormalS^previousS;
				code = (codeForNormalO<<(2*subdivision)) | codeForNormalS;
				previousO = codeForNormalO;
				previousS = codeForNormalS;
					
				_code2Normal(code,dNormal,subdivision);

				outIFS_->normal_[i] = dNormal[0];
				outIFS_->normal_[i+1] = dNormal[1];
				outIFS_->normal_[i+2] = dNormal[2];


			}


		}
	}


}

void SVADecoder::DecoderColor()
{
	if(numberOfColor == 0)
		return;
	int i;
	for(i = 0; i < 3; i++)
	{
		previousQuant[i] = 0;
	}
	if(entropytype == BP)
	{
		
		unsigned int nLoop, prefixSize, nBPL, nPayload, nSign, nQuantValue;
		int difValue;
		prefixSize = ReadBitsFromStorage(LEN_PREFIX);

		for(nLoop = 0; nLoop < 3 * numberOfCoord; nLoop++)
		{
			nBPL=ReadBitsFromStorage(prefixSize);
			if(nBPL > 2)
			{
				nPayload = ReadBitsFromStorage(nBPL-1);
			}
			else
			{
				nPayload = 0;
			}

			difValue = BPLTable[nBPL] + nPayload;
			if(difValue != 0)
			{
				nSign = ReadBitsFromStorage(1);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}
			nQuantValue = InverseCircularDifferenceForColor(previousQuant[nLoop % 3],difValue);
			
			previousQuant[nLoop % 3] = nQuantValue;
			
			//dmlab
			//coord.push_back(InverseQuantization(nQuantValue,(nLoop%3)));
			outIFS_->color_[nLoop] =  InverseQuantization(nQuantValue, quantRangeForColor[nLoop%3] ,quantMinForColor[nLoop%3], quantValueforColor);
			

			 //~dmlab
		}
	}
	else
	{
		unsigned int nLoop, nSign;
		int difValue, nQuantValue;
		nQuantValue=0;
//		m_pReadStream = m_pReadStream+unCurrentBufferPosition;
		unsigned int code_bytes = 0;
		ReadFromStream(m_pReadStream, &code_bytes, sizeof(unsigned int));

		unsigned char * code_buffer = new unsigned char [code_bytes];                                                   
		ReadFromStream(m_pReadStream, code_buffer, sizeof(unsigned char)*code_bytes);

		
		Adaptive_Data_Model mModel(1<<QPforGeometry);
		Adaptive_Data_Model mSign (2);
		Arithmetic_Codec acd(code_bytes, code_buffer);
		acd.start_decoder();	// initialize decoder

		
		

		for(nLoop = 0; nLoop < 3 * numberOfCoord; nLoop++)
		{
			difValue = acd.decode(mModel);
			if(difValue != 0)
			{
				nSign = acd.decode(mSign);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}

			nQuantValue = InverseCircularDifferenceForColor(previousQuant[nLoop % 3],difValue);
			
			previousQuant[nLoop % 3] = nQuantValue;
			
			//dmlab
			//coord.push_back(InverseQuantization(nQuantValue,(nLoop%3)));
			outIFS_->color_[nLoop] =  InverseQuantization(nQuantValue, quantRangeForColor[nLoop%3] ,quantMinForColor[nLoop%3], quantValueforColor);
			 //~dmlab
		}

	}
}

void SVADecoder::DecoderTexCoord()
{
	if(numberOfTexCoord == 0)
		return;

	ReadFromStream(m_pReadStream, &u, sizeof(int));
	ReadFromStream(m_pReadStream, &v, sizeof(int));

	for(int i=0; i<3; i++)
	{
		previousQuant[i] = 0;
	}
	if(entropytype == BP)
	{
		
		unsigned int nLoop, prefixSize, nBPL, nPayload, nSign, nQuantValue;
		int difValue;
		prefixSize = ReadBitsFromStorage(LEN_PREFIX);

		for(nLoop = 0; nLoop <2 * numberOfTexCoord; nLoop++)
		{
			nBPL=ReadBitsFromStorage(prefixSize);
			if(nBPL > 2)
			{
				nPayload = ReadBitsFromStorage(nBPL-1);
			}
			else
			{
				nPayload = 0;
			}

			difValue = BPLTable[nBPL] + nPayload;
			if(difValue != 0)
			{
				nSign = ReadBitsFromStorage(1);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}
			nQuantValue = InverseCircularDifferenceForTexCoord(previousQuant[nLoop % 2],difValue);
			
			previousQuant[nLoop % 2] = nQuantValue;
			
			//dmlab
			//coord.push_back(InverseQuantization(nQuantValue,(nLoop%3)));
			if(nLoop%2 ==0) outIFS_->texCoord_[nLoop] =  InverseQuantization(nQuantValue, quantRangeForTexCoord[nLoop%2] ,quantMinForTexCoord[nLoop%2], u-1);
			else outIFS_->texCoord_[nLoop] =  InverseQuantization(nQuantValue, quantRangeForTexCoord[nLoop%2] ,quantMinForTexCoord[nLoop%2], v-1);
			//outIFS_->texCoord_[nLoop] =  InverseQuantization(nQuantValue, quantRangeForTexCoord[nLoop%2] ,quantMinForTexCoord[nLoop%2], quantValueforTexCoord);
			

			 //~dmlab
		}
	}
	else
	{
		unsigned int nLoop, nSign;
		int difValue, nQuantValue;
		nQuantValue=0;
//		m_pReadStream = m_pReadStream+unCurrentBufferPosition;
		unsigned int code_bytes = 0;
		ReadFromStream(m_pReadStream, &code_bytes, sizeof(unsigned int));

		unsigned char * code_buffer = new unsigned char [code_bytes];                                                   
		ReadFromStream(m_pReadStream, code_buffer, sizeof(unsigned char)*code_bytes);

		
		Adaptive_Data_Model mModel(1<<QPforGeometry);
		Adaptive_Data_Model mSign (2);
		Arithmetic_Codec acd(code_bytes, code_buffer);
		acd.start_decoder();	// initialize decoder
		

		for(nLoop = 0; nLoop < 2 * numberOfTexCoord; nLoop++)
		{
			difValue = acd.decode(mModel);
			if(difValue != 0)
			{
				nSign = acd.decode(mSign);
				if(nSign == 1)
				{
					difValue = -difValue;
				}
			}

			nQuantValue = InverseCircularDifferenceForGeometry(previousQuant[nLoop % 2],difValue);
			
			previousQuant[nLoop % 2] = nQuantValue;
			
			//dmlab
			//coord.push_back(InverseQuantization(nQuantValue,(nLoop%3)));
			if(nLoop%2 ==0) outIFS_->texCoord_[nLoop] =  InverseQuantization(nQuantValue, quantRangeForTexCoord[nLoop%2] ,quantMinForTexCoord[nLoop%2], u-1);
			else outIFS_->texCoord_[nLoop] =  InverseQuantization(nQuantValue, quantRangeForTexCoord[nLoop%2] ,quantMinForTexCoord[nLoop%2], v-1);
			 //~dmlab
		}

	}
}

void SVADecoder::DecoderNormalIndex()

{
	if(numberOfNormalIndex == 0)
		return;
	
	if(entropytype == BP)
	{
		unsigned int nLoop, prefixSize, nBPL, nSign;
		int difValue;
		FDmode = (int)ReadBitsFromStorage(LEN_BIT);
		if(FDmode == 0)
		{
			FDvalue = (int)ReadBitsFromStorage(LEN_BIT);
		}
		prefixSize = ReadBitsFromStorage(LEN_PREFIX);
		unsigned int nType, nFaceDirection, nNormalIndex, nPosition, nRotation;
		int nDifNormalIndex[3];
		unsigned int tCount, nCount;
		unsigned int nIndex;
		unsigned int tNormalIndex[3];
		unsigned int tChange;

		for(nLoop = 0 ; nLoop < 3; nLoop++)
		{
			nBPL = ReadBitsFromStorage(prefixSize);
			nNormalIndex = BPLTable[nBPL];
			if(nBPL > 2)
			{
				nNormalIndex += ReadBitsFromStorage(nBPL - 1);
			}
			outIFS_->normalIndex_[nLoop] = (nNormalIndex);

		}
		
		if(FDmode == 0)
		{
			for(nLoop = 0 ; nLoop < (numberOfCoordIndex - 1); nLoop++)
			{
				for(tCount = 0; tCount < 5; tCount++)
				{
					if( ReadBitsFromStorage(1) == 1)
					{
						nType = tCount;
						break;
					}
				}
				switch(nType)
				{
				case 0:
					nPosition = ReadBitsFromStorage(1);
					if(nPosition == 1)
					{
						nPosition += ReadBitsFromStorage(1);
					}
					//nFaceDirection = ReadBitsFromStorage(1);
					nBPL = ReadBitsFromStorage(prefixSize);
					nDifNormalIndex[0] = BPLTable[nBPL];
					if(nBPL > 2)
					{
						nDifNormalIndex[0] += ReadBitsFromStorage(nBPL - 1);
					}
					nDifNormalIndex[0]+=1;
					nSign = ReadBitsFromStorage(1);
					if(nSign == 1)
					{
						nDifNormalIndex[0] = -(nDifNormalIndex[0]);
					}

					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}

					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						if(nIndex == nPosition)
						{
							tNormalIndex[nIndex] = InverseCircularDifferenceForNormalIndex(outIFS_->normalIndex_[(3*nLoop)+nIndex],nDifNormalIndex[0]);
						}
						else
						{
							tNormalIndex[nIndex] = outIFS_->normalIndex_[(3*nLoop)+nIndex];
						}
					}
					//dmlab
				
					if(FDvalue == 1)
					{
						tChange = tNormalIndex[(nPosition+1)%3];
						tNormalIndex[(nPosition+1)%3] = tNormalIndex[(nPosition+2)%3] ;
						tNormalIndex[(nPosition+2)%3] = tChange;
					}
					InverseRotation(tNormalIndex,nRotation);

					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						//coordIndex.push_back(tNormalIndex[nIndex]);
						outIFS_->normalIndex_[((nLoop+1)*3)+nIndex] = tNormalIndex[nIndex];
					}
					break;

				case 1:
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifNormalIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifNormalIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nDifNormalIndex[nIndex]+=1;
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifNormalIndex[nIndex] = -(nDifNormalIndex[nIndex]);
						}
						outIFS_->normalIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForNormalIndex(outIFS_->coordIndex_[(3*nLoop)+nIndex],nDifNormalIndex[nIndex]));
					}

					break;
				case 2:
					nPosition = ReadBitsFromStorage(1);
					if(nPosition == 1)
					{
						nPosition += ReadBitsFromStorage(1);
					}
					for(nIndex = 0; nIndex < 2; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifNormalIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifNormalIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nDifNormalIndex[nIndex]+=1;
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifNormalIndex[nIndex] = -(nDifNormalIndex[nIndex]);
						}
					}
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}

					nCount = 0;
					tNormalIndex[nPosition] = outIFS_->normalIndex_[(3*nLoop)+nPosition];
					for(nIndex = 1; nIndex < 3; nIndex++)
					{
						tNormalIndex[((nPosition+nIndex)%3)] = InverseCircularDifferenceForNormalIndex(outIFS_->normalIndex_[(3*nLoop)+((nPosition+nIndex)%3)],nDifNormalIndex[nCount++]);
					}

					InverseRotation(tNormalIndex,nRotation);
					
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						outIFS_->normalIndex_[((nLoop+1)*3)+nIndex] = (tNormalIndex[nIndex]);
					}
					break;
				case 3:
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifNormalIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifNormalIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifNormalIndex[nIndex] = -(nDifNormalIndex[nIndex]);
						}
						outIFS_->normalIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForNormalIndex(outIFS_->normalIndex_[(3*nLoop)+nIndex],nDifNormalIndex[nIndex]));
					}
					break;
				case 4:
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}
					if(FDvalue == 1)
					{
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							tNormalIndex[nIndex] = outIFS_->normalIndex_[(3*nLoop)+2-nIndex];	
						}
						InverseRotation(tNormalIndex,nRotation);
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							outIFS_->normalIndex_[((nLoop+1)*3)+nIndex] = (tNormalIndex[nIndex]);
						}
					}
					else
					{
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							tNormalIndex[nIndex] = outIFS_->normalIndex_[(3*nLoop)+nIndex];
						}
						InverseRotation(tNormalIndex,nRotation);
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							outIFS_->normalIndex_[((nLoop+1)*3)+nIndex] = (tNormalIndex[nIndex]);
						}
					}
					break;
				}
			}
		}
		else
		{
			for(nLoop = 0 ; nLoop < (numberOfCoordIndex - 1); nLoop++)
			{
				for(tCount = 0; tCount < 4; tCount++)
				{
					if( ReadBitsFromStorage(1) == 1)
					{
						nType = tCount;
						break;
					}
				}
				switch(nType)
				{
				case 0:
					nPosition = ReadBitsFromStorage(1);
					if(nPosition == 1)
					{
						nPosition += ReadBitsFromStorage(1);
					}
					nFaceDirection = ReadBitsFromStorage(1);
					nBPL = ReadBitsFromStorage(prefixSize);
					nDifNormalIndex[0] = BPLTable[nBPL];
					if(nBPL > 2)
					{
						nDifNormalIndex[0] += ReadBitsFromStorage(nBPL - 1);
					}
					nDifNormalIndex[0]+=1;
					nSign = ReadBitsFromStorage(1);
					if(nSign == 1)
					{
						nDifNormalIndex[0] = -(nDifNormalIndex[0]);
					}
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}

					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						if(nIndex == nPosition)
						{
							tNormalIndex[nIndex] = InverseCircularDifferenceForNormalIndex(outIFS_->normalIndex_[(3*nLoop)+nIndex],nDifNormalIndex[0]);
						}
						else
						{
							tNormalIndex[nIndex] = outIFS_->normalIndex_[(3*nLoop)+nIndex];
						}
					}
					//dmlab

					if(nFaceDirection == 1)
					{
						tChange = tNormalIndex[(nPosition+1)%3];
						tNormalIndex[(nPosition+1)%3] = tNormalIndex[(nPosition+2)%3] ;
						tNormalIndex[(nPosition+2)%3] = tChange;
					}
					InverseRotation(tNormalIndex,nRotation);
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						
						outIFS_->normalIndex_[((nLoop+1)*3)+nIndex] = tNormalIndex[nIndex];
					}
					break;

				case 1:
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifNormalIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifNormalIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nDifNormalIndex[nIndex]+=1;
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifNormalIndex[nIndex] = -(nDifNormalIndex[nIndex]);
						}
						//coordIndex.push_back(InverseCircularDifferenceForNormalIndex(coordIndex[(3*nLoop)+nIndex],nDifNormalIndex[nIndex]));
						outIFS_->normalIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForNormalIndex(outIFS_->normalIndex_[(3*nLoop)+nIndex],nDifNormalIndex[nIndex]));
					}

					break;
				case 2:
					nPosition = ReadBitsFromStorage(1);
					if(nPosition == 1)
					{
						nPosition += ReadBitsFromStorage(1);
					}
					for(nIndex = 0; nIndex < 2; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifNormalIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifNormalIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nDifNormalIndex[nIndex]+=1;
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifNormalIndex[nIndex] = -(nDifNormalIndex[nIndex]);
						}
					}
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}
					nCount = 0;
					tNormalIndex[nPosition] = outIFS_->normalIndex_[(3*nLoop)+nPosition];
					for(nIndex = 1; nIndex < 3; nIndex++)
					{
						tNormalIndex[((nPosition+nIndex)%3)] = InverseCircularDifferenceForNormalIndex(outIFS_->normalIndex_[(3*nLoop)+((nPosition+nIndex)%3)],nDifNormalIndex[nCount++]);
					}
					InverseRotation(tNormalIndex,nRotation);
					
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						outIFS_->normalIndex_[((nLoop+1)*3)+nIndex] = (tNormalIndex[nIndex]);
					}
					break;
				case 3:
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifNormalIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifNormalIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifNormalIndex[nIndex] = -(nDifNormalIndex[nIndex]);
						}
						outIFS_->normalIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForNormalIndex(outIFS_->normalIndex_[(3*nLoop)+nIndex],nDifNormalIndex[nIndex]));
					}
					break;
				case 4:
					nFaceDirection = ReadBitsFromStorage(1);
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}
					if(nFaceDirection == 1)
					{
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							tNormalIndex[nIndex] = outIFS_->normalIndex_[(3*nLoop)+2-nIndex];	
						}
						InverseRotation(tNormalIndex,nRotation);
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							outIFS_->normalIndex_[((nLoop+1)*3)+nIndex] = (tNormalIndex[nIndex]);
						}
					}
					else
					{
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							tNormalIndex[nIndex] = outIFS_->normalIndex_[(3*nLoop)+nIndex];
						}
						InverseRotation(tNormalIndex,nRotation);
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							outIFS_->normalIndex_[((nLoop+1)*3)+nIndex] = (tNormalIndex[nIndex]);
						}
					}
					break;

				}
			}

		}
	}
	else
	{
		unsigned int nLoop, nSign;
		int difValue;
		unsigned int nType, nFaceDirection, nNormalIndex, nPosition, nRotation;
		int nDifNormalIndex[3];
		unsigned int tCount, nCount;
		unsigned int nIndex;
		unsigned int tNormalIndex[3];
		unsigned int tChange;

		unsigned int code_bytes = 0;
		ReadFromStream(m_pReadStream, &code_bytes, sizeof(unsigned int));
		unsigned char * code_buffer = new unsigned char [code_bytes];                                                   
		ReadFromStream(m_pReadStream, code_buffer, sizeof(unsigned char)*code_bytes);
		
		Adaptive_Data_Model mModel(1024);
		Adaptive_Data_Model mhasnext (2);
		Adaptive_Data_Model mSign (2);
		Adaptive_Data_Model mType (5);
		Adaptive_Data_Model mFD (2);
		Adaptive_Data_Model mPos (3);
		Adaptive_Data_Model mRotate (3);
		Arithmetic_Codec acd(code_bytes, code_buffer);
		acd.start_decoder();	// initialize decoder

		int data[4];
		data[0] = 1;
		data[1] = 1<<10;
		data[2] = 1<<20;
		data[3] = 1<<30;
		int count = 0;

		for(nLoop = 0 ; nLoop < 3; nLoop++)
		{
			
			difValue = 0;
			while(1)
			{
				difValue += data[count] * acd.decode(mModel);
				count++;
				int temp = acd.decode(mhasnext);

				if( temp ==0)
					break;
			}
			count = 0;
			outIFS_->normalIndex_[nLoop] = (difValue);

		}
		for(nLoop = 0 ; nLoop < (numberOfCoordIndex - 1); nLoop++)
		{
			nType = acd.decode(mType);
			
			switch(nType)
			{
			case 0:
				nPosition = acd.decode(mPos);
				nFaceDirection =acd.decode(mFD);
				
				difValue = 0;
				while(1)
				{
					difValue += data[count] * acd.decode(mModel);
					count++;
					int temp = acd.decode(mhasnext);

					if( temp ==0)
						break;
				}
				difValue +=1;
				count = 0;
				nSign = acd.decode(mSign);
				if(nSign == 1)
				{
					difValue = -(difValue);
				}

				nRotation = acd.decode(mRotate);
				nDifNormalIndex[0] = difValue;
			
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					if(nIndex == nPosition)
					{
						tNormalIndex[nIndex] = InverseCircularDifferenceForNormalIndex(outIFS_->normalIndex_[(3*nLoop)+nIndex],nDifNormalIndex[0]);
					}
					else
					{
						tNormalIndex[nIndex] = outIFS_->normalIndex_[(3*nLoop)+nIndex];
					}
				}
				//dmlab
				if(nFaceDirection == 1)
				{
					tChange = tNormalIndex[(nPosition+1)%3];
					tNormalIndex[(nPosition+1)%3] = tNormalIndex[(nPosition+2)%3] ;
					tNormalIndex[(nPosition+2)%3] = tChange;
				}
				InverseRotation(tNormalIndex,nRotation);
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					//coordIndex.push_back(tNormalIndex[nIndex]);
					outIFS_->normalIndex_[((nLoop+1)*3)+nIndex] = tNormalIndex[nIndex];
				}
				break;

			case 1:
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					difValue = 0;
					while(1)
					{
						difValue += data[count] * acd.decode(mModel);
						count++;
						int temp = acd.decode(mhasnext);

						if( temp ==0)
							break;
					}
					difValue +=1;
					count = 0;
					nSign = acd.decode(mSign);
					if(nSign == 1)
					{
						difValue = -(difValue);
					}
					nDifNormalIndex[nIndex] = difValue;
					//coordIndex.push_back(InverseCircularDifferenceForNormalIndex(coordIndex[(3*nLoop)+nIndex],nDifNormalIndex[nIndex]));
					outIFS_->normalIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForNormalIndex(outIFS_->normalIndex_[(3*nLoop)+nIndex],nDifNormalIndex[nIndex]));
				}

				break;
			case 2:
				nPosition = acd.decode(mPos);
				for(nIndex = 0; nIndex < 2; nIndex++)
				{
					difValue = 0;
					while(1)
					{
						difValue += data[count] * acd.decode(mModel);
						count++;
						int temp = acd.decode(mhasnext);

						if( temp ==0)
							break;
					}
					count = 0;
					difValue +=1;
					nSign = acd.decode(mSign);
					if(nSign == 1)
					{
						difValue = -(difValue);
					}
					nDifNormalIndex[nIndex] = difValue;
					
				}
				nRotation = acd.decode(mRotate);
				nCount = 0;
				tNormalIndex[nPosition] = outIFS_->normalIndex_[(3*nLoop)+nPosition];
				for(nIndex = 1; nIndex < 3; nIndex++)
				{
					tNormalIndex[((nPosition+nIndex)%3)] = InverseCircularDifferenceForNormalIndex(outIFS_->normalIndex_[(3*nLoop)+((nPosition+nIndex)%3)],nDifNormalIndex[nCount++]);
				}
				InverseRotation(tNormalIndex,nRotation);
				
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					outIFS_->normalIndex_[((nLoop+1)*3)+nIndex] = (tNormalIndex[nIndex]);
				}
				break;
			case 3:
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					difValue = 0;
					while(1)
					{
						difValue += data[count] * acd.decode(mModel);
						count++;
						int temp = acd.decode(mhasnext);

						if( temp ==0)
							break;
					}
					count = 0;
					nSign = acd.decode(mSign);
					if(nSign == 1)
					{
						difValue = -(difValue);
					}
					nDifNormalIndex[nIndex] = difValue;

					outIFS_->normalIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForNormalIndex(outIFS_->normalIndex_[(3*nLoop)+nIndex],nDifNormalIndex[nIndex]));
				}
				break;
			case 4:
				nFaceDirection =  acd.decode(mFD);
				nRotation =  acd.decode(mRotate);
				if(nFaceDirection == 1)
				{
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						tNormalIndex[nIndex] = outIFS_->normalIndex_[(3*nLoop)+2-nIndex];	
					}
					InverseRotation(tNormalIndex,nRotation);
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						outIFS_->normalIndex_[((nLoop+1)*3)+nIndex] = (tNormalIndex[nIndex]);
					}
				}
				else
				{
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						tNormalIndex[nIndex] = outIFS_->normalIndex_[(3*nLoop)+nIndex];
					}
					InverseRotation(tNormalIndex,nRotation);
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						outIFS_->normalIndex_[((nLoop+1)*3)+nIndex] = (tNormalIndex[nIndex]);
					}
				}
				break;

			}
		}

	}

}


void SVADecoder::DecoderColorIndex()
{
	if(numberOfColorIndex == 0)
		return;
	if(entropytype == BP)
	{
		unsigned int nLoop, prefixSize, nBPL, nSign;
		int difValue;
		FDmode = (int)ReadBitsFromStorage(LEN_BIT);
		if(FDmode == 0)
		{
			FDvalue = (int)ReadBitsFromStorage(LEN_BIT);
		}
		prefixSize = ReadBitsFromStorage(LEN_PREFIX);
		unsigned int nType, nFaceDirection, nColorIndex, nPosition, nRotation;
		int nDifColorIndex[3];
		unsigned int tCount, nCount;
		unsigned int nIndex;
		unsigned int tColorIndex[3];
		unsigned int tChange;

		for(nLoop = 0 ; nLoop < 3; nLoop++)
		{
			nBPL = ReadBitsFromStorage(prefixSize);
			nColorIndex = BPLTable[nBPL];
			if(nBPL > 2)
			{
				nColorIndex += ReadBitsFromStorage(nBPL - 1);
			}
			outIFS_->colorIndex_[nLoop] = (nColorIndex);

		}
		
		if(FDmode == 0)
		{
			for(nLoop = 0 ; nLoop < (numberOfCoordIndex - 1); nLoop++)
			{
				for(tCount = 0; tCount < 5; tCount++)
				{
					if( ReadBitsFromStorage(1) == 1)
					{
						nType = tCount;
						break;
					}
				}
				switch(nType)
				{
				case 0:
					nPosition = ReadBitsFromStorage(1);
					if(nPosition == 1)
					{
						nPosition += ReadBitsFromStorage(1);
					}
					//nFaceDirection = ReadBitsFromStorage(1);
					nBPL = ReadBitsFromStorage(prefixSize);
					nDifColorIndex[0] = BPLTable[nBPL];
					if(nBPL > 2)
					{
						nDifColorIndex[0] += ReadBitsFromStorage(nBPL - 1);
					}
					nDifColorIndex[0]+=1;
					nSign = ReadBitsFromStorage(1);
					if(nSign == 1)
					{
						nDifColorIndex[0] = -(nDifColorIndex[0]);
					}

					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}

					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						if(nIndex == nPosition)
						{
							tColorIndex[nIndex] = InverseCircularDifferenceForColorIndex(outIFS_->colorIndex_[(3*nLoop)+nIndex],nDifColorIndex[0]);
						}
						else
						{
							tColorIndex[nIndex] = outIFS_->colorIndex_[(3*nLoop)+nIndex];
						}
					}
					//dmlab
				
					if(FDvalue == 1)
					{
						tChange = tColorIndex[(nPosition+1)%3];
						tColorIndex[(nPosition+1)%3] = tColorIndex[(nPosition+2)%3] ;
						tColorIndex[(nPosition+2)%3] = tChange;
					}
					InverseRotation(tColorIndex,nRotation);

					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						//coordIndex.push_back(tColorIndex[nIndex]);
						outIFS_->colorIndex_[((nLoop+1)*3)+nIndex] = tColorIndex[nIndex];
					}
					break;

				case 1:
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifColorIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifColorIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nDifColorIndex[nIndex]+=1;
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifColorIndex[nIndex] = -(nDifColorIndex[nIndex]);
						}
						outIFS_->colorIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForColorIndex(outIFS_->coordIndex_[(3*nLoop)+nIndex],nDifColorIndex[nIndex]));
					}

					break;
				case 2:
					nPosition = ReadBitsFromStorage(1);
					if(nPosition == 1)
					{
						nPosition += ReadBitsFromStorage(1);
					}
					for(nIndex = 0; nIndex < 2; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifColorIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifColorIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nDifColorIndex[nIndex]+=1;
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifColorIndex[nIndex] = -(nDifColorIndex[nIndex]);
						}
					}
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}

					nCount = 0;
					tColorIndex[nPosition] = outIFS_->colorIndex_[(3*nLoop)+nPosition];
					for(nIndex = 1; nIndex < 3; nIndex++)
					{
						tColorIndex[((nPosition+nIndex)%3)] = InverseCircularDifferenceForColorIndex(outIFS_->colorIndex_[(3*nLoop)+((nPosition+nIndex)%3)],nDifColorIndex[nCount++]);
					}

					InverseRotation(tColorIndex,nRotation);
					
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						outIFS_->colorIndex_[((nLoop+1)*3)+nIndex] = (tColorIndex[nIndex]);
					}
					break;
				case 3:
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifColorIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifColorIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifColorIndex[nIndex] = -(nDifColorIndex[nIndex]);
						}
						outIFS_->colorIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForColorIndex(outIFS_->colorIndex_[(3*nLoop)+nIndex],nDifColorIndex[nIndex]));
					}
					break;
				case 4:
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}
					if(FDvalue == 1)
					{
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							tColorIndex[nIndex] = outIFS_->colorIndex_[(3*nLoop)+2-nIndex];	
						}
						InverseRotation(tColorIndex,nRotation);
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							outIFS_->colorIndex_[((nLoop+1)*3)+nIndex] = (tColorIndex[nIndex]);
						}
					}
					else
					{
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							tColorIndex[nIndex] = outIFS_->colorIndex_[(3*nLoop)+nIndex];
						}
						InverseRotation(tColorIndex,nRotation);
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							outIFS_->colorIndex_[((nLoop+1)*3)+nIndex] = (tColorIndex[nIndex]);
						}
					}
					break;
				}
			}
		}
		else
		{
			for(nLoop = 0 ; nLoop < (numberOfCoordIndex - 1); nLoop++)
			{
				for(tCount = 0; tCount < 4; tCount++)
				{
					if( ReadBitsFromStorage(1) == 1)
					{
						nType = tCount;
						break;
					}
				}
				switch(nType)
				{
				case 0:
					nPosition = ReadBitsFromStorage(1);
					if(nPosition == 1)
					{
						nPosition += ReadBitsFromStorage(1);
					}
					nFaceDirection = ReadBitsFromStorage(1);
					nBPL = ReadBitsFromStorage(prefixSize);
					nDifColorIndex[0] = BPLTable[nBPL];
					if(nBPL > 2)
					{
						nDifColorIndex[0] += ReadBitsFromStorage(nBPL - 1);
					}
					nDifColorIndex[0]+=1;
					nSign = ReadBitsFromStorage(1);
					if(nSign == 1)
					{
						nDifColorIndex[0] = -(nDifColorIndex[0]);
					}
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}

					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						if(nIndex == nPosition)
						{
							tColorIndex[nIndex] = InverseCircularDifferenceForColorIndex(outIFS_->colorIndex_[(3*nLoop)+nIndex],nDifColorIndex[0]);
						}
						else
						{
							tColorIndex[nIndex] = outIFS_->colorIndex_[(3*nLoop)+nIndex];
						}
					}
					//dmlab

					if(nFaceDirection == 1)
					{
						tChange = tColorIndex[(nPosition+1)%3];
						tColorIndex[(nPosition+1)%3] = tColorIndex[(nPosition+2)%3] ;
						tColorIndex[(nPosition+2)%3] = tChange;
					}
					InverseRotation(tColorIndex,nRotation);
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						
						outIFS_->colorIndex_[((nLoop+1)*3)+nIndex] = tColorIndex[nIndex];
					}
					break;

				case 1:
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifColorIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifColorIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nDifColorIndex[nIndex]+=1;
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifColorIndex[nIndex] = -(nDifColorIndex[nIndex]);
						}
						//coordIndex.push_back(InverseCircularDifferenceForColorIndex(coordIndex[(3*nLoop)+nIndex],nDifColorIndex[nIndex]));
						outIFS_->colorIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForColorIndex(outIFS_->colorIndex_[(3*nLoop)+nIndex],nDifColorIndex[nIndex]));
					}

					break;
				case 2:
					nPosition = ReadBitsFromStorage(1);
					if(nPosition == 1)
					{
						nPosition += ReadBitsFromStorage(1);
					}
					for(nIndex = 0; nIndex < 2; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifColorIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifColorIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nDifColorIndex[nIndex]+=1;
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifColorIndex[nIndex] = -(nDifColorIndex[nIndex]);
						}
					}
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}
					nCount = 0;
					tColorIndex[nPosition] = outIFS_->colorIndex_[(3*nLoop)+nPosition];
					for(nIndex = 1; nIndex < 3; nIndex++)
					{
						tColorIndex[((nPosition+nIndex)%3)] = InverseCircularDifferenceForColorIndex(outIFS_->colorIndex_[(3*nLoop)+((nPosition+nIndex)%3)],nDifColorIndex[nCount++]);
					}
					InverseRotation(tColorIndex,nRotation);
					
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						outIFS_->colorIndex_[((nLoop+1)*3)+nIndex] = (tColorIndex[nIndex]);
					}
					break;
				case 3:
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifColorIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifColorIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifColorIndex[nIndex] = -(nDifColorIndex[nIndex]);
						}
						outIFS_->colorIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForColorIndex(outIFS_->colorIndex_[(3*nLoop)+nIndex],nDifColorIndex[nIndex]));
					}
					break;
				case 4:
					nFaceDirection = ReadBitsFromStorage(1);
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}
					if(nFaceDirection == 1)
					{
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							tColorIndex[nIndex] = outIFS_->colorIndex_[(3*nLoop)+2-nIndex];	
						}
						InverseRotation(tColorIndex,nRotation);
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							outIFS_->colorIndex_[((nLoop+1)*3)+nIndex] = (tColorIndex[nIndex]);
						}
					}
					else
					{
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							tColorIndex[nIndex] = outIFS_->colorIndex_[(3*nLoop)+nIndex];
						}
						InverseRotation(tColorIndex,nRotation);
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							outIFS_->colorIndex_[((nLoop+1)*3)+nIndex] = (tColorIndex[nIndex]);
						}
					}
					break;

				}
			}

		}
	}
	else
	{
		unsigned int nLoop, nSign;
		int difValue;
		unsigned int nType, nFaceDirection, nColorIndex, nPosition, nRotation;
		int nDifColorIndex[3];
		unsigned int tCount, nCount;
		unsigned int nIndex;
		unsigned int tColorIndex[3];
		unsigned int tChange;

		unsigned int code_bytes = 0;
		ReadFromStream(m_pReadStream, &code_bytes, sizeof(unsigned int));
		unsigned char * code_buffer = new unsigned char [code_bytes];                                                   
		ReadFromStream(m_pReadStream, code_buffer, sizeof(unsigned char)*code_bytes);
		
		Adaptive_Data_Model mModel(1024);
		Adaptive_Data_Model mhasnext (2);
		Adaptive_Data_Model mSign (2);
		Adaptive_Data_Model mType (5);
		Adaptive_Data_Model mFD (2);
		Adaptive_Data_Model mPos (3);
		Adaptive_Data_Model mRotate (3);
		Arithmetic_Codec acd(code_bytes, code_buffer);
		acd.start_decoder();	// initialize decoder

		int data[4];
		data[0] = 1;
		data[1] = 1<<10;
		data[2] = 1<<20;
		data[3] = 1<<30;
		int count = 0;

		for(nLoop = 0 ; nLoop < 3; nLoop++)
		{
			
			difValue = 0;
			while(1)
			{
				difValue += data[count] * acd.decode(mModel);
				count++;
				int temp = acd.decode(mhasnext);

				if( temp ==0)
					break;
			}
			count = 0;
			outIFS_->colorIndex_[nLoop] = (difValue);

		}
		for(nLoop = 0 ; nLoop < (numberOfCoordIndex - 1); nLoop++)
		{
			nType = acd.decode(mType);
			
			switch(nType)
			{
			case 0:
				nPosition = acd.decode(mPos);
				nFaceDirection =acd.decode(mFD);
				
				difValue = 0;
				while(1)
				{
					difValue += data[count] * acd.decode(mModel);
					count++;
					int temp = acd.decode(mhasnext);

					if( temp ==0)
						break;
				}
				difValue +=1;
				count = 0;
				nSign = acd.decode(mSign);
				if(nSign == 1)
				{
					difValue = -(difValue);
				}

				nRotation = acd.decode(mRotate);
				nDifColorIndex[0] = difValue;
			
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					if(nIndex == nPosition)
					{
						tColorIndex[nIndex] = InverseCircularDifferenceForColorIndex(outIFS_->colorIndex_[(3*nLoop)+nIndex],nDifColorIndex[0]);
					}
					else
					{
						tColorIndex[nIndex] = outIFS_->colorIndex_[(3*nLoop)+nIndex];
					}
				}
				//dmlab
				if(nFaceDirection == 1)
				{
					tChange = tColorIndex[(nPosition+1)%3];
					tColorIndex[(nPosition+1)%3] = tColorIndex[(nPosition+2)%3] ;
					tColorIndex[(nPosition+2)%3] = tChange;
				}
				InverseRotation(tColorIndex,nRotation);
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					//coordIndex.push_back(tColorIndex[nIndex]);
					outIFS_->colorIndex_[((nLoop+1)*3)+nIndex] = tColorIndex[nIndex];
				}
				break;

			case 1:
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					difValue = 0;
					while(1)
					{
						difValue += data[count] * acd.decode(mModel);
						count++;
						int temp = acd.decode(mhasnext);

						if( temp ==0)
							break;
					}
					difValue +=1;
					count = 0;
					nSign = acd.decode(mSign);
					if(nSign == 1)
					{
						difValue = -(difValue);
					}
					nDifColorIndex[nIndex] = difValue;
					//coordIndex.push_back(InverseCircularDifferenceForColorIndex(coordIndex[(3*nLoop)+nIndex],nDifColorIndex[nIndex]));
					outIFS_->colorIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForColorIndex(outIFS_->colorIndex_[(3*nLoop)+nIndex],nDifColorIndex[nIndex]));
				}

				break;
			case 2:
				nPosition = acd.decode(mPos);
				for(nIndex = 0; nIndex < 2; nIndex++)
				{
					difValue = 0;
					while(1)
					{
						difValue += data[count] * acd.decode(mModel);
						count++;
						int temp = acd.decode(mhasnext);

						if( temp ==0)
							break;
					}
					count = 0;
					difValue +=1;
					nSign = acd.decode(mSign);
					if(nSign == 1)
					{
						difValue = -(difValue);
					}
					nDifColorIndex[nIndex] = difValue;
					
				}
				nRotation = acd.decode(mRotate);
				nCount = 0;
				tColorIndex[nPosition] = outIFS_->colorIndex_[(3*nLoop)+nPosition];
				for(nIndex = 1; nIndex < 3; nIndex++)
				{
					tColorIndex[((nPosition+nIndex)%3)] = InverseCircularDifferenceForColorIndex(outIFS_->colorIndex_[(3*nLoop)+((nPosition+nIndex)%3)],nDifColorIndex[nCount++]);
				}
				InverseRotation(tColorIndex,nRotation);
				
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					outIFS_->colorIndex_[((nLoop+1)*3)+nIndex] = (tColorIndex[nIndex]);
				}
				break;
			case 3:
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					difValue = 0;
					while(1)
					{
						difValue += data[count] * acd.decode(mModel);
						count++;
						int temp = acd.decode(mhasnext);

						if( temp ==0)
							break;
					}
					count = 0;
					nSign = acd.decode(mSign);
					if(nSign == 1)
					{
						difValue = -(difValue);
					}
					nDifColorIndex[nIndex] = difValue;

					outIFS_->colorIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForColorIndex(outIFS_->colorIndex_[(3*nLoop)+nIndex],nDifColorIndex[nIndex]));
				}
				break;
			case 4:
				nFaceDirection =  acd.decode(mFD);
				nRotation =  acd.decode(mRotate);
				if(nFaceDirection == 1)
				{
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						tColorIndex[nIndex] = outIFS_->colorIndex_[(3*nLoop)+2-nIndex];	
					}
					InverseRotation(tColorIndex,nRotation);
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						outIFS_->colorIndex_[((nLoop+1)*3)+nIndex] = (tColorIndex[nIndex]);
					}
				}
				else
				{
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						tColorIndex[nIndex] = outIFS_->colorIndex_[(3*nLoop)+nIndex];
					}
					InverseRotation(tColorIndex,nRotation);
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						outIFS_->colorIndex_[((nLoop+1)*3)+nIndex] = (tColorIndex[nIndex]);
					}
				}
				break;

			}
		}

	}
}

void SVADecoder::DecoderTexCoordIndex()
{
	if(numberOfTexCoordIndex == 0)
		return;
	if(entropytype == BP)
	{
		unsigned int nLoop, prefixSize, nBPL, nSign;
		int difValue;
		FDmode = (int)ReadBitsFromStorage(LEN_BIT);
		if(FDmode == 0)
		{
			FDvalue = (int)ReadBitsFromStorage(LEN_BIT);
		}
		prefixSize = ReadBitsFromStorage(LEN_PREFIX);
		unsigned int nType, nFaceDirection, nTexCoordIndex, nPosition, nRotation;
		int nDifTexCoordIndex[3];
		unsigned int tCount, nCount;
		unsigned int nIndex;
		unsigned int tTexCoordIndex[3];
		unsigned int tChange;

		for(nLoop = 0 ; nLoop < 3; nLoop++)
		{
			nBPL = ReadBitsFromStorage(prefixSize);
			nTexCoordIndex = BPLTable[nBPL];
			if(nBPL > 2)
			{
				nTexCoordIndex += ReadBitsFromStorage(nBPL - 1);
			}
			outIFS_->texCoordIndex_[nLoop] = (nTexCoordIndex);

		}
		
		if(FDmode == 0)
		{
			for(nLoop = 0 ; nLoop < (numberOfCoordIndex - 1); nLoop++)
			{
				for(tCount = 0; tCount < 5; tCount++)
				{
					if( ReadBitsFromStorage(1) == 1)
					{
						nType = tCount;
						break;
					}
				}
				switch(nType)
				{
				case 0:
					nPosition = ReadBitsFromStorage(1);
					if(nPosition == 1)
					{
						nPosition += ReadBitsFromStorage(1);
					}
					//nFaceDirection = ReadBitsFromStorage(1);
					nBPL = ReadBitsFromStorage(prefixSize);
					nDifTexCoordIndex[0] = BPLTable[nBPL];
					if(nBPL > 2)
					{
						nDifTexCoordIndex[0] += ReadBitsFromStorage(nBPL - 1);
					}
					nDifTexCoordIndex[0]+=1;
					nSign = ReadBitsFromStorage(1);
					if(nSign == 1)
					{
						nDifTexCoordIndex[0] = -(nDifTexCoordIndex[0]);
					}

					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}

					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						if(nIndex == nPosition)
						{
							tTexCoordIndex[nIndex] = InverseCircularDifferenceForTexCoordIndex(outIFS_->texCoordIndex_[(3*nLoop)+nIndex],nDifTexCoordIndex[0]);
						}
						else
						{
							tTexCoordIndex[nIndex] = outIFS_->texCoordIndex_[(3*nLoop)+nIndex];
						}
					}
					//dmlab
				
					if(FDvalue == 1)
					{
						tChange = tTexCoordIndex[(nPosition+1)%3];
						tTexCoordIndex[(nPosition+1)%3] = tTexCoordIndex[(nPosition+2)%3] ;
						tTexCoordIndex[(nPosition+2)%3] = tChange;
					}
					InverseRotation(tTexCoordIndex,nRotation);

					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						//coordIndex.push_back(tTexCoordIndex[nIndex]);
						outIFS_->texCoordIndex_[((nLoop+1)*3)+nIndex] = tTexCoordIndex[nIndex];
					}
					break;

				case 1:
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifTexCoordIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifTexCoordIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nDifTexCoordIndex[nIndex]+=1;
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifTexCoordIndex[nIndex] = -(nDifTexCoordIndex[nIndex]);
						}
						outIFS_->texCoordIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForTexCoordIndex(outIFS_->coordIndex_[(3*nLoop)+nIndex],nDifTexCoordIndex[nIndex]));
					}

					break;
				case 2:
					nPosition = ReadBitsFromStorage(1);
					if(nPosition == 1)
					{
						nPosition += ReadBitsFromStorage(1);
					}
					for(nIndex = 0; nIndex < 2; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifTexCoordIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifTexCoordIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nDifTexCoordIndex[nIndex]+=1;
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifTexCoordIndex[nIndex] = -(nDifTexCoordIndex[nIndex]);
						}
					}
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}

					nCount = 0;
					tTexCoordIndex[nPosition] = outIFS_->texCoordIndex_[(3*nLoop)+nPosition];
					for(nIndex = 1; nIndex < 3; nIndex++)
					{
						tTexCoordIndex[((nPosition+nIndex)%3)] = InverseCircularDifferenceForTexCoordIndex(outIFS_->texCoordIndex_[(3*nLoop)+((nPosition+nIndex)%3)],nDifTexCoordIndex[nCount++]);
					}

					InverseRotation(tTexCoordIndex,nRotation);
					
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						outIFS_->texCoordIndex_[((nLoop+1)*3)+nIndex] = (tTexCoordIndex[nIndex]);
					}
					break;
				case 3:
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifTexCoordIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifTexCoordIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifTexCoordIndex[nIndex] = -(nDifTexCoordIndex[nIndex]);
						}
						outIFS_->texCoordIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForTexCoordIndex(outIFS_->texCoordIndex_[(3*nLoop)+nIndex],nDifTexCoordIndex[nIndex]));
					}
					break;
				case 4:
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}
					if(FDvalue == 1)
					{
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							tTexCoordIndex[nIndex] = outIFS_->texCoordIndex_[(3*nLoop)+2-nIndex];	
						}
						InverseRotation(tTexCoordIndex,nRotation);
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							outIFS_->texCoordIndex_[((nLoop+1)*3)+nIndex] = (tTexCoordIndex[nIndex]);
						}
					}
					else
					{
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							tTexCoordIndex[nIndex] = outIFS_->texCoordIndex_[(3*nLoop)+nIndex];
						}
						InverseRotation(tTexCoordIndex,nRotation);
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							outIFS_->texCoordIndex_[((nLoop+1)*3)+nIndex] = (tTexCoordIndex[nIndex]);
						}
					}
					break;
				}
			}
		}
		else
		{
			for(nLoop = 0 ; nLoop < (numberOfCoordIndex - 1); nLoop++)
			{
				for(tCount = 0; tCount < 4; tCount++)
				{
					if( ReadBitsFromStorage(1) == 1)
					{
						nType = tCount;
						break;
					}
				}
				switch(nType)
				{
				case 0:
					nPosition = ReadBitsFromStorage(1);
					if(nPosition == 1)
					{
						nPosition += ReadBitsFromStorage(1);
					}
					nFaceDirection = ReadBitsFromStorage(1);
					nBPL = ReadBitsFromStorage(prefixSize);
					nDifTexCoordIndex[0] = BPLTable[nBPL];
					if(nBPL > 2)
					{
						nDifTexCoordIndex[0] += ReadBitsFromStorage(nBPL - 1);
					}
					nDifTexCoordIndex[0]+=1;
					nSign = ReadBitsFromStorage(1);
					if(nSign == 1)
					{
						nDifTexCoordIndex[0] = -(nDifTexCoordIndex[0]);
					}
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}

					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						if(nIndex == nPosition)
						{
							tTexCoordIndex[nIndex] = InverseCircularDifferenceForTexCoordIndex(outIFS_->texCoordIndex_[(3*nLoop)+nIndex],nDifTexCoordIndex[0]);
						}
						else
						{
							tTexCoordIndex[nIndex] = outIFS_->texCoordIndex_[(3*nLoop)+nIndex];
						}
					}
					//dmlab

					if(nFaceDirection == 1)
					{
						tChange = tTexCoordIndex[(nPosition+1)%3];
						tTexCoordIndex[(nPosition+1)%3] = tTexCoordIndex[(nPosition+2)%3] ;
						tTexCoordIndex[(nPosition+2)%3] = tChange;
					}
					InverseRotation(tTexCoordIndex,nRotation);
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						
						outIFS_->texCoordIndex_[((nLoop+1)*3)+nIndex] = tTexCoordIndex[nIndex];
					}
					break;

				case 1:
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifTexCoordIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifTexCoordIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nDifTexCoordIndex[nIndex]+=1;
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifTexCoordIndex[nIndex] = -(nDifTexCoordIndex[nIndex]);
						}
						//coordIndex.push_back(InverseCircularDifferenceForTexCoordIndex(coordIndex[(3*nLoop)+nIndex],nDifTexCoordIndex[nIndex]));
						outIFS_->texCoordIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForTexCoordIndex(outIFS_->texCoordIndex_[(3*nLoop)+nIndex],nDifTexCoordIndex[nIndex]));
					}

					break;
				case 2:
					nPosition = ReadBitsFromStorage(1);
					if(nPosition == 1)
					{
						nPosition += ReadBitsFromStorage(1);
					}
					for(nIndex = 0; nIndex < 2; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifTexCoordIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifTexCoordIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nDifTexCoordIndex[nIndex]+=1;
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifTexCoordIndex[nIndex] = -(nDifTexCoordIndex[nIndex]);
						}
					}
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}
					nCount = 0;
					tTexCoordIndex[nPosition] = outIFS_->texCoordIndex_[(3*nLoop)+nPosition];
					for(nIndex = 1; nIndex < 3; nIndex++)
					{
						tTexCoordIndex[((nPosition+nIndex)%3)] = InverseCircularDifferenceForTexCoordIndex(outIFS_->texCoordIndex_[(3*nLoop)+((nPosition+nIndex)%3)],nDifTexCoordIndex[nCount++]);
					}
					InverseRotation(tTexCoordIndex,nRotation);
					
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						outIFS_->texCoordIndex_[((nLoop+1)*3)+nIndex] = (tTexCoordIndex[nIndex]);
					}
					break;
				case 3:
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifTexCoordIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifTexCoordIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifTexCoordIndex[nIndex] = -(nDifTexCoordIndex[nIndex]);
						}
						outIFS_->texCoordIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForTexCoordIndex(outIFS_->texCoordIndex_[(3*nLoop)+nIndex],nDifTexCoordIndex[nIndex]));
					}
					break;
				case 4:
					nFaceDirection = ReadBitsFromStorage(1);
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}
					if(nFaceDirection == 1)
					{
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							tTexCoordIndex[nIndex] = outIFS_->texCoordIndex_[(3*nLoop)+2-nIndex];	
						}
						InverseRotation(tTexCoordIndex,nRotation);
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							outIFS_->texCoordIndex_[((nLoop+1)*3)+nIndex] = (tTexCoordIndex[nIndex]);
						}
					}
					else
					{
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							tTexCoordIndex[nIndex] = outIFS_->texCoordIndex_[(3*nLoop)+nIndex];
						}
						InverseRotation(tTexCoordIndex,nRotation);
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							outIFS_->texCoordIndex_[((nLoop+1)*3)+nIndex] = (tTexCoordIndex[nIndex]);
						}
					}
					break;

				}
			}

		}
	}
	else
	{
		unsigned int nLoop, nSign;
		int difValue;
		unsigned int nType, nFaceDirection, nTexCoordIndex, nPosition, nRotation;
		int nDifTexCoordIndex[3];
		unsigned int tCount, nCount;
		unsigned int nIndex;
		unsigned int tTexCoordIndex[3];
		unsigned int tChange;

		unsigned int code_bytes = 0;
		ReadFromStream(m_pReadStream, &code_bytes, sizeof(unsigned int));
		unsigned char * code_buffer = new unsigned char [code_bytes];                                                   
		ReadFromStream(m_pReadStream, code_buffer, sizeof(unsigned char)*code_bytes);
		
		Adaptive_Data_Model mModel(1024);
		Adaptive_Data_Model mhasnext (2);
		Adaptive_Data_Model mSign (2);
		Adaptive_Data_Model mType (5);
		Adaptive_Data_Model mFD (2);
		Adaptive_Data_Model mPos (3);
		Adaptive_Data_Model mRotate (3);
		Arithmetic_Codec acd(code_bytes, code_buffer);
		acd.start_decoder();	// initialize decoder

		int data[4];
		data[0] = 1;
		data[1] = 1<<10;
		data[2] = 1<<20;
		data[3] = 1<<30;
		int count = 0;

		for(nLoop = 0 ; nLoop < 3; nLoop++)
		{
			
			difValue = 0;
			while(1)
			{
				difValue += data[count] * acd.decode(mModel);
				count++;
				int temp = acd.decode(mhasnext);

				if( temp ==0)
					break;
			}
			count = 0;
			outIFS_->texCoordIndex_[nLoop] = (difValue);

		}
		for(nLoop = 0 ; nLoop < (numberOfCoordIndex - 1); nLoop++)
		{
			nType = acd.decode(mType);
			
			switch(nType)
			{
			case 0:
				nPosition = acd.decode(mPos);
				nFaceDirection =acd.decode(mFD);
				
				difValue = 0;
				while(1)
				{
					difValue += data[count] * acd.decode(mModel);
					count++;
					int temp = acd.decode(mhasnext);

					if( temp ==0)
						break;
				}
				difValue +=1;
				count = 0;
				nSign = acd.decode(mSign);
				if(nSign == 1)
				{
					difValue = -(difValue);
				}

				nRotation = acd.decode(mRotate);
				nDifTexCoordIndex[0] = difValue;
			
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					if(nIndex == nPosition)
					{
						tTexCoordIndex[nIndex] = InverseCircularDifferenceForTexCoordIndex(outIFS_->texCoordIndex_[(3*nLoop)+nIndex],nDifTexCoordIndex[0]);
					}
					else
					{
						tTexCoordIndex[nIndex] = outIFS_->texCoordIndex_[(3*nLoop)+nIndex];
					}
				}
				//dmlab
				if(nFaceDirection == 1)
				{
					tChange = tTexCoordIndex[(nPosition+1)%3];
					tTexCoordIndex[(nPosition+1)%3] = tTexCoordIndex[(nPosition+2)%3] ;
					tTexCoordIndex[(nPosition+2)%3] = tChange;
				}
				InverseRotation(tTexCoordIndex,nRotation);
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					//coordIndex.push_back(tTexCoordIndex[nIndex]);
					outIFS_->texCoordIndex_[((nLoop+1)*3)+nIndex] = tTexCoordIndex[nIndex];
				}
				break;

			case 1:
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					difValue = 0;
					while(1)
					{
						difValue += data[count] * acd.decode(mModel);
						count++;
						int temp = acd.decode(mhasnext);

						if( temp ==0)
							break;
					}
					difValue +=1;
					count = 0;
					nSign = acd.decode(mSign);
					if(nSign == 1)
					{
						difValue = -(difValue);
					}
					nDifTexCoordIndex[nIndex] = difValue;
					//coordIndex.push_back(InverseCircularDifferenceForTexCoordIndex(coordIndex[(3*nLoop)+nIndex],nDifTexCoordIndex[nIndex]));
					outIFS_->texCoordIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForTexCoordIndex(outIFS_->texCoordIndex_[(3*nLoop)+nIndex],nDifTexCoordIndex[nIndex]));
				}

				break;
			case 2:
				nPosition = acd.decode(mPos);
				for(nIndex = 0; nIndex < 2; nIndex++)
				{
					difValue = 0;
					while(1)
					{
						difValue += data[count] * acd.decode(mModel);
						count++;
						int temp = acd.decode(mhasnext);

						if( temp ==0)
							break;
					}
					count = 0;
					difValue +=1;
					nSign = acd.decode(mSign);
					if(nSign == 1)
					{
						difValue = -(difValue);
					}
					nDifTexCoordIndex[nIndex] = difValue;
					
				}
				nRotation = acd.decode(mRotate);
				nCount = 0;
				tTexCoordIndex[nPosition] = outIFS_->texCoordIndex_[(3*nLoop)+nPosition];
				for(nIndex = 1; nIndex < 3; nIndex++)
				{
					tTexCoordIndex[((nPosition+nIndex)%3)] = InverseCircularDifferenceForTexCoordIndex(outIFS_->texCoordIndex_[(3*nLoop)+((nPosition+nIndex)%3)],nDifTexCoordIndex[nCount++]);
				}
				InverseRotation(tTexCoordIndex,nRotation);
				
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					outIFS_->texCoordIndex_[((nLoop+1)*3)+nIndex] = (tTexCoordIndex[nIndex]);
				}
				break;
			case 3:
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					difValue = 0;
					while(1)
					{
						difValue += data[count] * acd.decode(mModel);
						count++;
						int temp = acd.decode(mhasnext);

						if( temp ==0)
							break;
					}
					count = 0;
					nSign = acd.decode(mSign);
					if(nSign == 1)
					{
						difValue = -(difValue);
					}
					nDifTexCoordIndex[nIndex] = difValue;

					outIFS_->texCoordIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForTexCoordIndex(outIFS_->texCoordIndex_[(3*nLoop)+nIndex],nDifTexCoordIndex[nIndex]));
				}
				break;
			case 4:
				nFaceDirection =  acd.decode(mFD);
				nRotation =  acd.decode(mRotate);
				if(nFaceDirection == 1)
				{
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						tTexCoordIndex[nIndex] = outIFS_->texCoordIndex_[(3*nLoop)+2-nIndex];	
					}
					InverseRotation(tTexCoordIndex,nRotation);
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						outIFS_->texCoordIndex_[((nLoop+1)*3)+nIndex] = (tTexCoordIndex[nIndex]);
					}
				}
				else
				{
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						tTexCoordIndex[nIndex] = outIFS_->texCoordIndex_[(3*nLoop)+nIndex];
					}
					InverseRotation(tTexCoordIndex,nRotation);
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						outIFS_->texCoordIndex_[((nLoop+1)*3)+nIndex] = (tTexCoordIndex[nIndex]);
					}
				}
				break;

			}
		}

	}
}
void SVADecoder::InverseRotation(unsigned int *data, int size)
{
	int tData[3];
	int i;
	for(i=0; i<3; i++)
	{
		tData[i] = data[i];
	}
	for(i=0; i<3; i++)
	{
		data[i]= tData[(i-size+3)%3];
	}

}

void SVADecoder::DecoderConnectivity()
{
	if(entropytype == BP)
	{
		unsigned int nLoop, prefixSize, nBPL, nSign;
		int difValue;
		FDmode = (int)ReadBitsFromStorage(LEN_BIT);
		if(FDmode == 0)
		{
			FDvalue = (int)ReadBitsFromStorage(LEN_BIT);
		}
		prefixSize = ReadBitsFromStorage(LEN_PREFIX);
		unsigned int nType, nFaceDirection, nCoordIndex, nPosition, nRotation;
		int nDifCoordIndex[3];
		unsigned int tCount, nCount;
		unsigned int nIndex;
		unsigned int tCoordIndex[3];
		unsigned int tChange;

		for(nLoop = 0 ; nLoop < 3; nLoop++)
		{
			nBPL = ReadBitsFromStorage(prefixSize);
			nCoordIndex = BPLTable[nBPL];
			if(nBPL > 2)
			{
				nCoordIndex += ReadBitsFromStorage(nBPL - 1);
			}
			outIFS_->coordIndex_[nLoop] = (nCoordIndex);

		}
		
		if(FDmode == 0)
		{
			for(nLoop = 0 ; nLoop < (numberOfCoordIndex - 1); nLoop++)
			{
				for(tCount = 0; tCount < 5; tCount++)
				{
					if( ReadBitsFromStorage(1) == 1)
					{
						nType = tCount;
						break;
					}
				}
				switch(nType)
				{
				case 0:
					nPosition = ReadBitsFromStorage(1);
					if(nPosition == 1)
					{
						nPosition += ReadBitsFromStorage(1);
					}
					//nFaceDirection = ReadBitsFromStorage(1);
					nBPL = ReadBitsFromStorage(prefixSize);
					nDifCoordIndex[0] = BPLTable[nBPL];
					if(nBPL > 2)
					{
						nDifCoordIndex[0] += ReadBitsFromStorage(nBPL - 1);
					}
					nDifCoordIndex[0]+=1;
					nSign = ReadBitsFromStorage(1);
					if(nSign == 1)
					{
						nDifCoordIndex[0] = -(nDifCoordIndex[0]);
					}

					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}

					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						if(nIndex == nPosition)
						{
							tCoordIndex[nIndex] = InverseCircularDifferenceForConnectivity(outIFS_->coordIndex_[(3*nLoop)+nIndex],nDifCoordIndex[0]);
						}
						else
						{
							tCoordIndex[nIndex] = outIFS_->coordIndex_[(3*nLoop)+nIndex];
						}
					}
					//dmlab
					//tCoordIndex[(nPosition+1)%3] ^= tCoordIndex[(nPosition+2)%3] ^= tCoordIndex[(nPosition+1)%3] ^= tCoordIndex[(nPosition+2)%3];
					if(FDvalue == 1)
					{
						tChange = tCoordIndex[(nPosition+1)%3];
						tCoordIndex[(nPosition+1)%3] = tCoordIndex[(nPosition+2)%3] ;
						tCoordIndex[(nPosition+2)%3] = tChange;
					}
					InverseRotation(tCoordIndex,nRotation);

					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						//coordIndex.push_back(tCoordIndex[nIndex]);
						outIFS_->coordIndex_[((nLoop+1)*3)+nIndex] = tCoordIndex[nIndex];
					}
					break;

				case 1:
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifCoordIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifCoordIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nDifCoordIndex[nIndex]+=1;
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifCoordIndex[nIndex] = -(nDifCoordIndex[nIndex]);
						}
						//coordIndex.push_back(InverseCircularDifferenceForConnectivity(coordIndex[(3*nLoop)+nIndex],nDifCoordIndex[nIndex]));
						outIFS_->coordIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForConnectivity(outIFS_->coordIndex_[(3*nLoop)+nIndex],nDifCoordIndex[nIndex]));
					}

					break;
				case 2:
					nPosition = ReadBitsFromStorage(1);
					if(nPosition == 1)
					{
						nPosition += ReadBitsFromStorage(1);
					}
					for(nIndex = 0; nIndex < 2; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifCoordIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifCoordIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nDifCoordIndex[nIndex]+=1;
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifCoordIndex[nIndex] = -(nDifCoordIndex[nIndex]);
						}
					}
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}

					nCount = 0;
					tCoordIndex[nPosition] = outIFS_->coordIndex_[(3*nLoop)+nPosition];
					for(nIndex = 1; nIndex < 3; nIndex++)
					{
						tCoordIndex[((nPosition+nIndex)%3)] = InverseCircularDifferenceForConnectivity(outIFS_->coordIndex_[(3*nLoop)+((nPosition+nIndex)%3)],nDifCoordIndex[nCount++]);
					}

					InverseRotation(tCoordIndex,nRotation);
					
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						outIFS_->coordIndex_[((nLoop+1)*3)+nIndex] = (tCoordIndex[nIndex]);
					}
					break;
				case 3:
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifCoordIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifCoordIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifCoordIndex[nIndex] = -(nDifCoordIndex[nIndex]);
						}
						outIFS_->coordIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForConnectivity(outIFS_->coordIndex_[(3*nLoop)+nIndex],nDifCoordIndex[nIndex]));
					}
					break;
				case 4:
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}
					if(FDvalue == 1)
					{
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							tCoordIndex[nIndex] = outIFS_->coordIndex_[(3*nLoop)+2-nIndex];	
						}
						InverseRotation(tCoordIndex,nRotation);
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							outIFS_->coordIndex_[((nLoop+1)*3)+nIndex] = (tCoordIndex[nIndex]);
						}
					}
					else
					{
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							tCoordIndex[nIndex] = outIFS_->coordIndex_[(3*nLoop)+nIndex];
						}
						InverseRotation(tCoordIndex,nRotation);
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							outIFS_->coordIndex_[((nLoop+1)*3)+nIndex] = (tCoordIndex[nIndex]);
						}
					}
					break;
				}
			}
		}
		else
		{
			for(nLoop = 0 ; nLoop < (numberOfCoordIndex - 1); nLoop++)
			{
				for(tCount = 0; tCount < 4; tCount++)
				{
					if( ReadBitsFromStorage(1) == 1)
					{
						nType = tCount;
						break;
					}
				}
				switch(nType)
				{
				case 0:
					nPosition = ReadBitsFromStorage(1);
					if(nPosition == 1)
					{
						nPosition += ReadBitsFromStorage(1);
					}
					nFaceDirection = ReadBitsFromStorage(1);
					nBPL = ReadBitsFromStorage(prefixSize);
					nDifCoordIndex[0] = BPLTable[nBPL];
					if(nBPL > 2)
					{
						nDifCoordIndex[0] += ReadBitsFromStorage(nBPL - 1);
					}
					nDifCoordIndex[0]+=1;
					nSign = ReadBitsFromStorage(1);
					if(nSign == 1)
					{
						nDifCoordIndex[0] = -(nDifCoordIndex[0]);
					}
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}

					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						if(nIndex == nPosition)
						{
							tCoordIndex[nIndex] = InverseCircularDifferenceForConnectivity(outIFS_->coordIndex_[(3*nLoop)+nIndex],nDifCoordIndex[0]);
						}
						else
						{
							tCoordIndex[nIndex] = outIFS_->coordIndex_[(3*nLoop)+nIndex];
						}
					}
					//dmlab
					//tCoordIndex[(nPosition+1)%3] ^= tCoordIndex[(nPosition+2)%3] ^= tCoordIndex[(nPosition+1)%3] ^= tCoordIndex[(nPosition+2)%3];
					if(nFaceDirection == 1)
					{
						tChange = tCoordIndex[(nPosition+1)%3];
						tCoordIndex[(nPosition+1)%3] = tCoordIndex[(nPosition+2)%3] ;
						tCoordIndex[(nPosition+2)%3] = tChange;
					}
					InverseRotation(tCoordIndex,nRotation);
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						//coordIndex.push_back(tCoordIndex[nIndex]);
						outIFS_->coordIndex_[((nLoop+1)*3)+nIndex] = tCoordIndex[nIndex];
					}
					break;

				case 1:
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifCoordIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifCoordIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nDifCoordIndex[nIndex]+=1;
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifCoordIndex[nIndex] = -(nDifCoordIndex[nIndex]);
						}
						//coordIndex.push_back(InverseCircularDifferenceForConnectivity(coordIndex[(3*nLoop)+nIndex],nDifCoordIndex[nIndex]));
						outIFS_->coordIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForConnectivity(outIFS_->coordIndex_[(3*nLoop)+nIndex],nDifCoordIndex[nIndex]));
					}

					break;
				case 2:
					nPosition = ReadBitsFromStorage(1);
					if(nPosition == 1)
					{
						nPosition += ReadBitsFromStorage(1);
					}
					for(nIndex = 0; nIndex < 2; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifCoordIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifCoordIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nDifCoordIndex[nIndex]+=1;
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifCoordIndex[nIndex] = -(nDifCoordIndex[nIndex]);
						}
					}
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}
					nCount = 0;
					tCoordIndex[nPosition] = outIFS_->coordIndex_[(3*nLoop)+nPosition];
					for(nIndex = 1; nIndex < 3; nIndex++)
					{
						tCoordIndex[((nPosition+nIndex)%3)] = InverseCircularDifferenceForConnectivity(outIFS_->coordIndex_[(3*nLoop)+((nPosition+nIndex)%3)],nDifCoordIndex[nCount++]);
					}
					InverseRotation(tCoordIndex,nRotation);
					
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						outIFS_->coordIndex_[((nLoop+1)*3)+nIndex] = (tCoordIndex[nIndex]);
					}
					break;
				case 3:
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						nBPL = ReadBitsFromStorage(prefixSize);
						nDifCoordIndex[nIndex] = BPLTable[nBPL];
						if(nBPL > 2)
						{
							nDifCoordIndex[nIndex] += ReadBitsFromStorage(nBPL - 1);
						}
						nSign = ReadBitsFromStorage(1);
						if(nSign == 1)
						{
							nDifCoordIndex[nIndex] = -(nDifCoordIndex[nIndex]);
						}
						outIFS_->coordIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForConnectivity(outIFS_->coordIndex_[(3*nLoop)+nIndex],nDifCoordIndex[nIndex]));
					}
					break;
				case 4:
					nFaceDirection = ReadBitsFromStorage(1);
					nRotation = ReadBitsFromStorage(1);
					if(nRotation == 1)
					{
						nRotation += ReadBitsFromStorage(1);
					}
					if(nFaceDirection == 1)
					{
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							tCoordIndex[nIndex] = outIFS_->coordIndex_[(3*nLoop)+2-nIndex];	
						}
						InverseRotation(tCoordIndex,nRotation);
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							outIFS_->coordIndex_[((nLoop+1)*3)+nIndex] = (tCoordIndex[nIndex]);
						}
					}
					else
					{
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							tCoordIndex[nIndex] = outIFS_->coordIndex_[(3*nLoop)+nIndex];
						}
						InverseRotation(tCoordIndex,nRotation);
						for(nIndex = 0; nIndex < 3; nIndex++)
						{
							outIFS_->coordIndex_[((nLoop+1)*3)+nIndex] = (tCoordIndex[nIndex]);
						}
					}
					break;

				}
			}

		}
	}
	else
	{
		unsigned int nLoop, nSign;
		int difValue;
		unsigned int nType, nFaceDirection, nCoordIndex, nPosition, nRotation;
		int nDifCoordIndex[3];
		unsigned int tCount, nCount;
		unsigned int nIndex;
		unsigned int tCoordIndex[3];
		unsigned int tChange;

		unsigned int code_bytes = 0;
		ReadFromStream(m_pReadStream, &code_bytes, sizeof(unsigned int));
		unsigned char * code_buffer = new unsigned char [code_bytes];                                                   
		ReadFromStream(m_pReadStream, code_buffer, sizeof(unsigned char)*code_bytes);
		
		Adaptive_Data_Model mModel(1024);
		Adaptive_Data_Model mhasnext (2);
		Adaptive_Data_Model mSign (2);
		Adaptive_Data_Model mType (5);
		Adaptive_Data_Model mFD (2);
		Adaptive_Data_Model mPos (3);
		Adaptive_Data_Model mRotate (3);
		Arithmetic_Codec acd(code_bytes, code_buffer);
		acd.start_decoder();	// initialize decoder

		int data[4];
		data[0] = 1;
		data[1] = 1<<10;
		data[2] = 1<<20;
		data[3] = 1<<30;
		int count = 0;

		for(nLoop = 0 ; nLoop < 3; nLoop++)
		{
			
			difValue = 0;
			while(1)
			{
				difValue += data[count] * acd.decode(mModel);
				count++;
				int temp = acd.decode(mhasnext);

				if( temp ==0)
					break;
			}
			count = 0;
			outIFS_->coordIndex_[nLoop] = (difValue);

		}
		for(nLoop = 0 ; nLoop < (numberOfCoordIndex - 1); nLoop++)
		{
			nType = acd.decode(mType);
			
			switch(nType)
			{
			case 0:
				nPosition = acd.decode(mPos);
				nFaceDirection =acd.decode(mFD);
				
				difValue = 0;
				while(1)
				{
					difValue += data[count] * acd.decode(mModel);
					count++;
					int temp = acd.decode(mhasnext);

					if( temp ==0)
						break;
				}
				difValue +=1;
				count = 0;
				nSign = acd.decode(mSign);
				if(nSign == 1)
				{
					difValue = -(difValue);
				}

				nRotation = acd.decode(mRotate);
				nDifCoordIndex[0] = difValue;
			
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					if(nIndex == nPosition)
					{
						tCoordIndex[nIndex] = InverseCircularDifferenceForConnectivity(outIFS_->coordIndex_[(3*nLoop)+nIndex],nDifCoordIndex[0]);
					}
					else
					{
						tCoordIndex[nIndex] = outIFS_->coordIndex_[(3*nLoop)+nIndex];
					}
				}
				//dmlab
				//tCoordIndex[(nPosition+1)%3] ^= tCoordIndex[(nPosition+2)%3] ^= tCoordIndex[(nPosition+1)%3] ^= tCoordIndex[(nPosition+2)%3];
				if(nFaceDirection == 1)
				{
					tChange = tCoordIndex[(nPosition+1)%3];
					tCoordIndex[(nPosition+1)%3] = tCoordIndex[(nPosition+2)%3] ;
					tCoordIndex[(nPosition+2)%3] = tChange;
				}
				InverseRotation(tCoordIndex,nRotation);
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					//coordIndex.push_back(tCoordIndex[nIndex]);
					outIFS_->coordIndex_[((nLoop+1)*3)+nIndex] = tCoordIndex[nIndex];
				}
				break;

			case 1:
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					difValue = 0;
					while(1)
					{
						difValue += data[count] * acd.decode(mModel);
						count++;
						int temp = acd.decode(mhasnext);

						if( temp ==0)
							break;
					}
					difValue +=1;
					count = 0;
					nSign = acd.decode(mSign);
					if(nSign == 1)
					{
						difValue = -(difValue);
					}
					nDifCoordIndex[nIndex] = difValue;
					//coordIndex.push_back(InverseCircularDifferenceForConnectivity(coordIndex[(3*nLoop)+nIndex],nDifCoordIndex[nIndex]));
					outIFS_->coordIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForConnectivity(outIFS_->coordIndex_[(3*nLoop)+nIndex],nDifCoordIndex[nIndex]));
				}

				break;
			case 2:
				nPosition = acd.decode(mPos);
				for(nIndex = 0; nIndex < 2; nIndex++)
				{
					difValue = 0;
					while(1)
					{
						difValue += data[count] * acd.decode(mModel);
						count++;
						int temp = acd.decode(mhasnext);

						if( temp ==0)
							break;
					}
					count = 0;
					difValue +=1;
					nSign = acd.decode(mSign);
					if(nSign == 1)
					{
						difValue = -(difValue);
					}
					nDifCoordIndex[nIndex] = difValue;
					
				}
				nRotation = acd.decode(mRotate);
				nCount = 0;
				tCoordIndex[nPosition] = outIFS_->coordIndex_[(3*nLoop)+nPosition];
				for(nIndex = 1; nIndex < 3; nIndex++)
				{
					tCoordIndex[((nPosition+nIndex)%3)] = InverseCircularDifferenceForConnectivity(outIFS_->coordIndex_[(3*nLoop)+((nPosition+nIndex)%3)],nDifCoordIndex[nCount++]);
				}
				InverseRotation(tCoordIndex,nRotation);
				
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					outIFS_->coordIndex_[((nLoop+1)*3)+nIndex] = (tCoordIndex[nIndex]);
				}
				break;
			case 3:
				for(nIndex = 0; nIndex < 3; nIndex++)
				{
					difValue = 0;
					while(1)
					{
						difValue += data[count] * acd.decode(mModel);
						count++;
						int temp = acd.decode(mhasnext);

						if( temp ==0)
							break;
					}
					count = 0;
					nSign = acd.decode(mSign);
					if(nSign == 1)
					{
						difValue = -(difValue);
					}
					nDifCoordIndex[nIndex] = difValue;

					outIFS_->coordIndex_[((nLoop+1)*3)+nIndex] = (InverseCircularDifferenceForConnectivity(outIFS_->coordIndex_[(3*nLoop)+nIndex],nDifCoordIndex[nIndex]));
				}
				break;
			case 4:
				nFaceDirection =  acd.decode(mFD);
				nRotation =  acd.decode(mRotate);
				if(nFaceDirection == 1)
				{
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						tCoordIndex[nIndex] = outIFS_->coordIndex_[(3*nLoop)+2-nIndex];	
					}
					InverseRotation(tCoordIndex,nRotation);
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						outIFS_->coordIndex_[((nLoop+1)*3)+nIndex] = (tCoordIndex[nIndex]);
					}
				}
				else
				{
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						tCoordIndex[nIndex] = outIFS_->coordIndex_[(3*nLoop)+nIndex];
					}
					InverseRotation(tCoordIndex,nRotation);
					for(nIndex = 0; nIndex < 3; nIndex++)
					{
						outIFS_->coordIndex_[((nLoop+1)*3)+nIndex] = (tCoordIndex[nIndex]);
					}
				}
				break;

			}
		}

	}

}



unsigned int SVADecoder::ReadBitsFromStorage(unsigned int unReadLength)
{
	unsigned int n = unReadLength;
	int c=(n + 14 - unCurrentBitPosition) >> 3;

	union { 
		long long l;
		unsigned char c[8];
	} buf;
	buf.l = 0;

	int i=0;
	while(n) {
		i++;
		buf.c[c - i] = *(m_pReadStream + unCurrentBufferPosition);
		if(unCurrentBitPosition >= n) {
			unCurrentBitPosition -= n;
			break;
		} else {
			n -= unCurrentBitPosition + 1;
			unCurrentBitPosition = 7;
			unCurrentBufferPosition++;
		}
	}

	return (buf.l >> ((unCurrentBitPosition + 1) % 8)) & ~(~static_cast<unsigned int>(0) << unReadLength);
}

void SVADecoder::BPLInit()
{
	int i;
	BPLTable = new unsigned int [32];
	BPLTable[0] = 0;
	BPLTable[1] = 1;
	BPLTable[2] = 2;

	for(i = 3; i <LEN_BPLMAX; i++)
	{
		BPLTable[i] = pow(2.0,(double)(i-1))-1;
	}

}

