//-------------------------------------------------------------------------
// 
// Copyright ETRI, Hanyang University 2008, 2009
// Licensed Materials, Program Property of ETRI & Hanyang University
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.
// ETRI and Hanyang University retain full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//    AUTHORS:
//    Seung Wook Lee (ETRI)
//    Daiyong Kim (Hanyang University)
//    KyoungSoo Son (Hanyang University)
//
//    http://www.etri.re.kr/
//    http://dmlab.hanyang.ac.kr/
//  
//    Seung Wook Lee     (tajinet@etri.re.kr)  
//    Daiyong Kim        (dykim@dmlab.hanyang.ac.kr)    
//    KyoungSoo Son      (bonokensin@dmlab.hanyang.ac.kr)
//
//-------------------------------------------------------------------------



#include <iostream>
#include <time.h>
#include <string>
#include <iostream>
#include <fstream>
#include <sstream> 
#include <math.h>
#include <vector>
#include <map>
using namespace std;

#include <assert.h>

#include "IndexedFaceSet.h"
#include "SVA_Encoder.h"
#include "utils.h"
#include "arithmetic_codec.h"

#define MAX_VLAUE 1000000
#define MIN_VLAUE -1000000
#define INIT_UCHAR 0x00
#define LEN_BYTE	8
#define LEN_QP		5
#define LEN_PREFIX	5
#define LEN_BIT		1
#define DEFAULT		0
#define DIFF		1
#define XOR			2
#define BP			1
#define AC			2


#define SC3DMCStartCode 0x00001F1


SVAEncoder::SVAEncoder()
{
	dumpBit =INIT_UCHAR;
	bitPostion = 0;
	MaxDifferentialForGeometry = 0;
	MaxDifferentialForConnectivity = 0;
	BPLTable = NULL;
	payloadTable = NULL;

	
}

SVAEncoder::~SVAEncoder()
{
	if(BPLTable != NULL)
		delete [] BPLTable;
	if(payloadTable  != NULL)
		delete [] payloadTable;
	
}

void SVAEncoder::EncodeDataBuffer(unsigned char *compressedStream, unsigned int &compressedStreamSize)
{
	Initialize();
	QuantizationForGeometry(QPforGeomerty);
	ConnectivityAnalyze();
	if(nColor != 0)
		QuantizationColor(QPforColor);		
	if(nTexCoord != 0)
		QuantizationTexCoord(u, v);

	if(entropytype == BP)
		BPCInit();


	m_compressedStream = compressedStream; 
	m_compressedStreamSize = compressedStreamSize;

	
	Encode_Header();
	Encode_Geometry();
	Encode_Connectivity();
	
	Encode_Normal();						
	Encode_NormalIndex();					
	
	Encode_TexCoord();						
	Encode_TexCoordIndex();					
	Encode_Color();						
	Encode_ColorIndex();

	if(entropytype == BP)
		Flush();

	compressedStreamSize = m_compressedStreamSize;
}

void SVAEncoder::VectorToBuffer(unsigned char *compressedStream, unsigned int &compressedStreamSize)
{
	for(int i=0; i< dumpBitstream.size(); i++)
	{
		compressedStream[compressedStreamSize++] = dumpBitstream[i];
	}
}

void SVAEncoder::ParsingConfigure(char *config)
{
	char* tgbuf = NULL;
	int count = 0;
	int tempQP[4] = {0};
	while(config != NULL)
	{
		tgbuf = strtok(config, " ");
		tempQP[count++] = atoi(tgbuf);
		if(count >= 4)
			break;
	}
	QPforGeomerty = tempQP[0];
	QPforNormal = tempQP[1];
	QPforColor = tempQP[2];
	QPforTexCoord = tempQP[3];
}

void SVAEncoder::Initialize()
{

	codingtype = DIFF;
	entropytype = BP;
	
	dumpBit =INIT_UCHAR;
	bitPostion = 0;
	MaxDifferentialForGeometry = MIN_VLAUE;
	MaxDifferentialForConnectivity = MIN_VLAUE;
	BPLTable = NULL;
	payloadTable = NULL;
	
	nOtherAtrribute = 0;
	nOtherAtrributeIndex = 0;

	
	nCoord = m_ifs->GetNCoord();
	nCoordIndex = m_ifs->GetNCoordIndex();
	nNormal = m_ifs->GetNNormal();
	nNormalIndex = m_ifs->GetNNormalIndex();
	nColor = m_ifs->GetNColor();
	nColorIndex = m_ifs->GetNColorIndex();
	nTexCoord = m_ifs->GetNTexCoord();
	nTexCoordIndex = m_ifs->GetNTexCoordIndex();

	Coord = m_ifs->GetCoord();
	CoordIndex = m_ifs->GetCoordIndex();
	Normal = m_ifs->GetNormal();								
	NormalIndex = m_ifs->GetNormalIndex();
	TexCoord = m_ifs->GetTexCoord();					
	TexCoordIndex = m_ifs->GetTexCoordIndex();					
	Color = m_ifs->GetColor();
	ColorIndex = m_ifs->GetColorIndex();

	
	for(int i = 0; i < 3; i++)
	{
		
		quantMinGeometry[i] = m_ifs->minCoord[i];
		quantMinNormal[i] = m_ifs->minNormal[i];
		quantMinColor[i] = m_ifs->minColor[i];	
		quantMinOtherAttribute[i]= 0.0;
		quantRangeGeometry[i] = m_ifs->rangeCoord[i];
		quantRangeNormal[i] = m_ifs->rangeNormal[i];
		quantRangeColor[i] = m_ifs->rangeColor[i];
	
	}
	for(int i = 0; i < 2; i++)
	{
		quantMinTexCoord[i] = m_ifs->minTexCoord[i];
		quantRangeTexCoord[i] = m_ifs->rangeTexCoord[i];	
	}


					
	quantRangeOtherAttribute = 0;	

	QPforGeomerty = m_params.m_nQCoordBits;
	QPforNormal = m_params.m_nQNormalBits;
	QPforColor = m_params.m_nQColorBits;
	QPforTexCoord = m_params.m_nQTexCoordBits;

}

int SVAEncoder::Encode_SVA(IndexedFaceSet &myIFS, EncodingParams &m_params)
{


	// Quantization Part;

	Initialize();
	QuantizationForGeometry(QPforGeomerty);
	ConnectivityAnalyze();
	BPCInit();

	Encode_CommonHeader(myIFS,m_params);
	Encode_Header();
	Encode_Geometry();
	Encode_Connectivity();
	Flush();


	return 1;
}
int SVAEncoder::Rotation(int* data, int move)
{
	move=move%3;
	if(move<0)
		move+=3;
	int i;
	int tempStorage[5];
	for(i = 0 ; i < 5 ; i++)
	{
		tempStorage [i] = data[(i%3)];
	}
	for(i = 0 ; i < 3 ; i++)
	{
		data [i] = tempStorage[i+move];
	}

	return move;

}

void SVAEncoder::InverseRotation(int *data, int size)
{
	int tData[3];
	int i;
	for(i=0; i<3; i++)
	{
		tData[i] = data[i];
	}
	for(i=0; i<3; i++)
	{
		data[(i+size)%3]= tData[i];
	}

}

void SVAEncoder::BPCInit()
{

	int maxValue = 0;
	if(nCoord > 0)
	{
		maxValue = (maxValue > (1<<QPforGeomerty) ) ? maxValue : (1<<QPforGeomerty);
	}
	if(nNormal > 0)
	{
		maxValue = (maxValue > (1<<QPforNormal) ) ? maxValue : (1<<QPforNormal);
	}
	if(nColor > 0)
	{
		maxValue = (maxValue > (1<<QPforColor) ) ? maxValue : (1<<QPforColor);
	}
	if(nTexCoord > 0)
	{
		maxValue = (maxValue > (1<<QPforTexCoord) ) ? maxValue : (1<<QPforTexCoord);
	}
	if(nCoordIndex > 0)
	{
		maxValue = (maxValue > (nCoord<<1) ) ? maxValue : (nCoord<<1);
	}
	if(nNormalIndex > 0)
	{
		maxValue = (maxValue > (nNormal<<1) ) ? maxValue : (nNormal<<1);
	}
	if(nColorIndex > 0)
	{
		maxValue = (maxValue > (nColor<<1) ) ? maxValue : (nColor<<1);
	}
	if(nTexCoordIndex > 0)
	{
		maxValue = (maxValue > (nTexCoord<<1) ) ? maxValue : (nTexCoord<<1);
	}
	BPLTable = new int [maxValue];
	payloadTable = new int [maxValue];
	int i;
	for(i = 0 ; i < 3; i++)
	{
		BPLTable[i] = i;
		payloadTable[i] = -1;
	}
	int payloadLength = 4;
	int minValueInPresentBPL = 3;
	int presentBPL = 3;
	for( ; i<maxValue; i++)
	{
		if(i == payloadLength+minValueInPresentBPL)
		{
			minValueInPresentBPL += payloadLength;
			payloadLength *= 2;
			presentBPL++;
		}
		BPLTable[i] = presentBPL;
		payloadTable[i] = i - minValueInPresentBPL;
	}

}

void SVAEncoder::ConnectivityAnalyze()
{

	int symbol;
	int flip=0;
	int k, l, j;
	int DifferetialCI[2];
	int rotcount=0;
	int nRot;


	difCoordIndex.push_back(CoordIndex[0]); // first face 
	if(abs(CoordIndex[0]) > MaxDifferentialForConnectivity)
	{
		MaxDifferentialForConnectivity = CoordIndex[0];
	}
	difCoordIndex.push_back(CoordIndex[1]);
	if(abs(CoordIndex[1]) > MaxDifferentialForConnectivity)
	{
		MaxDifferentialForConnectivity = CoordIndex[1];
	}
	difCoordIndex.push_back(CoordIndex[2]); 
	if(abs(CoordIndex[2]) > MaxDifferentialForConnectivity)
	{
		MaxDifferentialForConnectivity = CoordIndex[2];
	}

	//From second face
	for(j = 1; j < nCoordIndex; j++)
	{

		//Count the overlapped points
		int sizeofOverlapped=0;
		int myType=0;
		int cheecktype4 =0;

		if(cheecktype4 == 1)
		{
			cheecktype4 =0;
			sizeofOverlapped = 4;

		}
		else 
		{
			for(k=0;k<3;k++)
			{
				for(l=k+1;l<3;l++)
				{
					if(CoordIndex[j*3+k] == CoordIndex[j*3+l])
					{
						cheecktype4 == 1;
						sizeofOverlapped = 4;
						break;
					}
				}
			}
			if(sizeofOverlapped != 4)
			{
				for(k=0;k<3;k++)
				{
					for(l=0;l<3;l++)
					{
						if(CoordIndex[j*3+k] == CoordIndex[(j-1)*3+l])
						{
							sizeofOverlapped++;
							break;
						}
					}
				}
			}
		}

		// Changing type information based on its frequency.
		// Type 2 ==> Type 0
		// Type 0 ==> Type 1
		// Type 1 ==> Type 2
		// Type 3 ==> Type 3

		

		
		switch(sizeofOverlapped)
		{
		case 0: //Type C encode 3 symbols
			myType = 1;
			
			symbol = CircularDifferenceForConnectivity(CoordIndex[(j-1)*3], CoordIndex[j*3]);
			difCoordIndex.push_back(symbol);
			if(abs(symbol)-1 > MaxDifferentialForConnectivity)
			{
				MaxDifferentialForConnectivity = abs(symbol)-1;
			}

			symbol = CircularDifferenceForConnectivity(CoordIndex[(j-1)*3+1], CoordIndex[j*3+1]);
			difCoordIndex.push_back(symbol);
			if(abs(symbol)-1 > MaxDifferentialForConnectivity)
			{
				MaxDifferentialForConnectivity = abs(symbol)-1;
			}
			
			symbol = CircularDifferenceForConnectivity(CoordIndex[(j-1)*3+2], CoordIndex[j*3+2]);
			difCoordIndex.push_back(symbol);
			if(abs(symbol)-1 > MaxDifferentialForConnectivity)
			{
				MaxDifferentialForConnectivity = abs(symbol)-1;
			}

			break;
		case 1: //Type B encode 2 symbols 
			myType = 2;

			for(k=0;k<3;k++)
			{
				for(l=0;l<3;l++)
				{
					if(CoordIndex[j*3+k] == CoordIndex[(j-1)*3+l])						
					{
						//Rotation
						nRot = Rotation((CoordIndex+j*3), k-l);
						rotate.push_back(nRot);
						
						


						//encode point that indicates same symbol						;
						position.push_back(l);

						//encode 2 symbols						
						symbol = CircularDifferenceForConnectivity(CoordIndex[(j-1)*3+(l+1)%3], CoordIndex[j*3+(l+1)%3]);
						difCoordIndex.push_back(symbol);
						if(abs(symbol)-1 > MaxDifferentialForConnectivity)
						{
							MaxDifferentialForConnectivity = abs(symbol)-1;
						}
						symbol = CircularDifferenceForConnectivity(CoordIndex[(j-1)*3+(l+2)%3], CoordIndex[j*3+(l+2)%3]);
						difCoordIndex.push_back(symbol);
						if(abs(symbol)-1 > MaxDifferentialForConnectivity)
						{
							MaxDifferentialForConnectivity = abs(symbol)-1;
						}

						k=3;
						InverseRotation((CoordIndex+j*3), nRot);
						break;
					}
				}
			}

			break;



		case 2: //Type A encode 1 symbol
			//DifferetialCI[0] indicates different symbol's position in the current face
			myType = 0;
			for(k=0;k<3;k++)
			{
				for(l=0;l<3;l++)
					if(CoordIndex[j*3+k] == CoordIndex[(j-1)*3+l])
						break;
				if(l==3)
					DifferetialCI[0]=k;
			}
			//DifferetialCI[1] indicates different symbol's position in the current face
			for(k=0;k<3;k++)
			{
				for(l=0;l<3;l++)
					if(CoordIndex[j*3+l] == CoordIndex[(j-1)*3+k])
						break;
				if(l==3)
					DifferetialCI[1]=k;
			}
			//rotate to match DifferetialCI[0] and DifferetialCI[1]
			nRot = Rotation((CoordIndex+j*3), DifferetialCI[0]-DifferetialCI[1]);
			rotate.push_back(nRot);
		

			flip=1;
			
			for(k=0; k <3; k++)
			{
				if(CoordIndex[j*3+k] == CoordIndex[(j-1)*3+k])
				{
					flip=0;
					break;
				}
			}
			//coding flip flag
			faceDirection.push_back(flip);

			//encode pos that indicates different symbol
			position.push_back(DifferetialCI[1]);


			//encode 1 symbol
			symbol = CircularDifferenceForConnectivity(CoordIndex[(j-1)*3+DifferetialCI[1]],CoordIndex[j*3+DifferetialCI[1]]);
			difCoordIndex.push_back(symbol);
			if(abs(symbol)-1 > MaxDifferentialForConnectivity)
			{
				MaxDifferentialForConnectivity = abs(symbol)-1;
			}
			InverseRotation((CoordIndex+j*3), nRot);
			
			break;

		
		case 3: //Type D encode only flip bit
			myType = 4;
			rotcount=0;
			while(CoordIndex[j*3+1] != CoordIndex[(j-1)*3+1])
			{
				Rotation((CoordIndex+j*3), 1);
				rotcount++;
			}
			rotate.push_back(rotcount);
			
			

			if(	CoordIndex[j*3] == CoordIndex[(j-1)*3] )
				flip=0;
			else
				flip=1;
			faceDirection.push_back(flip);
			InverseRotation((CoordIndex+j*3), rotcount);
			//coding flip flag			
			

			break;
		case 4:
			myType = 3;
			symbol = CircularDifferenceForConnectivity(CoordIndex[(j-1)*3], CoordIndex[j*3]);
			difCoordIndex.push_back(symbol);
			if(abs(symbol) > MaxDifferentialForConnectivity)
			{
				MaxDifferentialForConnectivity = abs(symbol);
			}

			symbol = CircularDifferenceForConnectivity(CoordIndex[(j-1)*3+1], CoordIndex[j*3+1]);
			difCoordIndex.push_back(symbol);
			if(abs(symbol) > MaxDifferentialForConnectivity)
			{
				MaxDifferentialForConnectivity = abs(symbol);
			}
			
			symbol = CircularDifferenceForConnectivity(CoordIndex[(j-1)*3+2], CoordIndex[j*3+2]);
			difCoordIndex.push_back(symbol);
			if(abs(symbol) > MaxDifferentialForConnectivity)
			{
				MaxDifferentialForConnectivity = abs(symbol);
			}

			break;

		}
		type.push_back(myType); //type information

	}

	//change
	if(faceDirection.size() == 0 )
	{
		FDmode = 1;
	}
	else
	{
		FDvalue = faceDirection[0];
		FDmode = 0;
		for(j = 1 ; j< faceDirection.size(); j++)
		{
			if(FDvalue != faceDirection[j])
			{
				FDmode = 1;
				break;
			}

		}

	}


}

void SVAEncoder::Analyze_NormalIndex()
{
	
	int symbol;
	int flip=0;
	int k, l, j;
	int DifferetialCI[2];
	int rotcount=0;
	int nRot;

	type.clear();
	difIndex.clear(); 
	position.clear();
	faceDirection.clear();
	rotate.clear();;

	MaxDifferentialForNormalIndex = 0;


	difIndex.push_back(NormalIndex[0]); // first face 
	if(abs(CoordIndex[0]) > MaxDifferentialForNormalIndex)
	{
		MaxDifferentialForNormalIndex = NormalIndex[0];
	}
	difIndex.push_back(NormalIndex[1]);
	if(abs(NormalIndex[1]) > MaxDifferentialForNormalIndex)
	{
		MaxDifferentialForNormalIndex = NormalIndex[1];
	}
	difIndex.push_back(NormalIndex[2]); 
	if(abs(NormalIndex[2]) > MaxDifferentialForNormalIndex)
	{
		MaxDifferentialForNormalIndex = NormalIndex[2];
	}

	//From second face
	for(j = 1; j < nNormalIndex; j++)
	{

		//Count the overlapped points
		int sizeofOverlapped=0;
		int myType=0;
		int cheecktype4 =0;

		if(cheecktype4 == 1)
		{
			cheecktype4 =0;
			sizeofOverlapped = 4;

		}
		else 
		{
			for(k=0;k<3;k++)
			{
				for(l=k+1;l<3;l++)
				{
					if(NormalIndex[j*3+k] == NormalIndex[j*3+l])
					{
						cheecktype4 == 1;
						sizeofOverlapped = 4;
						break;
					}
				}
			}
			if(sizeofOverlapped != 4)
			{
				for(k=0;k<3;k++)
				{
					for(l=0;l<3;l++)
					{
						if(NormalIndex[j*3+k] == NormalIndex[(j-1)*3+l])
						{
							sizeofOverlapped++;
							break;
						}
					}
				}
			}
		}

		// Changing type information based on its frequency.
		// Type 2 ==> Type 0
		// Type 0 ==> Type 1
		// Type 1 ==> Type 2
		// Type 3 ==> Type 3

		

		
		switch(sizeofOverlapped)
		{
		case 0: //Type C encode 3 symbols
			myType = 1;
			
			symbol = CircularDifferenceForNormalIndex(NormalIndex[(j-1)*3], NormalIndex[j*3]);
			difIndex.push_back(symbol);
			if(abs(symbol)-1 > MaxDifferentialForNormalIndex)
			{
				MaxDifferentialForNormalIndex = abs(symbol)-1;
			}

			symbol = CircularDifferenceForNormalIndex(NormalIndex[(j-1)*3+1], NormalIndex[j*3+1]);
			difIndex.push_back(symbol);
			if(abs(symbol)-1 > MaxDifferentialForNormalIndex)
			{
				MaxDifferentialForNormalIndex = abs(symbol)-1;
			}
			
			symbol = CircularDifferenceForNormalIndex(NormalIndex[(j-1)*3+2], NormalIndex[j*3+2]);
			difIndex.push_back(symbol);
			if(abs(symbol)-1 > MaxDifferentialForNormalIndex)
			{
				MaxDifferentialForNormalIndex = abs(symbol)-1;
			}

			break;
		case 1: //Type B encode 2 symbols 
			myType = 2;

			for(k=0;k<3;k++)
			{
				for(l=0;l<3;l++)
					if(NormalIndex[j*3+k] == NormalIndex[(j-1)*3+l])						
					{
						//Rotation
						nRot = Rotation((NormalIndex+j*3), k-l);
						rotate.push_back(Rotation((NormalIndex+j*3), k-l));
					


						//encode point that indicates same symbol						;
						position.push_back(l);

						//encode 2 symbols						
						symbol = CircularDifferenceForNormalIndex(NormalIndex[(j-1)*3+(l+1)%3], NormalIndex[j*3+(l+1)%3]);
						difIndex.push_back(symbol);
						if(abs(symbol)-1 > MaxDifferentialForNormalIndex)
						{
							MaxDifferentialForNormalIndex = abs(symbol)-1;
						}
						symbol = CircularDifferenceForNormalIndex(NormalIndex[(j-1)*3+(l+2)%3], NormalIndex[j*3+(l+2)%3]);
						difIndex.push_back(symbol);
						if(abs(symbol)-1 > MaxDifferentialForNormalIndex)
						{
							MaxDifferentialForNormalIndex = abs(symbol)-1;
						}

						k=3;
						InverseRotation((NormalIndex+j*3),nRot);
						break;
					}
			}

			break;



		case 2: //Type A encode 1 symbol
			//DifferetialCI[0] indicates different symbol's position in the current face
			myType = 0;
			for(k=0;k<3;k++)
			{
				for(l=0;l<3;l++)
					if(NormalIndex[j*3+k] == NormalIndex[(j-1)*3+l])
						break;
				if(l==3)
					DifferetialCI[0]=k;
			}
			//DifferetialCI[1] indicates different symbol's position in the current face
			for(k=0;k<3;k++)
			{
				for(l=0;l<3;l++)
					if(NormalIndex[j*3+l] == NormalIndex[(j-1)*3+k])
						break;
				if(l==3)
					DifferetialCI[1]=k;
			}
			//rotate to match DifferetialCI[0] and DifferetialCI[1]

			nRot = Rotation((NormalIndex+j*3), DifferetialCI[0]-DifferetialCI[1]);
			rotate.push_back(nRot);
		


			flip=1;
			
			for(k=0; k <3; k++)
			{
				if(NormalIndex[j*3+k] == NormalIndex[(j-1)*3+k])
				{
					flip=0;
					break;
				}
			}
			//coding flip flag
			faceDirection.push_back(flip);

			//encode pos that indicates different symbol
			position.push_back(DifferetialCI[1]);


			//encode 1 symbol
			symbol = CircularDifferenceForNormalIndex(NormalIndex[(j-1)*3+DifferetialCI[1]],NormalIndex[j*3+DifferetialCI[1]]);
			difIndex.push_back(symbol);
			if(abs(symbol)-1 > MaxDifferentialForNormalIndex)
			{
				MaxDifferentialForNormalIndex = abs(symbol)-1;
			}
			InverseRotation((NormalIndex+j*3),nRot);
			
			break;

		
		case 3: //Type D encode only flip bit
			myType = 4;
			rotcount=0;
			while(NormalIndex[j*3+1] != NormalIndex[(j-1)*3+1])
			{
				Rotation((NormalIndex+j*3), 1);
				rotcount++;
			}
			rotate.push_back(rotcount);
			

			if(	NormalIndex[j*3] == NormalIndex[(j-1)*3] )
				flip=0;
			else
				flip=1;
			faceDirection.push_back(flip);
			InverseRotation((NormalIndex+j*3),rotcount);
			//coding flip flag			
			

			break;
		case 4:
			myType = 3;
			symbol = CircularDifferenceForNormalIndex(NormalIndex[(j-1)*3], NormalIndex[j*3]);
			difIndex.push_back(symbol);
			if(abs(symbol) > MaxDifferentialForNormalIndex)
			{
				MaxDifferentialForNormalIndex = abs(symbol);
			}

			symbol = CircularDifferenceForNormalIndex(NormalIndex[(j-1)*3+1], NormalIndex[j*3+1]);
			difIndex.push_back(symbol);
			if(abs(symbol) > MaxDifferentialForNormalIndex)
			{
				MaxDifferentialForNormalIndex = abs(symbol);
			}
			
			symbol = CircularDifferenceForNormalIndex(NormalIndex[(j-1)*3+2], NormalIndex[j*3+2]);
			difIndex.push_back(symbol);
			if(abs(symbol) > MaxDifferentialForNormalIndex)
			{
				MaxDifferentialForNormalIndex = abs(symbol);
			}

			break;

		}
		type.push_back(myType); //type information

	}

	//change
	if(faceDirection.size() == 0 )
	{
		FDmode = 1;
	}
	else
	{
		FDvalue = faceDirection[0];
		FDmode = 0;
		for(j = 1 ; j< faceDirection.size(); j++)
		{
			if(FDvalue != faceDirection[j])
			{
				FDmode = 1;
				break;
			}

		}
	}


}


void SVAEncoder:: Encode_CommonHeader(IndexedFaceSet &m_ifs, EncodingParams &m_params)
{

}


void SVAEncoder::Encode_Header()
{
	Write2Stream(m_compressedStream, m_compressedStreamSize, &codingtype, sizeof(unsigned char));
	Write2Stream(m_compressedStream, m_compressedStreamSize, &entropytype, sizeof(unsigned char));	

}

int SVAEncoder::CircularDifferenceForConnectivity(int prev, int cur)
{
	int data;
	if(cur > prev)
	{
		data = (prev - 0 )+(nCoord - cur);

	}
	else
	{
		data = (cur - 0 )+(nCoord - prev);
	}

	if(abs(cur - prev) <= data)
		return cur-prev;
	else if(cur < prev)
	{
		return (data);
	}
	else
	{
		return (-data);
	}
}

int SVAEncoder::CircularDifferenceForNormalIndex(int prev, int cur)
{
	int data;
	if(cur > prev)
	{
		data = (prev - 0 )+(nNormal - cur);

	}
	else
	{
		data = (cur - 0 )+(nNormal - prev);
	}

	if(abs(cur - prev) <= data)
		return cur-prev;
	else if(cur < prev)
	{
		return (data);
	}
	else
	{
		return (-data);
	}
}

int SVAEncoder::CircularDifferenceForColorIndex(int prev, int cur)
{
	int data;
	if(cur > prev)
	{
		data = (prev - 0 )+(nColor - cur);

	}
	else
	{
		data = (cur - 0 )+(nColor - prev);
	}

	if(abs(cur - prev) <= data)
		return cur-prev;
	else if(cur < prev)
	{
		return (data);
	}
	else
	{
		return (-data);
	}
}

int SVAEncoder::CircularDifferenceForTexCoordIndex(int prev, int cur)
{
	int data;
	if(cur > prev)
	{
		data = (prev - 0 )+(nTexCoord - cur);

	}
	else
	{
		data = (cur - 0 )+(nTexCoord - prev);
	}

	if(abs(cur - prev) <= data)
		return cur-prev;
	else if(cur < prev)
	{
		return (data);
	}
	else
	{
		return (-data);
	}
}


int SVAEncoder::CircularDifferenceForGeometry(int prev, int cur)
{
	int data;
	if(cur > prev)
	{
		data = (prev - 0 )+(quantValueForGeometry + 1  - cur);

	}
	else
	{
		data = (cur - 0 )+(quantValueForGeometry + 1 - prev);
	}

	if(abs(cur - prev) <= data)
		return cur-prev;
	else if(cur < prev)
	{
		return (data);
	}
	else
	{
		return (-data);
	}
}

int SVAEncoder::CircularDifferenceForColor(int prev, int cur)
{
	int data;
	if(cur > prev)
	{
		data = (prev - 0 )+(quantValueForColor + 1  - cur);

	}
	else
	{
		data = (cur - 0 )+(quantValueForColor + 1 - prev);
	}

	if(abs(cur - prev) <= data)
		return cur-prev;
	else if(cur < prev)
	{
		return (data);
	}
	else
	{
		return (-data);
	}
}

int SVAEncoder::CircularDifferenceForTexCoord(int prev, int cur, int flag)
{
	int data;
	if(cur > prev)
	{
		if(flag == 0) data = (prev - 0 )+(quantValueForTexCoord_u + 1  - cur);
		else data = (prev - 0 )+(quantValueForTexCoord_v + 1  - cur);

	}
	else
	{
		if(flag == 0) data = (cur - 0 )+(quantValueForTexCoord_u + 1 - prev);
		else data = (cur - 0 )+(quantValueForTexCoord_v + 1 - prev);
	}

	if(abs(cur - prev) <= data)
		return cur-prev;
	else if(cur < prev)
	{
		return (data);
	}
	else
	{
		return (-data);
	}
}

void SVAEncoder::Encode_Geometry()
{
	if(entropytype == BP)
	{
		unsigned int nMaxCheckBPL = int(log((float)MaxDifferentialForGeometry-1) / log((float)2)) +1; 
		unsigned int prefix_size = int(log((float)(nMaxCheckBPL+1)) / log((float)2)) +1;	

		Insert_Data(prefix_size, LEN_PREFIX);

		int nSymbol, nSign;
		unsigned int i;
		for(i = 0 ; i < quantCoord.size(); i++)
		{

			nSymbol = quantCoord[i];
		
			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;


			Insert_Data(BPLTable[nSymbol],prefix_size);

			if(payloadTable[nSymbol] != -1)
				Insert_Data(payloadTable[nSymbol], (BPLTable[nSymbol]-1));
			if(nSign != 2)
			{
				Insert_Data(nSign, 1);
			}

		}
	}
	else
	{

		int nSymbol, nSign;
		unsigned int i;
		unsigned int NMAX = (int) quantCoord.size() * 10 + 100;
		Adaptive_Data_Model mModel(1<<QPforGeomerty);
		Adaptive_Data_Model mSign (2);
		Arithmetic_Codec ace(NMAX, NULL);
		ace.start_encoder();
		for(i = 0 ; i < quantCoord.size(); i++)
		{

			nSymbol = quantCoord[i];
		
			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;
			ace.encode(nSymbol, mModel);
			if(nSign != 2)
				ace.encode(nSign, mSign);

		}
		ace.write_to_mem(m_compressedStream, m_compressedStreamSize);
	}
}

void SVAEncoder::Encode_Normal()
{
	if(nNormal== 0)
		return;
	int subdivision = (QPforNormal - 3) / 2;
	int i;
	for(i = 0; i < nNormal; i++)
	{
		quantNormal.push_back(_normal2code(Normal,subdivision));
		Normal +=3 ;
	}
	int codeForNormalO;
	int codeForNormalS;
	int quantized;
	int code;
	int previous = 0;
	int compareO = (7 << (2*subdivision));
	int compareS = (1 << (2*subdivision))-1;
	float qNormal[3];
	int nSign;

	if(entropytype == BP)
	{
		if(codingtype == DIFF)
		{
			int* BPtable;
			int* PayLoadtable;
			BPtable = new int [1 << (2*subdivision)];
			PayLoadtable = new int [1 << (2*subdivision)];
			for(i = 0; i<= 3 ; i++)
			{
				BPtable[i] = i;
				PayLoadtable[i] = -1;
			}
			int BPL = 4;
			int BPrange = 0;
			for(i = 4; i< (1 << (2*subdivision)) ; i++)
			{
				BPtable[i] = BPL;
				PayLoadtable[i] = BPrange++;
				
				if(BPrange >= 1<<(BPL-2))
				{
					BPL++;
					BPrange=0;
				}
			}
			int prefix = (int)(log((double)BPtable[(1 << (2*subdivision))-1]/log(2.0))) +1;
			Insert_Data(prefix,5);
				
			FILE* m_fp;
			unsigned char data;
			for(unsigned int i=0;i<dumpBitstream.size();i++)
			{
				data = dumpBitstream[i];
				fwrite(&data,sizeof(unsigned char),1,m_fp);
				
			}	
			

			

			for(i = 0; i < quantNormal.size(); i++)
			{
				quantized = quantNormal[i];
				codeForNormalO = (((quantized&compareO)>>(2*subdivision))) - (((previous&compareO)>>(2*subdivision)));
				codeForNormalS = (quantized&compareS) - (previous&compareS);
				
	

				
				previous = quantized;	
				if(abs(codeForNormalO) == 7)
				{
					if(codeForNormalO < 0)
					{
						codeForNormalO= - codeForNormalO;
						nSign = 1;
					}	
					else
					{
						nSign = 0;

					}
					Insert_Data(0,codeForNormalO);
					Insert_Data(nSign,1);

				}
				else
				{
					if(codeForNormalO < 0)
					{
						codeForNormalO= - codeForNormalO;
						nSign = 1;
					}
					else
					{
						nSign = 0;
					}
			
					Insert_Data(1,(codeForNormalO+1));
					if(codeForNormalO != 0)
					{
						Insert_Data(nSign,1);
					}
				}

				if(codeForNormalS < 0)
				{
					codeForNormalS = -codeForNormalS;
					nSign = 1;
				}
				else if(codeForNormalS > 0)
				{
					nSign = 0;
				}
				else
				{
					nSign = 2;
				}

				Insert_Data(BPtable[codeForNormalS],prefix);
				if(PayLoadtable[codeForNormalS] != -1)
				{
					Insert_Data(PayLoadtable[codeForNormalS],BPtable[codeForNormalS]-2);
				}
				if(nSign != 2)
				{
					Insert_Data(nSign,1);
				}
			}

			delete [] BPtable;
			delete [] PayLoadtable;


		}
		else
		{
			int* BPtable;
			int* PayLoadtable;
			BPtable = new int [1 << (2*subdivision+1)];
			PayLoadtable = new int [1 << (2*subdivision+1)];
			

			for(i = 0; i<= 3 ; i++)
			{
				BPtable[i] = i;
				PayLoadtable[i] = -1;
			}
			int BPL = 4;
			int BPrange = 0;
			for(i = 4; i< (1 << (2*subdivision+1)) ; i++)
			{
				BPtable[i] = BPL;
				PayLoadtable[i] = BPrange++;
				
				if(BPrange >= 1<<(BPL-2))
				{
					BPL++;
					BPrange=0;
				}
			}
			int prefix = (int)(log((double)BPtable[(1 << (2*subdivision+1))]/log(2.0))) +1;
			Insert_Data(prefix,5);

			for(i = 0; i < quantNormal.size(); i++)
			{
				quantized = quantNormal[i];
				codeForNormalO = (((quantized&compareO)>>(2*subdivision))) ^ (((previous&compareO)>>(2*subdivision)));
				codeForNormalS = (quantized&compareS) ^ (previous&compareS);

				
				previous = quantized;	
				if(abs(codeForNormalO) == 7)
				{
					Insert_Data(0,codeForNormalO);
				}
				else
				{
					Insert_Data(1,(codeForNormalO+1));
				}
				Insert_Data(BPtable[codeForNormalS],prefix);
				if(PayLoadtable[codeForNormalS] != -1)
				{
					Insert_Data(PayLoadtable[codeForNormalS],BPtable[codeForNormalS]-2);
				}
			}

			delete [] BPtable;
			delete [] PayLoadtable;

		}

	}
	else
	{
		if(codingtype == DIFF)
		{
			
			unsigned int NMAX = (int) quantNormal.size() * 10 + 100;

			Adaptive_Data_Model mModelO(1<<3);
			Adaptive_Data_Model mModelS(1<<(2*subdivision));
			Adaptive_Data_Model mSign (2);
	
			Arithmetic_Codec ace(NMAX, NULL);
			ace.start_encoder();
		
			for(i = 0; i < quantNormal.size(); i++)
			{
				quantized = quantNormal[i];
				codeForNormalO = (((quantized&compareO)>>(2*subdivision))) - (((previous&compareO)>>(2*subdivision)));
				codeForNormalS = (quantized&compareS) - (previous&compareS);

				previous = quantized;	
				if(codeForNormalO < 0)
				{
					codeForNormalO = -codeForNormalO;
					nSign = 1;
				}
				else if(codeForNormalO > 0)
					nSign = 0;
				else
					nSign = 2;
				ace.encode(codeForNormalO, mModelO);
				if(nSign != 2)
					ace.encode(nSign, mSign);

				if(codeForNormalS < 0)
				{
					codeForNormalS = -codeForNormalS;
					nSign = 1;
				}
				else if(codeForNormalS > 0)
					nSign = 0;
				else
					nSign = 2;
				ace.encode(codeForNormalS, mModelS);
				if(nSign != 2)
					ace.encode(nSign, mSign);
			}
			ace.write_to_mem(m_compressedStream, m_compressedStreamSize);

		}
		else
		{
			unsigned int NMAX = (int) quantNormal.size() * 10 + 100;

			Adaptive_Data_Model mModelO(1<<3);
			Adaptive_Data_Model mModelS(1<<(2*subdivision));
	
			Arithmetic_Codec ace(NMAX, NULL);
			ace.start_encoder();
		
			for(i = 0; i < quantNormal.size(); i++)
			{
				quantized = quantNormal[i];
				codeForNormalO = (((quantized&compareO)>>(2*subdivision))) ^ (((previous&compareO)>>(2*subdivision)));
				codeForNormalS = (quantized&compareS) ^ (previous&compareS);
				previous = quantized;	
			
				ace.encode(codeForNormalO, mModelO);
				ace.encode(codeForNormalS, mModelS);
			}
			ace.write_to_mem(m_compressedStream, m_compressedStreamSize);			
		}
	}

}

void SVAEncoder::Encode_Color()
{
	if(nColor == 0)
		return;
	// daiyong
	if(entropytype == BP)
	{
		unsigned int nMaxCheckBPL = int(log((float)MaxDifferentialForColor-1) / log((float)2)) +1; 
		unsigned int prefix_size = int(log((float)(nMaxCheckBPL+1)) / log((float)2)) +1;	

		Insert_Data(prefix_size, LEN_PREFIX);

		int nSymbol, nSign;
		unsigned int i;
		for(i = 0 ; i < quantColor.size(); i++)
		{

			nSymbol = quantColor[i];
		
			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;


			Insert_Data(BPLTable[nSymbol],prefix_size);

			if(payloadTable[nSymbol] != -1)
				Insert_Data(payloadTable[nSymbol], (BPLTable[nSymbol]-1));
			if(nSign != 2)
			{
				Insert_Data(nSign, 1);
			}

		}
	}
	else
	{

		int nSymbol, nSign;
		unsigned int i;
		unsigned int NMAX = (int) quantColor.size() * 10 + 100;
		Adaptive_Data_Model mModel(1<<QPforColor);
		Adaptive_Data_Model mSign (2);
		Arithmetic_Codec ace(NMAX, NULL);
		ace.start_encoder();
		for(i = 0 ; i < quantColor.size(); i++)
		{

			nSymbol = quantColor[i];
		
			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;
			ace.encode(nSymbol, mModel);
			if(nSign != 2)
				ace.encode(nSign, mSign);

		}
		ace.write_to_mem(m_compressedStream, m_compressedStreamSize);
	}

	// ~daiyong
}

void SVAEncoder::Encode_TexCoord()
{
	// daiyong
	if(nTexCoord == 0)
		return;
	Write2Stream(m_compressedStream, m_compressedStreamSize, &u, sizeof(int));
	Write2Stream(m_compressedStream, m_compressedStreamSize, &v, sizeof(int));	
	if(entropytype == BP)
	{
		unsigned int nMaxCheckBPL = int(log((float)MaxDifferentialForTexCoord-1) / log((float)2)) +1; 
		unsigned int prefix_size = int(log((float)(nMaxCheckBPL+1)) / log((float)2)) +1;	

		Insert_Data(prefix_size, LEN_PREFIX);

		int nSymbol, nSign;
		unsigned int i;
		for(i = 0 ; i < quantTexCoord.size(); i++)
		{

			nSymbol = quantTexCoord[i];
		
			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;


			Insert_Data(BPLTable[nSymbol],prefix_size);

			if(payloadTable[nSymbol] != -1)
				Insert_Data(payloadTable[nSymbol], (BPLTable[nSymbol]-1));
			if(nSign != 2)
			{
				Insert_Data(nSign, 1);
			}

		}
	}
	else
	{

		int nSymbol, nSign;
		unsigned int i;
		unsigned int NMAX = (int) quantTexCoord.size() * 10 + 100;
		Adaptive_Data_Model mModel(1<<QPforTexCoord);
		Adaptive_Data_Model mSign (2);
		Arithmetic_Codec ace(NMAX, NULL);
		ace.start_encoder();
		for(i = 0 ; i < quantTexCoord.size(); i++)
		{

			nSymbol = quantTexCoord[i];
		
			if(nSymbol < 0)
			{
				nSymbol = -nSymbol;
				nSign = 1;
			}
			else if(nSymbol > 0)
				nSign = 0;
			else
				nSign = 2;
			ace.encode(nSymbol, mModel);
			if(nSign != 2)
				ace.encode(nSign, mSign);

		}
		ace.write_to_mem(m_compressedStream, m_compressedStreamSize);
	}


	// ~daiyong
}

unsigned long SVAEncoder::_normal2code(const float *normal,unsigned int subdivision)
{
	  float vx = normal[0], vy = normal[1], vz = normal[2];
	  unsigned long octantCode = 0;

	  // Generate bits b30, b29, b28 of code (0: positive, 1: negative)
	  if (vx < 0) { vx = -vx; octantCode  = 0x00000004; }
	  if (vy < 0) { vy = -vy; octantCode |= 0x00000002; }
	  if (vz < 0) { vz = -vz; octantCode |= 0x00000001; }

	  // Convert coordinates to integer values
	  unsigned long maxCoord = 1 << subdivision;
	  float s = maxCoord / (vx + vy + vz);
	  unsigned long x = (unsigned long)(vx*s);
	  unsigned long y = (unsigned long)(vy*s);
	  unsigned long z = (unsigned long)(maxCoord-vx*s-vy*s); //  long z = vz * s;

	  short upsideDown = 0;

	  if (maxCoord - x - y - z == 2) upsideDown = 1;
	  else if (x == maxCoord ) x -= 1;
	  else if ( x + y == maxCoord ) y -= 1;

	  unsigned long tricode = y * (2 * maxCoord - 1) - y * (y - 1) +    2 * x + upsideDown;

	  unsigned long code =  (octantCode << 2*subdivision) | tricode;
 
	  return code;
}



void SVAEncoder::Encode_Connectivity()
{
	if(entropytype == BP)
	{
		unsigned int nMaxCheckBPL = int(log((float)MaxDifferentialForConnectivity-1) / log((float)2)) +1; 
		unsigned int prefix_size = int(log((float)(nMaxCheckBPL+1)) / log((float)2)) +1;	
		int nBPL, nPayload, nSign;
		unsigned int i;
		unsigned int countForType =0;
		unsigned int countForFD =0;  // Not to support RunLength Coding
		unsigned int countForDifCI =0;
		unsigned int countForPos =0;
		unsigned int countForRotate =0;
		int nType, nPos, nFD, nDifCI, nRotate;

		Insert_Data(FDmode, LEN_BIT);
		if(FDmode == 0)
			Insert_Data(FDvalue, LEN_BIT);
		Insert_Data(prefix_size, LEN_PREFIX);


		for( ;countForDifCI<3;countForDifCI++)
		{
			nBPL= BPLTable[difCoordIndex[countForDifCI]];
			nPayload = payloadTable[difCoordIndex[countForDifCI]];
			Insert_Data(nBPL,prefix_size);
			if(nPayload != -1)
				Insert_Data(nPayload, (nBPL-1));
		}

		for( ; countForType < type.size(); countForType++)
		{
			nType = type[countForType];
			Insert_Data(1, nType+1);
			switch(nType)
			{
			case 0:
				nPos= position[countForPos++];
				nFD = faceDirection[countForFD++];
				
				nDifCI = difCoordIndex[countForDifCI++];
				if(nPos == 0 )
					Insert_Data(nPos, 1);
				else
					Insert_Data((nPos+1), 2);
				if(FDmode == 1)
					Insert_Data(nFD,1);
				//nsert_Data(nRotate, 2);
				if(nDifCI > 0)
					nSign = 0;
				else
				{
					nDifCI = -nDifCI;
					nSign = 1;
				}
				nBPL= BPLTable[nDifCI-1];
				nPayload = payloadTable[nDifCI-1];
				Insert_Data(nBPL,prefix_size);
				if(nPayload != -1)
					Insert_Data(nPayload, (nBPL-1));
				Insert_Data(nSign,1);
				nRotate = rotate[countForRotate++];
				if(nRotate == 0 )
					Insert_Data(nRotate, 1);
				else
					Insert_Data((nRotate+1), 2);

				break;
			case 1:
				for(i=0; i < 3; i++)
				{
					nDifCI = difCoordIndex[countForDifCI++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nBPL= BPLTable[nDifCI-1];
					nPayload = payloadTable[nDifCI-1];
					Insert_Data(nBPL,prefix_size);
					if(nPayload != -1)
						Insert_Data(nPayload, (nBPL-1));
					Insert_Data(nSign,1);
				}
				break;
			case 2:
				nPos= position[countForPos++];
				//nRotate = rotate[countForRotate++];
				if(nPos == 0 )
					Insert_Data(nPos, 1);
				else
					Insert_Data((nPos+1), 2);
				//nsert_Data(nRotate, 2);
				for(i=0; i < 2; i++)
				{
					nDifCI = difCoordIndex[countForDifCI++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nBPL= BPLTable[nDifCI-1];
					nPayload = payloadTable[nDifCI-1];
					Insert_Data(nBPL,prefix_size);
					if(nPayload != -1)
						Insert_Data(nPayload, (nBPL-1));
					Insert_Data(nSign,1);
				}
				nRotate = rotate[countForRotate++];
				if(nRotate == 0 )
					Insert_Data(nRotate, 1);
				else
					Insert_Data((nRotate+1), 2);
				break;
			case 3:
				for(i=0; i < 3; i++)
				{
					nDifCI = difCoordIndex[countForDifCI++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nBPL= BPLTable[nDifCI];
					nPayload = payloadTable[nDifCI];
					Insert_Data(nBPL,prefix_size);
					if(nPayload != -1)
						Insert_Data(nPayload, (nBPL-1));
					Insert_Data(nSign,1);
				}
				break;
			case 4:
				nFD = faceDirection[countForFD++];
				if(FDmode == 1)
					Insert_Data(nFD,1);
				nRotate = rotate[countForRotate++];
				if(nRotate == 0 )
					Insert_Data(nRotate, 1);
				else
					Insert_Data((nRotate+1), 2);
				
				break;
			}
		}
	}
	else
	{
		
		unsigned int i;
		unsigned int countForType =0;
		unsigned int countForFD =0;  // Not to support RunLength Coding
		unsigned int countForDifCI =0;
		unsigned int countForPos =0;
		unsigned int countForRotate =0;
		int nType, nPos, nFD, nDifCI, nRotate;

		unsigned int attr_num;
		int previousQuantValue[3];
		for(attr_num = 0; attr_num < 3; attr_num++)
		{
			previousQuantValue[attr_num] = 0;
		}

		int nSymbol, nSign, input;
		int lengthCoordIndex = int(log((double)nCoord-1)/log(2.0)) + 1;
		unsigned int NMAX = (int) nCoordIndex*3 * lengthCoordIndex + 100;

		Adaptive_Data_Model mModel(1024);
		Adaptive_Data_Model mhasnext (2);
		Adaptive_Data_Model mSign (2);
		Adaptive_Data_Model mType (5);
		Adaptive_Data_Model mFD (2);
		Adaptive_Data_Model mPos (3);
		Adaptive_Data_Model mRotate (3);

		Arithmetic_Codec ace(NMAX, NULL);
		ace.start_encoder();


		for( ;countForDifCI<3;countForDifCI++)
		{
			nSymbol=difCoordIndex[countForDifCI];
			while(1)
			{
				ace.encode(nSymbol%1024, mModel);
				nSymbol/=1024;
				if(nSymbol)
					ace.encode(1, mhasnext);
				else
					break;
			}
			ace.encode(0, mhasnext);

		}

		for( ; countForType < type.size(); countForType++)
		{
			nType = type[countForType];
			ace.encode(nType,mType);			
			switch(nType)
			{
			case 0:
				nPos= position[countForPos++];
				nFD = faceDirection[countForFD++];
				
				nDifCI = difCoordIndex[countForDifCI++];
				ace.encode(nPos, mPos);
				ace.encode(nFD, mFD);
				
				if(nDifCI > 0)
					nSign = 0;
				else
				{
					nDifCI = -nDifCI;
					nSign = 1;
				}
				nSymbol = nDifCI-1;
				while(1)
				{
					ace.encode(nSymbol%1024, mModel);
					nSymbol/=1024;
					if(nSymbol)
						ace.encode(1, mhasnext);
					else
						break;
				}
				ace.encode(0, mhasnext);
				ace.encode(nSign, mSign);
				nRotate = rotate[countForRotate++];
				ace.encode(nRotate, mRotate);

				break;
			case 1:
				for(i=0; i < 3; i++)
				{
					nDifCI = difCoordIndex[countForDifCI++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nSymbol = nDifCI-1;
					while(1)
					{
						ace.encode(nSymbol%1024, mModel);
						nSymbol/=1024;
						if(nSymbol)
							ace.encode(1, mhasnext);
						else
							break;
					}
					ace.encode(0, mhasnext);
					ace.encode(nSign, mSign);
				}
				break;
			case 2:
				nPos= position[countForPos++];
				ace.encode(nPos, mPos);
				for(i=0; i < 2; i++)
				{
					nDifCI = difCoordIndex[countForDifCI++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nSymbol = nDifCI-1;
					while(1)
					{
						ace.encode(nSymbol%1024, mModel);
						nSymbol/=1024;
						if(nSymbol)
							ace.encode(1, mhasnext);
						else
							break;
					}
					ace.encode(0, mhasnext);
					ace.encode(nSign, mSign);
				}
				nRotate = rotate[countForRotate++];
				ace.encode(nRotate, mRotate);
				break;
			case 3:
				for(i=0; i < 3; i++)
				{
					nDifCI = difCoordIndex[countForDifCI++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nSymbol = nDifCI;
					while(1)
					{
						ace.encode(nSymbol%1024, mModel);
						nSymbol/=1024;
						if(nSymbol)
							ace.encode(1, mhasnext);
						else
							break;
					}
					ace.encode(0, mhasnext);
					ace.encode(nSign, mSign);
				}
				break;
			case 4:
				nFD = faceDirection[countForFD++];
				ace.encode(nFD, mFD);
				nRotate = rotate[countForRotate++];
				ace.encode(nRotate, mRotate);
				break;
			}
		}
		ace.write_to_mem(m_compressedStream, m_compressedStreamSize);

	}

}

void SVAEncoder::Analyze_ColorIndex()
{
	int symbol;
	int flip=0;
	int k, l, j;
	int DifferetialCI[2];
	int rotcount=0;
	int nRot;

	type.clear();
	difIndex.clear(); 
	position.clear();
	faceDirection.clear();
	rotate.clear();;

	MaxDifferentialForColorIndex = 0;


	difIndex.push_back(ColorIndex[0]); // first face 
	if(abs(CoordIndex[0]) > MaxDifferentialForColorIndex)
	{
		MaxDifferentialForColorIndex = ColorIndex[0];
	}
	difIndex.push_back(ColorIndex[1]);
	if(abs(ColorIndex[1]) > MaxDifferentialForColorIndex)
	{
		MaxDifferentialForColorIndex = ColorIndex[1];
	}
	difIndex.push_back(ColorIndex[2]); 
	if(abs(ColorIndex[2]) > MaxDifferentialForColorIndex)
	{
		MaxDifferentialForColorIndex = ColorIndex[2];
	}

	//From second face
	for(j = 1; j < nColorIndex; j++)
	{

		//Count the overlapped points
		int sizeofOverlapped=0;
		int myType=0;
		int cheecktype4 =0;

		if(cheecktype4 == 1)
		{
			cheecktype4 =0;
			sizeofOverlapped = 4;

		}
		else 
		{
			for(k=0;k<3;k++)
			{
				for(l=k+1;l<3;l++)
				{
					if(ColorIndex[j*3+k] == ColorIndex[j*3+l])
					{
						cheecktype4 == 1;
						sizeofOverlapped = 4;
						break;
					}
				}
			}
			if(sizeofOverlapped != 4)
			{
				for(k=0;k<3;k++)
				{
					for(l=0;l<3;l++)
					{
						if(ColorIndex[j*3+k] == ColorIndex[(j-1)*3+l])
						{
							sizeofOverlapped++;
							break;
						}
					}
				}
			}
		}

		// Changing type information based on its frequency.
		// Type 2 ==> Type 0
		// Type 0 ==> Type 1
		// Type 1 ==> Type 2
		// Type 3 ==> Type 3

		

		
		switch(sizeofOverlapped)
		{
		case 0: //Type C encode 3 symbols
			myType = 1;
			
			symbol = CircularDifferenceForColorIndex(ColorIndex[(j-1)*3], ColorIndex[j*3]);
			difIndex.push_back(symbol);
			if(abs(symbol)-1 > MaxDifferentialForColorIndex)
			{
				MaxDifferentialForColorIndex = abs(symbol)-1;
			}

			symbol = CircularDifferenceForColorIndex(ColorIndex[(j-1)*3+1], ColorIndex[j*3+1]);
			difIndex.push_back(symbol);
			if(abs(symbol)-1 > MaxDifferentialForColorIndex)
			{
				MaxDifferentialForColorIndex = abs(symbol)-1;
			}
			
			symbol = CircularDifferenceForColorIndex(ColorIndex[(j-1)*3+2], ColorIndex[j*3+2]);
			difIndex.push_back(symbol);
			if(abs(symbol)-1 > MaxDifferentialForColorIndex)
			{
				MaxDifferentialForColorIndex = abs(symbol)-1;
			}

			break;
		case 1: //Type B encode 2 symbols 
			myType = 2;

			for(k=0;k<3;k++)
			{
				for(l=0;l<3;l++)
					if(ColorIndex[j*3+k] == ColorIndex[(j-1)*3+l])						
					{
						//Rotation
						nRot = Rotation((ColorIndex+j*3), k-l);
						rotate.push_back(nRot);
					

						//encode point that indicates same symbol						;
						position.push_back(l);

						//encode 2 symbols						
						symbol = CircularDifferenceForColorIndex(ColorIndex[(j-1)*3+(l+1)%3], ColorIndex[j*3+(l+1)%3]);
						difIndex.push_back(symbol);
						if(abs(symbol)-1 > MaxDifferentialForColorIndex)
						{
							MaxDifferentialForColorIndex = abs(symbol)-1;
						}
						symbol = CircularDifferenceForColorIndex(ColorIndex[(j-1)*3+(l+2)%3], ColorIndex[j*3+(l+2)%3]);
						difIndex.push_back(symbol);
						if(abs(symbol)-1 > MaxDifferentialForColorIndex)
						{
							MaxDifferentialForColorIndex = abs(symbol)-1;
						}

						k=3;
						InverseRotation((ColorIndex+j*3),nRot);
						break;
					}
			}

			break;



		case 2: //Type A encode 1 symbol
			//DifferetialCI[0] indicates different symbol's position in the current face
			myType = 0;
			for(k=0;k<3;k++)
			{
				for(l=0;l<3;l++)
					if(ColorIndex[j*3+k] == ColorIndex[(j-1)*3+l])
						break;
				if(l==3)
					DifferetialCI[0]=k;
			}
			//DifferetialCI[1] indicates different symbol's position in the current face
			for(k=0;k<3;k++)
			{
				for(l=0;l<3;l++)
					if(ColorIndex[j*3+l] == ColorIndex[(j-1)*3+k])
						break;
				if(l==3)
					DifferetialCI[1]=k;
			}
			//rotate to match DifferetialCI[0] and DifferetialCI[1]

			nRot = Rotation((ColorIndex+j*3), DifferetialCI[0]-DifferetialCI[1]);
			rotate.push_back(nRot);
		


			flip=1;
			
			for(k=0; k <3; k++)
			{
				if(ColorIndex[j*3+k] == ColorIndex[(j-1)*3+k])
				{
					flip=0;
					break;
				}
			}
			//coding flip flag
			faceDirection.push_back(flip);

			//encode pos that indicates different symbol
			position.push_back(DifferetialCI[1]);


			//encode 1 symbol
			symbol = CircularDifferenceForColorIndex(ColorIndex[(j-1)*3+DifferetialCI[1]],ColorIndex[j*3+DifferetialCI[1]]);
			difIndex.push_back(symbol);
			if(abs(symbol)-1 > MaxDifferentialForColorIndex)
			{
				MaxDifferentialForColorIndex = abs(symbol)-1;
			}
			InverseRotation((ColorIndex+j*3),nRot);
			break;

		
		case 3: //Type D encode only flip bit
			myType = 4;
			rotcount=0;
			while(ColorIndex[j*3+1] != ColorIndex[(j-1)*3+1])
			{
				Rotation((ColorIndex+j*3), 1);
				rotcount++;
			}
			rotate.push_back(rotcount);
		

			if(	ColorIndex[j*3] == ColorIndex[(j-1)*3] )
				flip=0;
			else
				flip=1;
			faceDirection.push_back(flip);
			InverseRotation((ColorIndex+j*3),rotcount);
			//coding flip flag			
			

			break;
		case 4:
			myType = 3;
			symbol = CircularDifferenceForColorIndex(ColorIndex[(j-1)*3], ColorIndex[j*3]);
			difIndex.push_back(symbol);
			if(abs(symbol) > MaxDifferentialForColorIndex)
			{
				MaxDifferentialForColorIndex = abs(symbol);
			}

			symbol = CircularDifferenceForColorIndex(ColorIndex[(j-1)*3+1], ColorIndex[j*3+1]);
			difIndex.push_back(symbol);
			if(abs(symbol) > MaxDifferentialForColorIndex)
			{
				MaxDifferentialForColorIndex = abs(symbol);
			}
			
			symbol = CircularDifferenceForColorIndex(ColorIndex[(j-1)*3+2], ColorIndex[j*3+2]);
			difIndex.push_back(symbol);
			if(abs(symbol) > MaxDifferentialForColorIndex)
			{
				MaxDifferentialForColorIndex = abs(symbol);
			}

			break;

		}
		type.push_back(myType); //type information

	}

	//change
	if(faceDirection.size() == 0 )
	{
		FDmode = 1;
	}
	else
	{
		FDvalue = faceDirection[0];
		FDmode = 0;
		for(j = 1 ; j< faceDirection.size(); j++)
		{
			if(FDvalue != faceDirection[j])
			{
				FDmode = 1;
				break;
			}

		}
	}

}

void SVAEncoder::Analyze_TexCoordIndex()
{
	int symbol;
	int flip=0;
	int k, l, j;
	int DifferetialCI[2];
	int rotcount=0;
	int nRot;

	type.clear();
	difIndex.clear(); 
	position.clear();
	faceDirection.clear();
	rotate.clear();;

	MaxDifferentialForTexCoordIndex = 0;


	difIndex.push_back(TexCoordIndex[0]); // first face 
	if(abs(CoordIndex[0]) > MaxDifferentialForTexCoordIndex)
	{
		MaxDifferentialForTexCoordIndex = TexCoordIndex[0];
	}
	difIndex.push_back(TexCoordIndex[1]);
	if(abs(TexCoordIndex[1]) > MaxDifferentialForTexCoordIndex)
	{
		MaxDifferentialForTexCoordIndex = TexCoordIndex[1];
	}
	difIndex.push_back(TexCoordIndex[2]); 
	if(abs(TexCoordIndex[2]) > MaxDifferentialForTexCoordIndex)
	{
		MaxDifferentialForTexCoordIndex = TexCoordIndex[2];
	}

	//From second face
	for(j = 1; j < nTexCoordIndex; j++)
	{

		//Count the overlapped points
		int sizeofOverlapped=0;
		int myType=0;
		int cheecktype4 =0;

		if(cheecktype4 == 1)
		{
			cheecktype4 =0;
			sizeofOverlapped = 4;

		}
		else 
		{
			for(k=0;k<3;k++)
			{
				for(l=k+1;l<3;l++)
				{
					if(TexCoordIndex[j*3+k] == TexCoordIndex[j*3+l])
					{
						cheecktype4 == 1;
						sizeofOverlapped = 4;
						break;
					}
				}
			}
			if(sizeofOverlapped != 4)
			{
				for(k=0;k<3;k++)
				{
					for(l=0;l<3;l++)
					{
						if(TexCoordIndex[j*3+k] == TexCoordIndex[(j-1)*3+l])
						{
							sizeofOverlapped++;
							break;
						}
					}
				}
			}
		}

		// Changing type information based on its frequency.
		// Type 2 ==> Type 0
		// Type 0 ==> Type 1
		// Type 1 ==> Type 2
		// Type 3 ==> Type 3

		

		
		switch(sizeofOverlapped)
		{
		case 0: //Type C encode 3 symbols
			myType = 1;
			
			symbol = CircularDifferenceForTexCoordIndex(TexCoordIndex[(j-1)*3], TexCoordIndex[j*3]);
			difIndex.push_back(symbol);
			if(abs(symbol)-1 > MaxDifferentialForTexCoordIndex)
			{
				MaxDifferentialForTexCoordIndex = abs(symbol)-1;
			}

			symbol = CircularDifferenceForTexCoordIndex(TexCoordIndex[(j-1)*3+1], TexCoordIndex[j*3+1]);
			difIndex.push_back(symbol);
			if(abs(symbol)-1 > MaxDifferentialForTexCoordIndex)
			{
				MaxDifferentialForTexCoordIndex = abs(symbol)-1;
			}
			
			symbol = CircularDifferenceForTexCoordIndex(TexCoordIndex[(j-1)*3+2], TexCoordIndex[j*3+2]);
			difIndex.push_back(symbol);
			if(abs(symbol)-1 > MaxDifferentialForTexCoordIndex)
			{
				MaxDifferentialForTexCoordIndex = abs(symbol)-1;
			}

			break;
		case 1: //Type B encode 2 symbols 
			myType = 2;

			for(k=0;k<3;k++)
			{
				for(l=0;l<3;l++)
					if(TexCoordIndex[j*3+k] == TexCoordIndex[(j-1)*3+l])						
					{
						//Rotation
						nRot = Rotation((TexCoordIndex+j*3), k-l);
						rotate.push_back(nRot);
				



						//encode point that indicates same symbol						;
						position.push_back(l);

						//encode 2 symbols						
						symbol = CircularDifferenceForTexCoordIndex(TexCoordIndex[(j-1)*3+(l+1)%3], TexCoordIndex[j*3+(l+1)%3]);
						difIndex.push_back(symbol);
						if(abs(symbol)-1 > MaxDifferentialForTexCoordIndex)
						{
							MaxDifferentialForTexCoordIndex = abs(symbol)-1;
						}
						symbol = CircularDifferenceForTexCoordIndex(TexCoordIndex[(j-1)*3+(l+2)%3], TexCoordIndex[j*3+(l+2)%3]);
						difIndex.push_back(symbol);
						if(abs(symbol)-1 > MaxDifferentialForTexCoordIndex)
						{
							MaxDifferentialForTexCoordIndex = abs(symbol)-1;
						}

						k=3;
						InverseRotation((TexCoordIndex+j*3),nRot);
						break;
					}
			}

			break;



		case 2: //Type A encode 1 symbol
			//DifferetialCI[0] indicates different symbol's position in the current face
			myType = 0;
			for(k=0;k<3;k++)
			{
				for(l=0;l<3;l++)
					if(TexCoordIndex[j*3+k] == TexCoordIndex[(j-1)*3+l])
						break;
				if(l==3)
					DifferetialCI[0]=k;
			}
			//DifferetialCI[1] indicates different symbol's position in the current face
			for(k=0;k<3;k++)
			{
				for(l=0;l<3;l++)
					if(TexCoordIndex[j*3+l] == TexCoordIndex[(j-1)*3+k])
						break;
				if(l==3)
					DifferetialCI[1]=k;
			}
			//rotate to match DifferetialCI[0] and DifferetialCI[1]

			nRot = Rotation((TexCoordIndex+j*3), DifferetialCI[0]-DifferetialCI[1]);
			rotate.push_back(nRot);
			

			flip=1;
			
			for(k=0; k <3; k++)
			{
				if(TexCoordIndex[j*3+k] == TexCoordIndex[(j-1)*3+k])
				{
					flip=0;
					break;
				}
			}
			//coding flip flag
			faceDirection.push_back(flip);

			//encode pos that indicates different symbol
			position.push_back(DifferetialCI[1]);


			//encode 1 symbol
			symbol = CircularDifferenceForTexCoordIndex(TexCoordIndex[(j-1)*3+DifferetialCI[1]],TexCoordIndex[j*3+DifferetialCI[1]]);
			difIndex.push_back(symbol);
			if(abs(symbol)-1 > MaxDifferentialForTexCoordIndex)
			{
				MaxDifferentialForTexCoordIndex = abs(symbol)-1;
			}
			InverseRotation((TexCoordIndex+j*3),nRot);
			
			break;

		
		case 3: //Type D encode only flip bit
			myType = 4;
			rotcount=0;
			while(TexCoordIndex[j*3+1] != TexCoordIndex[(j-1)*3+1])
			{
				Rotation((TexCoordIndex+j*3), 1);
				rotcount++;
			}
			rotate.push_back(rotcount);
		

			if(	TexCoordIndex[j*3] == TexCoordIndex[(j-1)*3] )
				flip=0;
			else
				flip=1;
			faceDirection.push_back(flip);
			InverseRotation((TexCoordIndex+j*3),rotcount);
			//coding flip flag			
			

			break;
		case 4:
			myType = 3;
			symbol = CircularDifferenceForTexCoordIndex(TexCoordIndex[(j-1)*3], TexCoordIndex[j*3]);
			difIndex.push_back(symbol);
			if(abs(symbol) > MaxDifferentialForTexCoordIndex)
			{
				MaxDifferentialForTexCoordIndex = abs(symbol);
			}

			symbol = CircularDifferenceForTexCoordIndex(TexCoordIndex[(j-1)*3+1], TexCoordIndex[j*3+1]);
			difIndex.push_back(symbol);
			if(abs(symbol) > MaxDifferentialForTexCoordIndex)
			{
				MaxDifferentialForTexCoordIndex = abs(symbol);
			}
			
			symbol = CircularDifferenceForTexCoordIndex(TexCoordIndex[(j-1)*3+2], TexCoordIndex[j*3+2]);
			difIndex.push_back(symbol);
			if(abs(symbol) > MaxDifferentialForTexCoordIndex)
			{
				MaxDifferentialForTexCoordIndex = abs(symbol);
			}

			break;

		}
		type.push_back(myType); //type information

	}

	//change
	if(faceDirection.size() == 0 )
	{
		FDmode = 1;
	}
	else
	{
		FDvalue = faceDirection[0];
		FDmode = 0;
		for(j = 1 ; j< faceDirection.size(); j++)
		{
			if(FDvalue != faceDirection[j])
			{
				FDmode = 1;
				break;
			}

		}
	}

}

void SVAEncoder::Encode_NormalIndex()
{
	if(nNormalIndex == 0)
		return;
	Analyze_NormalIndex();

	if(entropytype == BP)
	{
		unsigned int nMaxCheckBPL = int(log((float)MaxDifferentialForNormalIndex-1) / log((float)2)) +1; 
		unsigned int prefix_size = int(log((float)(nMaxCheckBPL+1)) / log((float)2)) +1;	
		int nBPL, nPayload, nSign;
		unsigned int i;
		unsigned int countForType =0;
		unsigned int countForFD =0;  // Not to support RunLength Coding
		unsigned int countForDifIndex =0;
		unsigned int countForPos =0;
		unsigned int countForRotate =0;
		int nType, nPos, nFD, nDifCI, nRotate;

		Insert_Data(FDmode, LEN_BIT);
		if(FDmode == 0)
			Insert_Data(FDvalue, LEN_BIT);
		Insert_Data(prefix_size, LEN_PREFIX);


		for( ;countForDifIndex<3;countForDifIndex++)
		{
			nBPL= BPLTable[difIndex[countForDifIndex]];
			nPayload = payloadTable[difIndex[countForDifIndex]];
			Insert_Data(nBPL,prefix_size);
			if(nPayload != -1)
				Insert_Data(nPayload, (nBPL-1));
		}

		for( ; countForType < type.size(); countForType++)
		{
			nType = type[countForType];
			Insert_Data(1, nType+1);
			switch(nType)
			{
			case 0:
				nPos= position[countForPos++];
				nFD = faceDirection[countForFD++];
				
				nDifCI = difIndex[countForDifIndex++];
				if(nPos == 0 )
					Insert_Data(nPos, 1);
				else
					Insert_Data((nPos+1), 2);
				if(FDmode == 1)
					Insert_Data(nFD,1);
				//nsert_Data(nRotate, 2);
				if(nDifCI > 0)
					nSign = 0;
				else
				{
					nDifCI = -nDifCI;
					nSign = 1;
				}
				nBPL= BPLTable[nDifCI-1];
				nPayload = payloadTable[nDifCI-1];
				Insert_Data(nBPL,prefix_size);
				if(nPayload != -1)
					Insert_Data(nPayload, (nBPL-1));
				Insert_Data(nSign,1);
				nRotate = rotate[countForRotate++];
				if(nRotate == 0 )
					Insert_Data(nRotate, 1);
				else
					Insert_Data((nRotate+1), 2);

				break;
			case 1:
				for(i=0; i < 3; i++)
				{
					nDifCI = difIndex[countForDifIndex++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nBPL= BPLTable[nDifCI-1];
					nPayload = payloadTable[nDifCI-1];
					Insert_Data(nBPL,prefix_size);
					if(nPayload != -1)
						Insert_Data(nPayload, (nBPL-1));
					Insert_Data(nSign,1);
				}
				break;
			case 2:
				nPos= position[countForPos++];
				//nRotate = rotate[countForRotate++];
				if(nPos == 0 )
					Insert_Data(nPos, 1);
				else
					Insert_Data((nPos+1), 2);
				//nsert_Data(nRotate, 2);
				for(i=0; i < 2; i++)
				{
					nDifCI = difIndex[countForDifIndex++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nBPL= BPLTable[nDifCI-1];
					nPayload = payloadTable[nDifCI-1];
					Insert_Data(nBPL,prefix_size);
					if(nPayload != -1)
						Insert_Data(nPayload, (nBPL-1));
					Insert_Data(nSign,1);
				}
				nRotate = rotate[countForRotate++];
				if(nRotate == 0 )
					Insert_Data(nRotate, 1);
				else
					Insert_Data((nRotate+1), 2);
				break;
			case 3:
				for(i=0; i < 3; i++)
				{
					nDifCI = difIndex[countForDifIndex++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nBPL= BPLTable[nDifCI];
					nPayload = payloadTable[nDifCI];
					Insert_Data(nBPL,prefix_size);
					if(nPayload != -1)
						Insert_Data(nPayload, (nBPL-1));
					Insert_Data(nSign,1);
				}
				break;
			case 4:
				nFD = faceDirection[countForFD++];
				if(FDmode == 1)
					Insert_Data(nFD,1);
				nRotate = rotate[countForRotate++];
				if(nRotate == 0 )
					Insert_Data(nRotate, 1);
				else
					Insert_Data((nRotate+1), 2);
				
				break;
			}
		}
	}
	else
	{
		
		unsigned int i;
		unsigned int countForType =0;
		unsigned int countForFD =0;  // Not to support RunLength Coding
		unsigned int countForDifIndex =0;
		unsigned int countForPos =0;
		unsigned int countForRotate =0;
		int nType, nPos, nFD, nDifCI, nRotate;

		unsigned int attr_num;
		int previousQuantValue[3];
		for(attr_num = 0; attr_num < 3; attr_num++)
		{
			previousQuantValue[attr_num] = 0;
		}

		int nSymbol, nSign, input;
		int lengthNormalIndex = int(log((double)nNormal-1)/log(2.0)) + 1;
		unsigned int NMAX = (int) nNormalIndex*3 * lengthNormalIndex + 100;

		Adaptive_Data_Model mModel(1024);
		Adaptive_Data_Model mhasnext (2);
		Adaptive_Data_Model mSign (2);
		Adaptive_Data_Model mType (5);
		Adaptive_Data_Model mFD (2);
		Adaptive_Data_Model mPos (3);
		Adaptive_Data_Model mRotate (3);

		Arithmetic_Codec ace(NMAX, NULL);
		ace.start_encoder();


		for( ;countForDifIndex<3;countForDifIndex++)
		{
			nSymbol=difIndex[countForDifIndex];
			while(1)
			{
				ace.encode(nSymbol%1024, mModel);
				nSymbol/=1024;
				if(nSymbol)
					ace.encode(1, mhasnext);
				else
					break;
			}
			ace.encode(0, mhasnext);

		}

		for( ; countForType < type.size(); countForType++)
		{
			nType = type[countForType];
			ace.encode(nType, mType);
			switch(nType)
			{
			case 0:
				nPos= position[countForPos++];
				nFD = faceDirection[countForFD++];
				
				nDifCI = difIndex[countForDifIndex++];
				ace.encode(nPos, mPos);
				ace.encode(nFD, mFD);
				
				if(nDifCI > 0)
					nSign = 0;
				else
				{
					nDifCI = -nDifCI;
					nSign = 1;
				}
				nSymbol = nDifCI-1;
				while(1)
				{
					ace.encode(nSymbol%1024, mModel);
					nSymbol/=1024;
					if(nSymbol)
						ace.encode(1, mhasnext);
					else
						break;
				}
				ace.encode(0, mhasnext);
				ace.encode(nSign, mSign);
				nRotate = rotate[countForRotate++];
				ace.encode(nRotate, mRotate);

				break;
			case 1:
				for(i=0; i < 3; i++)
				{
					nDifCI = difIndex[countForDifIndex++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nSymbol = nDifCI-1;
					while(1)
					{
						ace.encode(nSymbol%1024, mModel);
						nSymbol/=1024;
						if(nSymbol)
							ace.encode(1, mhasnext);
						else
							break;
					}
					ace.encode(0, mhasnext);
					ace.encode(nSign, mSign);
				}
				break;
			case 2:
				nPos= position[countForPos++];
				ace.encode(nPos, mPos);
				for(i=0; i < 2; i++)
				{
					nDifCI = difIndex[countForDifIndex++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nSymbol = nDifCI-1;
					while(1)
					{
						ace.encode(nSymbol%1024, mModel);
						nSymbol/=1024;
						if(nSymbol)
							ace.encode(1, mhasnext);
						else
							break;
					}
					ace.encode(0, mhasnext);
					ace.encode(nSign, mSign);
				}
				nRotate = rotate[countForRotate++];
				ace.encode(nRotate, mRotate);
				break;
			case 3:
				for(i=0; i < 3; i++)
				{
					nDifCI = difIndex[countForDifIndex++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nSymbol = nDifCI;
					while(1)
					{
						ace.encode(nSymbol%1024, mModel);
						nSymbol/=1024;
						if(nSymbol)
							ace.encode(1, mhasnext);
						else
							break;
					}
					ace.encode(0, mhasnext);
					ace.encode(nSign, mSign);
				}
				break;
			case 4:
				nFD = faceDirection[countForFD++];
				ace.encode(nFD, mFD);
				nRotate = rotate[countForRotate++];
				ace.encode(nRotate, mRotate);
				break;
			}
		}
		ace.write_to_mem(m_compressedStream, m_compressedStreamSize);
	}


}

void SVAEncoder::Encode_ColorIndex()

{
	if(nColorIndex == 0)
		return;
	Analyze_ColorIndex();

	if(entropytype == BP)
	{
		unsigned int nMaxCheckBPL = int(log((float)MaxDifferentialForColorIndex-1) / log((float)2)) +1; 
		unsigned int prefix_size = int(log((float)(nMaxCheckBPL+1)) / log((float)2)) +1;	
		int nBPL, nPayload, nSign;
		unsigned int i;
		unsigned int countForType =0;
		unsigned int countForFD =0;  // Not to support RunLength Coding
		unsigned int countForDifIndex =0;
		unsigned int countForPos =0;
		unsigned int countForRotate =0;
		int nType, nPos, nFD, nDifCI, nRotate;

		Insert_Data(FDmode, LEN_BIT);
		if(FDmode == 0)
			Insert_Data(FDvalue, LEN_BIT);
		Insert_Data(prefix_size, LEN_PREFIX);


		for( ;countForDifIndex<3;countForDifIndex++)
		{
			nBPL= BPLTable[difIndex[countForDifIndex]];
			nPayload = payloadTable[difIndex[countForDifIndex]];
			Insert_Data(nBPL,prefix_size);
			if(nPayload != -1)
				Insert_Data(nPayload, (nBPL-1));
		}

		for( ; countForType < type.size(); countForType++)
		{
			nType = type[countForType];
			Insert_Data(1, nType+1);
			switch(nType)
			{
			case 0:
				nPos= position[countForPos++];
				nFD = faceDirection[countForFD++];
				
				nDifCI = difIndex[countForDifIndex++];
				if(nPos == 0 )
					Insert_Data(nPos, 1);
				else
					Insert_Data((nPos+1), 2);
				if(FDmode == 1)
					Insert_Data(nFD,1);
				//nsert_Data(nRotate, 2);
				if(nDifCI > 0)
					nSign = 0;
				else
				{
					nDifCI = -nDifCI;
					nSign = 1;
				}
				nBPL= BPLTable[nDifCI-1];
				nPayload = payloadTable[nDifCI-1];
				Insert_Data(nBPL,prefix_size);
				if(nPayload != -1)
					Insert_Data(nPayload, (nBPL-1));
				Insert_Data(nSign,1);
				nRotate = rotate[countForRotate++];
				if(nRotate == 0 )
					Insert_Data(nRotate, 1);
				else
					Insert_Data((nRotate+1), 2);

				break;
			case 1:
				for(i=0; i < 3; i++)
				{
					nDifCI = difIndex[countForDifIndex++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nBPL= BPLTable[nDifCI-1];
					nPayload = payloadTable[nDifCI-1];
					Insert_Data(nBPL,prefix_size);
					if(nPayload != -1)
						Insert_Data(nPayload, (nBPL-1));
					Insert_Data(nSign,1);
				}
				break;
			case 2:
				nPos= position[countForPos++];
				//nRotate = rotate[countForRotate++];
				if(nPos == 0 )
					Insert_Data(nPos, 1);
				else
					Insert_Data((nPos+1), 2);
				//nsert_Data(nRotate, 2);
				for(i=0; i < 2; i++)
				{
					nDifCI = difIndex[countForDifIndex++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nBPL= BPLTable[nDifCI-1];
					nPayload = payloadTable[nDifCI-1];
					Insert_Data(nBPL,prefix_size);
					if(nPayload != -1)
						Insert_Data(nPayload, (nBPL-1));
					Insert_Data(nSign,1);
				}
				nRotate = rotate[countForRotate++];
				if(nRotate == 0 )
					Insert_Data(nRotate, 1);
				else
					Insert_Data((nRotate+1), 2);
				break;
			case 3:
				for(i=0; i < 3; i++)
				{
					nDifCI = difIndex[countForDifIndex++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nBPL= BPLTable[nDifCI];
					nPayload = payloadTable[nDifCI];
					Insert_Data(nBPL,prefix_size);
					if(nPayload != -1)
						Insert_Data(nPayload, (nBPL-1));
					Insert_Data(nSign,1);
				}
				break;
			case 4:
				nFD = faceDirection[countForFD++];
				if(FDmode == 1)
					Insert_Data(nFD,1);
				nRotate = rotate[countForRotate++];
				if(nRotate == 0 )
					Insert_Data(nRotate, 1);
				else
					Insert_Data((nRotate+1), 2);
				
				break;
			}
		}
	}
	else
	{
		
		unsigned int i;
		unsigned int countForType =0;
		unsigned int countForFD =0;  // Not to support RunLength Coding
		unsigned int countForDifIndex =0;
		unsigned int countForPos =0;
		unsigned int countForRotate =0;
		int nType, nPos, nFD, nDifCI, nRotate;

		unsigned int attr_num;
		int previousQuantValue[3];
		for(attr_num = 0; attr_num < 3; attr_num++)
		{
			previousQuantValue[attr_num] = 0;
		}

		int nSymbol, nSign, input;
		int lengthColorIndex = int(log((double)nColor-1)/log(2.0)) + 1;
		unsigned int NMAX = (int) nColorIndex*3 * lengthColorIndex + 100;

		Adaptive_Data_Model mModel(1024);
		Adaptive_Data_Model mhasnext (2);
		Adaptive_Data_Model mSign (2);
		Adaptive_Data_Model mType (5);
		Adaptive_Data_Model mFD (2);
		Adaptive_Data_Model mPos (3);
		Adaptive_Data_Model mRotate (3);

		Arithmetic_Codec ace(NMAX, NULL);
		ace.start_encoder();


		for( ;countForDifIndex<3;countForDifIndex++)
		{
			nSymbol=difIndex[countForDifIndex];
			while(1)
			{
				ace.encode(nSymbol%1024, mModel);
				nSymbol/=1024;
				if(nSymbol)
					ace.encode(1, mhasnext);
				else
					break;
			}
			ace.encode(0, mhasnext);

		}

		for( ; countForType < type.size(); countForType++)
		{
			nType = type[countForType];
			ace.encode(nType, mType);
			switch(nType)
			{
			case 0:
				nPos= position[countForPos++];
				nFD = faceDirection[countForFD++];
				
				nDifCI = difIndex[countForDifIndex++];
				ace.encode(nPos, mPos);
				ace.encode(nFD, mFD);
				
				if(nDifCI > 0)
					nSign = 0;
				else
				{
					nDifCI = -nDifCI;
					nSign = 1;
				}
				nSymbol = nDifCI-1;
				while(1)
				{
					ace.encode(nSymbol%1024, mModel);
					nSymbol/=1024;
					if(nSymbol)
						ace.encode(1, mhasnext);
					else
						break;
				}
				ace.encode(0, mhasnext);
				ace.encode(nSign, mSign);
				nRotate = rotate[countForRotate++];
				ace.encode(nRotate, mRotate);

				break;
			case 1:
				for(i=0; i < 3; i++)
				{
					nDifCI = difIndex[countForDifIndex++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nSymbol = nDifCI-1;
					while(1)
					{
						ace.encode(nSymbol%1024, mModel);
						nSymbol/=1024;
						if(nSymbol)
							ace.encode(1, mhasnext);
						else
							break;
					}
					ace.encode(0, mhasnext);
					ace.encode(nSign, mSign);
				}
				break;
			case 2:
				nPos= position[countForPos++];
				ace.encode(nPos, mPos);
				for(i=0; i < 2; i++)
				{
					nDifCI = difIndex[countForDifIndex++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nSymbol = nDifCI-1;
					while(1)
					{
						ace.encode(nSymbol%1024, mModel);
						nSymbol/=1024;
						if(nSymbol)
							ace.encode(1, mhasnext);
						else
							break;
					}
					ace.encode(0, mhasnext);
					ace.encode(nSign, mSign);
				}
				nRotate = rotate[countForRotate++];
				ace.encode(nRotate, mRotate);
				break;
			case 3:
				for(i=0; i < 3; i++)
				{
					nDifCI = difIndex[countForDifIndex++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nSymbol = nDifCI;
					while(1)
					{
						ace.encode(nSymbol%1024, mModel);
						nSymbol/=1024;
						if(nSymbol)
							ace.encode(1, mhasnext);
						else
							break;
					}
					ace.encode(0, mhasnext);
					ace.encode(nSign, mSign);
				}
				break;
			case 4:
				nFD = faceDirection[countForFD++];
				ace.encode(nFD, mFD);
				nRotate = rotate[countForRotate++];
				ace.encode(nRotate, mRotate);
				break;
			}
		}
		ace.write_to_mem(m_compressedStream, m_compressedStreamSize);
	}

}


void SVAEncoder::Encode_TexCoordIndex()
{
	if(nTexCoordIndex == 0)
		return;
	Analyze_TexCoordIndex();

	if(entropytype == BP)
	{
		unsigned int nMaxCheckBPL = int(log((float)MaxDifferentialForTexCoordIndex-1) / log((float)2)) +1; 
		unsigned int prefix_size = int(log((float)(nMaxCheckBPL+1)) / log((float)2)) +1;	
		int nBPL, nPayload, nSign;
		unsigned int i;
		unsigned int countForType =0;
		unsigned int countForFD =0;  // Not to support RunLength Coding
		unsigned int countForDifIndex =0;
		unsigned int countForPos =0;
		unsigned int countForRotate =0;
		int nType, nPos, nFD, nDifCI, nRotate;

		Insert_Data(FDmode, LEN_BIT);
		if(FDmode == 0)
			Insert_Data(FDvalue, LEN_BIT);
		Insert_Data(prefix_size, LEN_PREFIX);


		for( ;countForDifIndex<3;countForDifIndex++)
		{
			nBPL= BPLTable[difIndex[countForDifIndex]];
			nPayload = payloadTable[difIndex[countForDifIndex]];
			Insert_Data(nBPL,prefix_size);
			if(nPayload != -1)
				Insert_Data(nPayload, (nBPL-1));
		}

		for( ; countForType < type.size(); countForType++)
		{
			nType = type[countForType];
			Insert_Data(1, nType+1);
			switch(nType)
			{
			case 0:
				nPos= position[countForPos++];
				nFD = faceDirection[countForFD++];
				
				nDifCI = difIndex[countForDifIndex++];
				if(nPos == 0 )
					Insert_Data(nPos, 1);
				else
					Insert_Data((nPos+1), 2);
				if(FDmode == 1)
					Insert_Data(nFD,1);
				//nsert_Data(nRotate, 2);
				if(nDifCI > 0)
					nSign = 0;
				else
				{
					nDifCI = -nDifCI;
					nSign = 1;
				}
				nBPL= BPLTable[nDifCI-1];
				nPayload = payloadTable[nDifCI-1];
				Insert_Data(nBPL,prefix_size);
				if(nPayload != -1)
					Insert_Data(nPayload, (nBPL-1));
				Insert_Data(nSign,1);
				nRotate = rotate[countForRotate++];
				if(nRotate == 0 )
					Insert_Data(nRotate, 1);
				else
					Insert_Data((nRotate+1), 2);

				break;
			case 1:
				for(i=0; i < 3; i++)
				{
					nDifCI = difIndex[countForDifIndex++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nBPL= BPLTable[nDifCI-1];
					nPayload = payloadTable[nDifCI-1];
					Insert_Data(nBPL,prefix_size);
					if(nPayload != -1)
						Insert_Data(nPayload, (nBPL-1));
					Insert_Data(nSign,1);
				}
				break;
			case 2:
				nPos= position[countForPos++];
				//nRotate = rotate[countForRotate++];
				if(nPos == 0 )
					Insert_Data(nPos, 1);
				else
					Insert_Data((nPos+1), 2);
				//nsert_Data(nRotate, 2);
				for(i=0; i < 2; i++)
				{
					nDifCI = difIndex[countForDifIndex++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nBPL= BPLTable[nDifCI-1];
					nPayload = payloadTable[nDifCI-1];
					Insert_Data(nBPL,prefix_size);
					if(nPayload != -1)
						Insert_Data(nPayload, (nBPL-1));
					Insert_Data(nSign,1);
				}
				nRotate = rotate[countForRotate++];
				if(nRotate == 0 )
					Insert_Data(nRotate, 1);
				else
					Insert_Data((nRotate+1), 2);
				break;
			case 3:
				for(i=0; i < 3; i++)
				{
					nDifCI = difIndex[countForDifIndex++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nBPL= BPLTable[nDifCI];
					nPayload = payloadTable[nDifCI];
					Insert_Data(nBPL,prefix_size);
					if(nPayload != -1)
						Insert_Data(nPayload, (nBPL-1));
					Insert_Data(nSign,1);
				}
				break;
			case 4:
				nFD = faceDirection[countForFD++];
				if(FDmode == 1)
					Insert_Data(nFD,1);
				nRotate = rotate[countForRotate++];
				if(nRotate == 0 )
					Insert_Data(nRotate, 1);
				else
					Insert_Data((nRotate+1), 2);
				
				break;
			}
		}
	}
	else
	{
		
		unsigned int i;
		unsigned int countForType =0;
		unsigned int countForFD =0;  // Not to support RunLength Coding
		unsigned int countForDifIndex =0;
		unsigned int countForPos =0;
		unsigned int countForRotate =0;
		int nType, nPos, nFD, nDifCI, nRotate;

		unsigned int attr_num;
		
		int nSymbol, nSign, input;
		int lengthTexCoordIndex = int(log((double)nTexCoord-1)/log(2.0)) + 1;
		unsigned int NMAX = (int) nTexCoordIndex*3 * lengthTexCoordIndex + 100;

		Adaptive_Data_Model mModel(1024);
		Adaptive_Data_Model mhasnext (2);
		Adaptive_Data_Model mSign (2);
		Adaptive_Data_Model mType (5);
		Adaptive_Data_Model mFD (2);
		Adaptive_Data_Model mPos (3);
		Adaptive_Data_Model mRotate (3);

		Arithmetic_Codec ace(NMAX, NULL);
		ace.start_encoder();


		for( ;countForDifIndex<3;countForDifIndex++)
		{
			nSymbol=difIndex[countForDifIndex];
			while(1)
			{
				ace.encode(nSymbol%1024, mModel);
				nSymbol/=1024;
				if(nSymbol)
					ace.encode(1, mhasnext);
				else
					break;
			}
			ace.encode(0, mhasnext);

		}

		for( ; countForType < type.size(); countForType++)
		{
			nType = type[countForType];
			ace.encode(nType, mType);
			switch(nType)
			{
			case 0:
				nPos= position[countForPos++];
				nFD = faceDirection[countForFD++];
				
				nDifCI = difIndex[countForDifIndex++];
				ace.encode(nPos, mPos);
				ace.encode(nFD, mFD);
				
				if(nDifCI > 0)
					nSign = 0;
				else
				{
					nDifCI = -nDifCI;
					nSign = 1;
				}
				nSymbol = nDifCI-1;
				while(1)
				{
					ace.encode(nSymbol%1024, mModel);
					nSymbol/=1024;
					if(nSymbol)
						ace.encode(1, mhasnext);
					else
						break;
				}
				ace.encode(0, mhasnext);
				ace.encode(nSign, mSign);
				nRotate = rotate[countForRotate++];
				ace.encode(nRotate, mRotate);

				break;
			case 1:
				for(i=0; i < 3; i++)
				{
					nDifCI = difIndex[countForDifIndex++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nSymbol = nDifCI-1;
					while(1)
					{
						ace.encode(nSymbol%1024, mModel);
						nSymbol/=1024;
						if(nSymbol)
							ace.encode(1, mhasnext);
						else
							break;
					}
					ace.encode(0, mhasnext);
					ace.encode(nSign, mSign);
				}
				break;
			case 2:
				nPos= position[countForPos++];
				ace.encode(nPos, mPos);
				for(i=0; i < 2; i++)
				{
					nDifCI = difIndex[countForDifIndex++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nSymbol = nDifCI-1;
					while(1)
					{
						ace.encode(nSymbol%1024, mModel);
						nSymbol/=1024;
						if(nSymbol)
							ace.encode(1, mhasnext);
						else
							break;
					}
					ace.encode(0, mhasnext);
					ace.encode(nSign, mSign);
				}
				nRotate = rotate[countForRotate++];
				ace.encode(nRotate, mRotate);
				break;
			case 3:
				for(i=0; i < 3; i++)
				{
					nDifCI = difIndex[countForDifIndex++];
					if(nDifCI > 0)
						nSign = 0;
					else
					{
						nDifCI = -nDifCI;
						nSign = 1;
					}
					nSymbol = nDifCI;
					while(1)
					{
						ace.encode(nSymbol%1024, mModel);
						nSymbol/=1024;
						if(nSymbol)
							ace.encode(1, mhasnext);
						else
							break;
					}
					ace.encode(0, mhasnext);
					ace.encode(nSign, mSign);
				}
				break;
			case 4:
				nFD = faceDirection[countForFD++];
				ace.encode(nFD, mFD);
				nRotate = rotate[countForRotate++];
				ace.encode(nRotate, mRotate);
				break;
			}
		}
		ace.write_to_mem(m_compressedStream, m_compressedStreamSize);
	}

}
int SVAEncoder::QuantizationForGeometry(int QP)
{
	unsigned int i,j,k;
	int previousQuantValue[3];
	quantValueForGeometry = (int)pow(2.0,(double)QP) - 1 ;

	// Processing the Quantization
	int result, circular;
	MaxDifferentialForGeometry = 0;
	for(i = 0; i < 3; i++)
	{
		previousQuantValue[i] = 0;
	}
	for(i = 0; i < nCoord * 3; i++)
	{
		result = (int)(((Coord[i]-quantMinGeometry[i%3])/quantRangeGeometry[i%3])*(float)quantValueForGeometry + 0.5);
		circular = CircularDifferenceForGeometry(previousQuantValue[i%3],result);
		quantCoord.push_back(circular);
		previousQuantValue[i%3] = result;
		if(MaxDifferentialForGeometry < abs(circular))
		{
			MaxDifferentialForGeometry = abs(circular);
		}

	}			

	return 1;

}

int SVAEncoder::QuantizationColor(int QP)
{
	unsigned int i,j,k;
	int previousQuantValue[3];
	int result, circular;
	MaxDifferentialForColor = 0;
	quantValueForColor = (int)pow(2.0,(double)QP) - 1 ;
	for(i = 0; i < 3; i++)
	{
		previousQuantValue[i] = 0;
	}
	
	for(i = 0; i < nColor * 3; i++)
	{
		result = (int)(((Color[i]-quantMinColor[i%3])/quantRangeColor[i%3])*quantValueForColor + 0.5);
		circular = CircularDifferenceForColor(previousQuantValue[i%3],result);
		previousQuantValue[i%3]= result;
		quantColor.push_back(circular);
		if(MaxDifferentialForColor < abs(circular))
		{
			MaxDifferentialForColor = abs(circular);
		}
	}

	return 1;

}


int SVAEncoder::QuantizationTexCoord(int u, int v)
{
	
	unsigned int i,j,k;
	int previousQuantValue[2];
	int result, circular;
	MaxDifferentialForTexCoord = 0;
	//quantValueForTexCoord = (int)pow(2.0,(double)QP) - 1 ;
	quantValueForTexCoord_u = u ;
	quantValueForTexCoord_v = v ;

	for(i = 0; i < 2; i++)
	{
		previousQuantValue[i] = 0;
	}
	
	for(i = 0; i < nTexCoord *2; i++)
	{
		if(i%2 == 0) result = (int)(((TexCoord[i]-quantMinTexCoord[i%2])/quantRangeTexCoord[i%2])*quantValueForTexCoord_u + 0.5);
		else result = (int)(((TexCoord[i]-quantMinTexCoord[i%2])/quantRangeTexCoord[i%2])*quantValueForTexCoord_u + 0.5);

		circular = CircularDifferenceForColor(previousQuantValue[i%2],result);
		previousQuantValue[i%2]= result;
		quantTexCoord.push_back(circular);
		if(MaxDifferentialForTexCoord < abs(circular))
		{
			MaxDifferentialForTexCoord = abs(circular);
		}
	}

	
	return 1;

}

int SVAEncoder::DumpBitstream(char* filename)
{
	FILE* m_fp;
	m_fp = fopen(filename,"wb");
	unsigned char data;
	for(unsigned int i=0;i<dumpBitstream.size();i++)
	{
		data = dumpBitstream[i];
		fwrite(&data,sizeof(unsigned char),1,m_fp);
		
	}	
	int filesize =  ftell(m_fp);
	dumpBitstream.clear();
	quantCoord.clear();
	type.clear();
	difCoordIndex.clear(); 
	position.clear();
	faceDirection.clear();
	rotate.clear();

	fclose(m_fp);
	return filesize;
}


void SVAEncoder::Insert_Data(int symbol, int length)
{
	int i;
	int comparison = 1 << length;
	for(i=0; i < length; i++)
	{
		dumpBit <<= 1;
		comparison >>= 1;
		bitPostion++;
		dumpBit += (symbol & comparison) ?  1:0;
		if(bitPostion == LEN_BYTE)
		{
			bitPostion = 0;
			Write2Stream(m_compressedStream, m_compressedStreamSize, &dumpBit, sizeof(unsigned char));
			dumpBit = INIT_UCHAR;
		}
	}

}


void SVAEncoder::Flush()
{
	if(bitPostion == 0)
		return;
	int shift = LEN_BYTE - bitPostion;
	dumpBit <<= shift;
	Write2Stream(m_compressedStream, m_compressedStreamSize, &dumpBit, sizeof(unsigned char));
}

