//-------------------------------------------------------------------------
// 
// Copyright Institut Telecom 2008, 2009
// Licensed Materials, Program Property of Institut Telecom
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  Institut Telecom 
// retains full right to use the code for his/her own purpose, assign or donate 
// the code to a third party and to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//    AUTHORS:
//    Khaled Mammou
//
//
//    http://www-artemis.it-sudparis.eu/
//    
//    Khaled Mammou      (khaled_mamou@yahoo.fr)
//
//-------------------------------------------------------------------------
#include "TFANDecoder.h"


void TFANDecoder::DecodeTFANData(unsigned char * &compressedStream)
{
		if (m_ifs->GetNCoordIndex() >0) 
		{
			TFANConnectivityDecoder coordIndexDecoder;
			coordIndexDecoder.SetTriangleList(m_ifs->GetCoordIndex());
			coordIndexDecoder.SetNTriangles(m_ifs->GetNCoordIndex());
			coordIndexDecoder.SetNVertices(m_ifs->GetNCoord());
			coordIndexDecoder.SetVertexOrderPres(m_vertexOrderPres);
			coordIndexDecoder.SetTriangleOrderPres(m_triangleOrderPres);
			coordIndexDecoder.SetTrianglePermPres(0);
			coordIndexDecoder.Decode(compressedStream);

			TFANConnectivityDecoder normalIndexDecoder;
			if (m_ifs->GetNNormalIndex() >0) 
			{
				if (m_ifs->GetNormalPerVertex())
				{
					normalIndexDecoder.SetTriangleList(m_ifs->GetNormalIndex());
					normalIndexDecoder.SetNTriangles(m_ifs->GetNNormalIndex());
					normalIndexDecoder.SetNVertices(m_ifs->GetNNormal());
					normalIndexDecoder.SetVertexOrderPres(m_vertexOrderPres);
					normalIndexDecoder.SetTriangleOrderPres(1);
					normalIndexDecoder.SetTrianglePermPres(1);
					normalIndexDecoder.Decode(compressedStream);
				}
				else
				{
					//to be added
				}
			}

			TFANConnectivityDecoder colorIndexDecoder;
			if (m_ifs->GetNColorIndex() >0) 
			{
				if (m_ifs->GetColorPerVertex())
				{
					colorIndexDecoder.SetTriangleList(m_ifs->GetColorIndex());
					colorIndexDecoder.SetNTriangles(m_ifs->GetNColorIndex());
					colorIndexDecoder.SetNVertices(m_ifs->GetNColor());
					colorIndexDecoder.SetVertexOrderPres(m_vertexOrderPres);
					colorIndexDecoder.SetTriangleOrderPres(1);
					colorIndexDecoder.SetTrianglePermPres(1);
					colorIndexDecoder.Decode(compressedStream);
				}
				else
				{
					//to be added
				}
			}

			TFANConnectivityDecoder texCoordIndexDecoder;
			if (m_ifs->GetNTexCoordIndex() >0) 
			{
				texCoordIndexDecoder.SetTriangleList(m_ifs->GetTexCoordIndex());
				texCoordIndexDecoder.SetNTriangles(m_ifs->GetNTexCoordIndex());
				texCoordIndexDecoder.SetNVertices(m_ifs->GetNTexCoord());
				texCoordIndexDecoder.SetVertexOrderPres(m_vertexOrderPres);
				texCoordIndexDecoder.SetTriangleOrderPres(1);
				texCoordIndexDecoder.SetTrianglePermPres(1);
				texCoordIndexDecoder.Decode(compressedStream);
			}

			if (m_ifs->GetNCoord() >0)
			{
				TFANGeometryDecoder coordDecoder;
				coordDecoder.SetGeometryList(m_ifs->GetCoord());
				coordDecoder.SetNVertices(m_ifs->GetNCoord());
				coordDecoder.SetDim(3);
				coordDecoder.SetNQuantizationBits(m_header->QPforGeometry);
				coordDecoder.SetTriangleList(m_ifs->GetCoordIndex());
				coordDecoder.SetNTriangles(m_ifs->GetNCoordIndex());
				coordDecoder.SetTFANS(coordIndexDecoder.GetTFans());
				coordDecoder.SetV2T(coordIndexDecoder.GetV2T());
				coordDecoder.Decode(compressedStream);
				if (m_vertexOrderPres == 1) coordIndexDecoder.ReOrderTab(m_ifs->GetCoord(), 3, m_ifs->GetNCoord());
			}

			if (m_ifs->GetNNormal() >0)
			{
				if (m_ifs->GetNormalPerVertex())
				{
					if (m_ifs->GetNNormalIndex()==0)
					{
						TFANGeometryDecoder normalDecoder;
						normalDecoder.SetGeometryList(m_ifs->GetNormal());
						normalDecoder.SetNVertices(m_ifs->GetNNormal());
						normalDecoder.SetDim(3);
						normalDecoder.SetNQuantizationBits(m_header->QPforNormal);
						normalDecoder.SetTriangleList(m_ifs->GetCoordIndex());
						normalDecoder.SetNTriangles(m_ifs->GetNCoordIndex());
						normalDecoder.SetTFANS(coordIndexDecoder.GetTFans());
						normalDecoder.SetV2T(coordIndexDecoder.GetV2T());
						normalDecoder.Decode(compressedStream);
						if (m_vertexOrderPres == 1) coordIndexDecoder.ReOrderTab(m_ifs->GetNormal(), 3, m_ifs->GetNNormal());
					}
					else
					{
						TFANGeometryDecoder normalDecoder;
						normalDecoder.SetGeometryList(m_ifs->GetNormal());
						normalDecoder.SetNVertices(m_ifs->GetNNormal());
						normalDecoder.SetDim(3);
						normalDecoder.SetNQuantizationBits(m_header->QPforNormal);
						normalDecoder.SetTriangleList(m_ifs->GetNormalIndex());
						normalDecoder.SetNTriangles(m_ifs->GetNNormalIndex());
						normalDecoder.SetTFANS(normalIndexDecoder.GetTFans());
						normalDecoder.SetV2T(normalIndexDecoder.GetV2T());
						normalDecoder.Decode(compressedStream);
						if (m_vertexOrderPres == 1) normalIndexDecoder.ReOrderTab(m_ifs->GetNormal(), 3, m_ifs->GetNNormal());
					}
				}
				else
				{
					// to be added
				}
			}


			if (m_ifs->GetNColor() >0)
			{
				if (m_ifs->GetColorPerVertex())
				{
					if (m_ifs->GetNColorIndex()==0)
					{
						TFANGeometryDecoder colorDecoder;
						colorDecoder.SetGeometryList(m_ifs->GetColor());
						colorDecoder.SetNVertices(m_ifs->GetNColor());
						colorDecoder.SetDim(3);
						colorDecoder.SetNQuantizationBits(m_header->QPforColor);
						colorDecoder.SetTriangleList(m_ifs->GetCoordIndex());
						colorDecoder.SetNTriangles(m_ifs->GetNCoordIndex());
						colorDecoder.SetTFANS(coordIndexDecoder.GetTFans());
						colorDecoder.SetV2T(coordIndexDecoder.GetV2T());
						colorDecoder.Decode(compressedStream);
						if (m_vertexOrderPres == 1) coordIndexDecoder.ReOrderTab(m_ifs->GetColor(), 3, m_ifs->GetNColor());
					}
					else
					{
						TFANGeometryDecoder colorDecoder;
						colorDecoder.SetGeometryList(m_ifs->GetColor());
						colorDecoder.SetNVertices(m_ifs->GetNColor());
						colorDecoder.SetDim(3);
						colorDecoder.SetNQuantizationBits(m_header->QPforColor);
						colorDecoder.SetTriangleList(m_ifs->GetColorIndex());
						colorDecoder.SetNTriangles(m_ifs->GetNColorIndex());
						colorDecoder.SetTFANS(colorIndexDecoder.GetTFans());
						colorDecoder.SetV2T(colorIndexDecoder.GetV2T());
						colorDecoder.Decode(compressedStream);
						if (m_vertexOrderPres == 1) colorIndexDecoder.ReOrderTab(m_ifs->GetColor(), 3, m_ifs->GetNColor());
					}
				}
				else
				{
					// to be added
				}
			}


			if (m_ifs->GetNTexCoord() >0)
			{
				if (m_ifs->GetNTexCoordIndex()==0)
				{
					TFANGeometryDecoder texCoordDecoder;
					texCoordDecoder.SetGeometryList(m_ifs->GetTexCoord());
					texCoordDecoder.SetNVertices(m_ifs->GetNTexCoord());
					texCoordDecoder.SetDim(2);
					texCoordDecoder.SetNQuantizationBits(m_header->QPforTexCoord);
					texCoordDecoder.SetTriangleList(m_ifs->GetCoordIndex());
					texCoordDecoder.SetNTriangles(m_ifs->GetNCoordIndex());
					texCoordDecoder.SetTFANS(coordIndexDecoder.GetTFans());
					texCoordDecoder.SetV2T(coordIndexDecoder.GetV2T());
					texCoordDecoder.Decode(compressedStream);
					if (m_vertexOrderPres == 1) coordIndexDecoder.ReOrderTab(m_ifs->GetTexCoord(), 2, m_ifs->GetNTexCoord());
				}
				else
				{
					TFANGeometryDecoder texCoordDecoder;
					texCoordDecoder.SetGeometryList(m_ifs->GetTexCoord());
					texCoordDecoder.SetNVertices(m_ifs->GetNTexCoord());
					texCoordDecoder.SetDim(2);
					texCoordDecoder.SetTriangleList(m_ifs->GetTexCoordIndex());
					texCoordDecoder.SetNTriangles(m_ifs->GetNTexCoordIndex());
					texCoordDecoder.SetTFANS(texCoordIndexDecoder.GetTFans());
					texCoordDecoder.SetV2T(texCoordIndexDecoder.GetV2T());
					texCoordDecoder.Decode(compressedStream);
					if (m_vertexOrderPres == 1) texCoordIndexDecoder.ReOrderTab(m_ifs->GetTexCoord(), 2, m_ifs->GetNTexCoord());
				}
			}


			if (m_ifs->GetNNormalIndex() >0) 
			{
				if (m_vertexOrderPres == 1) normalIndexDecoder.ReOrderConnec(m_ifs->GetNormalIndex(), m_ifs->GetNNormalIndex());
				normalIndexDecoder.ReOrderPerm(m_ifs->GetNormalIndex(), m_ifs->GetNNormalIndex());
				normalIndexDecoder.ReOrderTab(m_ifs->GetNormalIndex(), 3, m_ifs->GetNNormalIndex());
				if (m_triangleOrderPres == 1) coordIndexDecoder.ReOrderTab(m_ifs->GetNormalIndex(), 3, m_ifs->GetNNormalIndex());
			}
			if (m_ifs->GetNColorIndex() >0) 
			{
				if (m_vertexOrderPres == 1) colorIndexDecoder.ReOrderConnec(m_ifs->GetColorIndex(), m_ifs->GetNColorIndex());
				colorIndexDecoder.ReOrderPerm(m_ifs->GetColorIndex(), m_ifs->GetNColorIndex());
				colorIndexDecoder.ReOrderTab(m_ifs->GetColorIndex(), 3, m_ifs->GetNColorIndex());
				if (m_triangleOrderPres == 1) coordIndexDecoder.ReOrderTab(m_ifs->GetColorIndex(), 3, m_ifs->GetNColorIndex());
			}
			if (m_ifs->GetNTexCoordIndex() >0) 
			{
				if (m_vertexOrderPres == 1) texCoordIndexDecoder.ReOrderConnec(m_ifs->GetTexCoordIndex(), m_ifs->GetNTexCoordIndex());
				texCoordIndexDecoder.ReOrderPerm(m_ifs->GetTexCoordIndex(), m_ifs->GetNTexCoordIndex());
				texCoordIndexDecoder.ReOrderTab(m_ifs->GetTexCoordIndex(), 3, m_ifs->GetNTexCoordIndex());
				if (m_triangleOrderPres == 1) coordIndexDecoder.ReOrderTab(m_ifs->GetTexCoordIndex(), 3, m_ifs->GetNTexCoordIndex());
			}			

			// reorder the coord array
			if (m_vertexOrderPres == 1) {
				coordIndexDecoder.ReOrderConnec(m_ifs->GetCoordIndex(), m_ifs->GetNCoordIndex());
			}
			// reorder the coordIndex
			if (m_triangleOrderPres == 1) {
				coordIndexDecoder.ReOrderTab(m_ifs->GetCoordIndex(), 3, m_ifs->GetNCoordIndex());
			}
		}
	}
