//-------------------------------------------------------------------------
// 
// Copyright Institut Telecom 2008, 2009
// Licensed Materials, Program Property of Institut Telecom
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  Institut Telecom 
// retains full right to use the code for his/her own purpose, assign or donate 
// the code to a third party and to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//    AUTHORS:
//    Khaled Mammou
//
//
//    http://www-artemis.it-sudparis.eu/
//    
//    Khaled Mammou      (khaled_mamou@yahoo.fr)
//
//-------------------------------------------------------------------------
#include "TFANGeometryDecoder.h"
#include "arithmetic_codec.h"
#include <math.h>

void TFANGeometryDecoder::Init()
{
	qTable_ = new int [m_nVertices * m_dim];
	errTable_= new int [m_nVertices * m_dim];
	tagVertex_ = new int [m_nVertices];
	for(int v = 0; v < m_nVertices; v++) {
		tagVertex_[v] = 0;
		for (int d = 0; d < m_dim; d++) {
			qTable_[v*m_dim+d] = 0;
			errTable_[v*m_dim+d] = 0;
		}
	}
	tagTriangle_ = new int [m_nTriangles];
	for(int t = 0; t < m_nTriangles; t++) {
		tagTriangle_[t] = 0;
	}
	min_ = new float[m_dim];
	max_ = new float[m_dim];
	for (int d = 0; d < m_dim; d++) {
		min_[d] = 0.0f;
		max_[d] = 0.0f;
	}
}

TFANGeometryDecoder::~TFANGeometryDecoder(void)
{
	if (qTable_ != NULL) delete [] qTable_;
	if (errTable_ != NULL) delete [] errTable_;
	if (tagVertex_ != NULL) delete [] tagVertex_;
	if (tagTriangle_ != NULL) delete [] tagTriangle_;
	if (min_ != NULL) delete [] min_;
	if (max_ != NULL) delete [] max_;
}
bool TFANGeometryDecoder::GetCoordIndex(int pos, int * coordIndex) {
	if (pos < m_nTriangles) {
		for(int h = 0; h < 3; h++) {
			coordIndex[h] = m_triangles[pos*3+h];
		}
		return true;
	}
	return false;
}
void TFANGeometryDecoder::AddNeighborVertex2Vertex(int v1, int v2)
{
	int found = 0;	
	for (IntVect::iterator posc = vertex2Vertex_[v1].begin(); posc != vertex2Vertex_[v1].end(); ++posc) {
		if ((*posc) == v2) {
			found = 1;
			break;
		}
	}
	if (found == 0)
		vertex2Vertex_[v1].push_back(v2);
}

void TFANGeometryDecoder::ComputeVertex2Vertex(){
	for (int f = 0; f < m_nTriangles; f++) {
		AddNeighborVertex2Vertex(m_triangles[f*3+0], m_triangles[f*3+1]);
		AddNeighborVertex2Vertex(m_triangles[f*3+0], m_triangles[f*3+2]);

		AddNeighborVertex2Vertex(m_triangles[f*3+1], m_triangles[f*3+0]);
		AddNeighborVertex2Vertex(m_triangles[f*3+1], m_triangles[f*3+2]);

		AddNeighborVertex2Vertex(m_triangles[f*3+2], m_triangles[f*3+0]);
		AddNeighborVertex2Vertex(m_triangles[f*3+2], m_triangles[f*3+1]);
	}
}
void TFANGeometryDecoder::AddNeighborVertex2Triangle(int v, int t)
{
	int found = 0;	
	for (IntVect::iterator posc = vertex2Triangle_[v].begin(); posc != vertex2Triangle_[v].end(); ++posc) {
		if ((*posc) == t) {
			found = 1;
			break;
		}
	}
	if (found == 0)
		vertex2Triangle_[v].push_back(t);
}

void TFANGeometryDecoder::ComputeVertex2Triangle(){
	for (int f = 0; f < m_nTriangles; f++) {
		AddNeighborVertex2Triangle(m_triangles[f*3+0], f);
		AddNeighborVertex2Triangle(m_triangles[f*3+1], f);
		AddNeighborVertex2Triangle(m_triangles[f*3+2], f);
	}
}
void TFANGeometryDecoder::AddNeighborTriangle2Triangle(int t1, int t2)
{
	int found = 0;	
	for (IntVect::iterator posc = triangle2Triangle_[t1].begin(); posc != triangle2Triangle_[t1].end(); ++posc) {
		if ((*posc) == t2) {
			found = 1;
			break;
		}
	}
	if (found == 0)
		triangle2Triangle_[t1].push_back(t2);
}

void TFANGeometryDecoder::ComputeTriangle2Triangle(){
	int coordIndex[3] = {-1, -1, -1};

	for (int t = 0; t < m_nTriangles; t++) {
		GetCoordIndex(t, coordIndex);
		for (int k = 0; k < 3; k++) {
			for (IntVect::iterator pt0 = vertex2Triangle_[coordIndex[k]].begin(); pt0 != vertex2Triangle_[coordIndex[k]].end(); ++pt0){
				if ((*pt0) != t){
					int tr0 = (*pt0);
					for (int i=0; i <  (int) vertex2Triangle_[coordIndex[(k+1)%3]].size();i++){
						int tr1 = vertex2Triangle_[coordIndex[(k+1)%3]][i];
						if (tr1 == tr0){
							AddNeighborTriangle2Triangle(t, tr1);
						}
					}
				}
			}
		}
	}
}

int TFANGeometryDecoder::Round(double a){
      if (a >= 0.0) return (int) (a+0.5);
      return -(int) (-a+0.5);
}

void TFANGeometryDecoder::Unquantize(){
      // Constant for uniform quantization
    double * delta = new double[m_dim];
	for (int d = 0; d < m_dim; d++){  
		delta[d] = 1.0;
		if ((max_[d]-min_[d]) != 0.0) delta[d] = (double) ((1 << m_nQBits) - 1) / (max_[d]-min_[d]);
	}

	for (int v = 0; v < m_nVertices; v++) {
	  for (int d = 0; d < m_dim; d++){
		m_geometry[v*m_dim+d] = (float) ((double) qTable_[v*m_dim+d] / delta[d] + min_[d]);
      }
	}
	delete [] delta;
}
void TFANGeometryDecoder::GetVisitedNeighbors(int v, IntVect & vN){
	for (int i = 0; i!= vertex2Vertex_[v].size(); i++) {
		if (tagVertex_[vertex2Vertex_[v][i]] != 0) {
			vN.push_back(vertex2Vertex_[v][i]);
		}
	}
}
int TFANGeometryDecoder::BaryPred(){
	ComputeVertex2Vertex();
	for(int v = 0; v < m_nVertices; v++) {
		BaryPredVertex(v);
	}
	return 0;
}
void TFANGeometryDecoder::BaryPredVertex(int v){
	if (tagVertex_[v] == 0) {
		IntVect vN;
		GetVisitedNeighbors(v, vN);
		if (vN.size() == 0) {
			tagVertex_[v] = 2;
			for (int i = 0; i < m_dim; i++) {
				qTable_[v*m_dim+i] = errTable_[v*m_dim+i];
			}
		}
		else {
			tagVertex_[v] = 1;
			for (int i = 0; i < m_dim; i++) {
				double bary = 0.0;
				for (int k = 0; k != vN.size(); k++) {
					bary += qTable_[vN[k]*m_dim+i];
				}
				bary/= vN.size();
				qTable_[v*m_dim+i] = errTable_[v*m_dim+i] + Round(bary);
			}
		}
	}
}

int TFANGeometryDecoder::DeltaPred(){

	for(int v = 0; v < m_nVertices; v++) {
		if ( tagVertex_[v] == 0) {
			if ( v==0) {
				tagVertex_[v] = 2;
				for (int i = 0; i < m_dim; i++) {
					qTable_[v*m_dim+i] = errTable_[v*m_dim+i];
				}
			}
			else {
				tagVertex_[v] = 1;
				for (int i = 0; i < m_dim; i++) {
					qTable_[v*m_dim+i] = errTable_[v*m_dim+i] + qTable_[(v-1)*m_dim+i];
				}
			}
		}
	}
	return 0;
}
int TFANGeometryDecoder::ParaPred(){
	ComputeVertex2Vertex();
	ComputeVertex2Triangle();
	ComputeTriangle2Triangle();
	int currentT[3] = {-1, -1, -1};
	int nextT[3] = {-1, -1, -1};
	for(int t = 0; t != m_nTriangles; t++) {
		if (tagTriangle_[t] == 0) {
			GetCoordIndex(t, currentT);
			for (int k = 0; k != 3; k++) {
				BaryPredVertex(currentT[k]);
			}
			IntVect LT;
			LT.push_back(t);
			tagTriangle_[t] = 1;
			int ct = -1;
			int nt = -1;
			while (LT.size() != 0){
				ct = LT[0]; LT.erase(LT.begin());
				GetCoordIndex(ct, currentT);
				for (int u = 0; u != triangle2Triangle_[ct].size(); u++){
					nt = triangle2Triangle_[ct][u];
					if (tagTriangle_[nt] == 0) {
						LT.push_back(nt);
						tagTriangle_[nt] = 1;
						GetCoordIndex(nt, nextT);
						// determin the commun edge
						IntVect cEdge;
						for (int k1 = 0; k1 != 3; k1++){
							for (int k2 = 0; k2 != 3; k2++){
								if (nextT[k2] == currentT[k1]) cEdge.push_back(nextT[k2]);
							}
						}


						// determin a and d
						int a = -1;
						int b = cEdge[0];
						int c = cEdge[1];
						int d = -1;
						for (int k1 = 0; k1 != 3; k1++){
							if ((currentT[k1] != b) && (currentT[k1] != c)) a = currentT[k1];
							if ((nextT[k1] != b) && (nextT[k1] != c)) d = nextT[k1];
						}

						if ( tagVertex_[d] == 0) {
							tagVertex_[d] = 1;
							for (int i = 0; i < m_dim; i++) {
								qTable_[d*m_dim+i] = errTable_[d*m_dim+i] - qTable_[a*m_dim+i] + qTable_[b*m_dim+i] + qTable_[c*m_dim+i];
							}
						}
					}
				}
			}
		}
	}
	// for isolated vertices
	for(int v = 0; v < m_nVertices; v++) {
		BaryPredVertex(v);
	}
	return 0;
}
int TFANGeometryDecoder::AdapPred(){

	ComputeVertex2Vertex();

	ComputeVertex2Triangle();

	ComputeTriangle2Triangle();


	int currentT[3] = {-1, -1, -1};
	int nextT[3] = {-1, -1, -1};
	for(int t = 0; t != m_nTriangles; t++) {
		if (tagTriangle_[t] == 0) {
			GetCoordIndex(t, currentT);
			for (int k = 0; k != 3; k++) {
				BaryPredVertex(currentT[k]);
			}
			IntVect LT;
			LT.push_back(t);
			tagTriangle_[t] = 1;
			int ct = -1;
			int nt = -1;
			while (LT.size() != 0){
				ct = LT[0]; LT.erase(LT.begin());
				GetCoordIndex(ct, currentT);
				for (int u = 0; u != triangle2Triangle_[ct].size(); u++){
					nt = triangle2Triangle_[ct][u];
					if (tagTriangle_[nt] == 0) {
						LT.push_back(nt);
						tagTriangle_[nt] = 1;
						GetCoordIndex(nt, nextT);
						// determin the commun edge
						IntVect cEdge;
						for (int k1 = 0; k1 != 3; k1++){
							for (int k2 = 0; k2 != 3; k2++){
								if (nextT[k2] == currentT[k1]) cEdge.push_back(nextT[k2]);
							}
						}


						// determin a and d
						int a = -1;
						int b = cEdge[0];
						int c = cEdge[1];
						int d = -1;
						for (int k1 = 0; k1 != 3; k1++){
							if ((currentT[k1] != b) && (currentT[k1] != c)) a = currentT[k1];
							if ((nextT[k1] != b) && (nextT[k1] != c)) d = nextT[k1];
						}

						if ( tagVertex_[d] == 0) {
							int predMod = vertexPredMode_[0]; vertexPredMode_.erase(vertexPredMode_.begin());
							if (predMod == 1) {// barycenter
								BaryPredVertex(d);
							}
							else {
								for (int i = 0; i < m_dim; i++) {
									qTable_[d*m_dim+i] = errTable_[d*m_dim+i] - qTable_[a*m_dim+i] + qTable_[b*m_dim+i] + qTable_[c*m_dim+i];
								}
								tagVertex_[d] = 1;						
							}
						}
					}
				}
			}
		}
	}
	// for isolated vertices
	for(int v = 0; v < m_nVertices; v++) {
		BaryPredVertex(v);
	}
	return 0;
}
int TFANGeometryDecoder::FastParaPred(){
	int a = 0;
	int b = 0;
	int c = 0;
	int d = 0;
	int e = -1;
	int ta = -1;
	int x = -1;
	int y = -1;
	int z = -1;
	for(int v = 0; v < tfans_->GetV(); v++) {
		if(tfans_->GetSize(v)>2){
			// first vertex
			a = tfans_->GetElement(v,0);
			if(tagVertex_[a]==0){
				for (int i = 0; i < m_dim; i++) {
					qTable_[a*m_dim+i] = errTable_[a*m_dim+i];
				}
				tagVertex_[a]=2;
			}

			// second vertex
			b = tfans_->GetElement(v,1);
			if(tagVertex_[b]==0){
				for (int i = 0; i < m_dim; i++) {
					qTable_[b*m_dim+i] =  errTable_[b*m_dim+i] + qTable_[a*m_dim+i];
				}
				tagVertex_[b]=1;
			}

			//third vertex
			c = tfans_->GetElement(v,2);
			if(tagVertex_[c]==0){
				e = -1;
				ta = -1;
				x = -1;
				y = -1;
				z = -1;

				for (int u = 0; u < V2T_->GetSize(a); u++) {
					ta = V2T_->GetElement(a, u);
					x = m_triangles[ta*3];
					y = m_triangles[ta*3+1];
					z = m_triangles[ta*3+2];
					if ( (x==a) && (y==b) && (z!=c) && (tagVertex_[z]!=0) ) { e = z; break;}
					if ( (x==a) && (z==b) && (y!=c) && (tagVertex_[y]!=0) ) { e = y; break;}

					if ( (y==a) && (x==b) && (z!=c) && (tagVertex_[z]!=0) ) { e = z; break;}
					if ( (y==a) && (z==b) && (x!=c) && (tagVertex_[x]!=0) ) { e = x; break;}

					if ( (z==a) && (x==b) && (y!=c) && (tagVertex_[y]!=0) ) { e = y; break;}
					if ( (z==a) && (y==b) && (x!=c) && (tagVertex_[x]!=0) ) { e = x; break;}
				}
				if (e==-1) {
					for (int i = 0; i < m_dim; i++) {
						qTable_[c*m_dim+i] = errTable_[c*m_dim+i] + qTable_[b*m_dim+i];
					}
					tagVertex_[c]=1;
				}
				else {
					for (int i = 0; i < m_dim; i++) {
						qTable_[c*m_dim+i] = errTable_[c*m_dim+i] + qTable_[b*m_dim+i] - qTable_[e*m_dim+i] + qTable_[a*m_dim+i];
					}
					tagVertex_[c]=3;
				}
			}

			// remaining vertices
			for(int k = 3; k < tfans_->GetSize(v); k++){
				d = tfans_->GetElement(v,k);
				if(tagVertex_[d]==0){
					for (int i = 0; i < m_dim; i++) {
						qTable_[d*m_dim+i] = errTable_[d*m_dim+i] + qTable_[c*m_dim+i] - qTable_[b*m_dim+i] + qTable_[a*m_dim+i];
					}
					b = c;
					c = d;
					tagVertex_[d]=3;
				}
			}
		}
	}
	for (int v=0; v < m_nVertices; v++) {
		if(tagVertex_[v]==0){
			for (int i = 0; i < m_dim; i++) {
				qTable_[v*m_dim+i] = errTable_[v*m_dim+i];
			}
			tagVertex_[v]=2;
		}
	}
	return 0;
}


int TFANGeometryDecoder::Compute(){
	switch(m_predMode) {
		case -1 :		// nothing
			for(int v = 0; v < m_nVertices; v++) {
				tagVertex_[v] = 2;
				for (int i = 0; i < m_dim; i++) {
					errTable_[v*m_dim+i] = qTable_[v*m_dim+i];
				}
			}

		break;
		case 0 :		// barycenter
			BaryPred();
		break;
		
		case 1 :		// delta
			DeltaPred();
		break;

		case 2 :		// para
			ParaPred();
		break;

		case 3 :		// Adaptive
			AdapPred();
		break;
		case 4 :		// Fast para
			FastParaPred();
		break;
	}
	Unquantize();

/*
	for(int v = 0; v < vertexPredMode_.size(); v++) {
		logger.write_2_log("%i\n", vertexPredMode_[v]);
	}
	for(int v = 0; v < m_nVertices; v++) {
		logger.write_2_log("%i\t(%i)\t", v, tagVertex_[v]);
		for (int i = 0; i < m_dim; i++) {
			logger.write_2_log("%i\t", errTable_[v*m_dim+i]);
//			if (abs(errTable_[v*m_dim+i])>64) tagVertex_[v] = 2;
		}
		logger.write_2_log("\n");
	}
*/
	return 0;
}
int TFANGeometryDecoder::Decode(unsigned char * &compressedStream){
	Init();
	unsigned char K = 0;
	unsigned char M = 0;
//	unsigned char q = 0;
	unsigned char predMode = 0;
	unsigned char encodingMode = 0;

	ReadFromStream(compressedStream, &encodingMode, sizeof(unsigned char));
//	ReadFromStream(compressedStream, &q, sizeof(unsigned char));
	ReadFromStream(compressedStream, &predMode, sizeof(unsigned char));
	ReadFromStream(compressedStream, &M, sizeof(unsigned char));
	ReadFromStream(compressedStream, &K, sizeof(unsigned char));
	ReadFromStream(compressedStream, min_, sizeof(float)*m_dim);
	ReadFromStream(compressedStream, max_, sizeof(float)*m_dim);
//	m_nQBits = q;
	m_predMode = predMode;


	
	unsigned int nbrPredModes = 0;
	if (m_predMode == 3) {
		ReadFromStream(compressedStream, &nbrPredModes, sizeof(unsigned int));
	}

	unsigned int code_bytes = 0;
	ReadFromStream(compressedStream, &code_bytes, sizeof(unsigned int));

	unsigned char * code_buffer = new unsigned char [code_bytes];                                                   
	ReadFromStream(compressedStream, code_buffer, sizeof(unsigned char)*code_bytes);

	Adaptive_Bit_Model bModelPredMode;
	Adaptive_Data_Model mModel(2*M+3);
	Adaptive_Bit_Model * repModel = new Adaptive_Bit_Model[m_dim];
	Arithmetic_Codec acd(code_bytes, code_buffer);
	acd.start_decoder();	// initialize decoder

	if (m_predMode == 3) {
		for(int v = 0; v < (int) nbrPredModes; v++) {
			vertexPredMode_.push_back(acd.decode( bModelPredMode));
		}
	}

	int dV = 0;

	if (encodingMode == 0) {

		for(int v = 0; v < m_nVertices; v++) {
			for (int i = 0; i < m_dim; i++) {
				dV = acd.decode(mModel);
				if (dV <= 2*M) {
					errTable_[v*m_dim+i] = dV - M;
				}
				else {
					if (dV == 2*M+1) {
						errTable_[v*m_dim+i] = - (int)(acd.ExpGolombDecode(K) + M + 1);
					} 
					else {
						errTable_[v*m_dim+i] = acd.ExpGolombDecode(K) + M + 1;
					}
				}			
			}
		}
	}
	else {// With RLE	
		for(int v = 0; v < m_nVertices; v++) {
			for (int i = 0; i < m_dim; i++) {
				if (acd.decode(repModel[i]) == 0 ){
					errTable_[v*m_dim+i] = errTable_[(v-1)*m_dim+i];
				}
				else {
					dV = acd.decode(mModel);
					if (dV <= 2*M) {
						errTable_[v*m_dim+i] = dV - M;
					}
					else {
						if (dV == 2*M+1) {
							errTable_[v*m_dim+i] = - (int)(acd.ExpGolombDecode(K) + M + 1);
						} 
						else {
							errTable_[v*m_dim+i] = acd.ExpGolombDecode(K) + M + 1;
						}
					}	
				}
			}
		}
	}

	delete [] code_buffer;
	delete [] repModel;

	Compute();
/*
	logger.write_2_log("Decoder\n");
	for(int v = 0; v < m_nVertices; v++) {
		logger.write_2_log("%i\t(%i)\t", v, tagVertex_[v]);
		for (int i = 0; i < m_dim; i++) {
			logger.write_2_log("%i\t", qTable_[v*m_dim+i]);
		}
		logger.write_2_log("\n");
	}
*/
	return 0;
}
