//-------------------------------------------------------------------------
// 
// Copyright Institut Telecom 2008, 2009
// Licensed Materials, Program Property of Institut Telecom
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.  Institut Telecom 
// retains full right to use the code for his/her own purpose, assign or donate 
// the code to a third party and to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//    AUTHORS:
//    Khaled Mammou
//
//
//    http://www-artemis.it-sudparis.eu/
//    
//    Khaled Mammou      (khaled_mamou@yahoo.fr)
//
//-------------------------------------------------------------------------
#include "TFANEncoder.h"
#include "TFANDecoder.h"
#include "utils.h"



void TFANEncoder::EncodeDataBuffer(unsigned char * compressedStream, unsigned int & compressedStreamSize)
{
	if (m_ifs->GetNCoordIndex() >0) 
	{
		TFANConnectivityEncoder coordIndexEncoder;
		coordIndexEncoder.SetTriangleList(m_ifs->GetCoordIndex());
		coordIndexEncoder.SetNTriangles(m_ifs->GetNCoordIndex());
		coordIndexEncoder.SetNVertices(m_ifs->GetNCoord());
		coordIndexEncoder.SetVertexOrderPres(m_params.m_vertexOrderPres);
		coordIndexEncoder.SetTriangleOrderPres(m_params.m_triangleOrderPres);
		coordIndexEncoder.SetTrianglePermPres(0);
		coordIndexEncoder.Encode(compressedStream, compressedStreamSize);
		// compute the decoder connectivity
		TFANConnectivityDecoder coordIndexDecoder;
		coordIndexDecoder.ReadStream(coordIndexEncoder.nbrFans_, coordIndexEncoder.degree_, 
								 coordIndexEncoder.cases_, coordIndexEncoder.vertices_, coordIndexEncoder.ops_,
								 m_ifs->GetCoordIndex(), m_ifs->GetNCoordIndex(), m_ifs->GetNCoord(),
								 coordIndexEncoder.GetMaxV2T());
		coordIndexDecoder.Decode();
		//===========================================




		TFANConnectivityEncoder normalIndexEncoder;
		TFANConnectivityDecoder normalIndexDecoder;
		if ((m_ifs->GetNNormalIndex()>0))
		{
			if (m_ifs->GetNormalPerVertex())
			{
				// reorder the normalIndex
				coordIndexEncoder.ReOrderTab(m_ifs->GetNormalIndex(), 3, m_ifs->GetNNormalIndex());
				coordIndexEncoder.ReOrderPerm(m_ifs->GetNormalIndex(), m_ifs->GetNNormalIndex());
				// encode the normalIndex
				normalIndexEncoder.SetTriangleList(m_ifs->GetNormalIndex());
				normalIndexEncoder.SetNTriangles(m_ifs->GetNNormalIndex());
				normalIndexEncoder.SetNVertices(m_ifs->GetNNormal());
				normalIndexEncoder.SetVertexOrderPres(m_params.m_vertexOrderPres);
				normalIndexEncoder.SetTriangleOrderPres(1);
				normalIndexEncoder.SetTrianglePermPres(1);
				normalIndexEncoder.Encode(compressedStream, compressedStreamSize);
				// compute the decoder connectivity
				normalIndexDecoder.ReadStream(normalIndexEncoder.nbrFans_, normalIndexEncoder.degree_, 
									 normalIndexEncoder.cases_, normalIndexEncoder.vertices_, normalIndexEncoder.ops_,
									 m_ifs->GetNormalIndex(), m_ifs->GetNNormalIndex(), m_ifs->GetNNormal(),
									 normalIndexEncoder.GetMaxV2T());
				normalIndexDecoder.Decode();
				//===========================================
			}
			else
			{
				// to be added
			}
		}
		
		TFANConnectivityEncoder colorIndexEncoder;
		TFANConnectivityDecoder colorIndexDecoder;
		if (m_ifs->GetNColorIndex()>0)
		{
			if (m_ifs->GetColorPerVertex())
			{
				// reorder the colorIndex
				coordIndexEncoder.ReOrderTab(m_ifs->GetColorIndex(), 3, m_ifs->GetNColorIndex());
				coordIndexEncoder.ReOrderPerm(m_ifs->GetColorIndex(), m_ifs->GetNColorIndex());
				// encode the colorIndex
				colorIndexEncoder.SetTriangleList(m_ifs->GetColorIndex());
				colorIndexEncoder.SetNTriangles(m_ifs->GetNColorIndex());
				colorIndexEncoder.SetNVertices(m_ifs->GetNColor());
				colorIndexEncoder.SetVertexOrderPres(m_params.m_vertexOrderPres);
				colorIndexEncoder.SetTriangleOrderPres(1);
				colorIndexEncoder.SetTrianglePermPres(1);
				colorIndexEncoder.Encode(compressedStream, compressedStreamSize);
				// compute the decoder connectivity
				colorIndexDecoder.ReadStream(colorIndexEncoder.nbrFans_, colorIndexEncoder.degree_, 
									 colorIndexEncoder.cases_, colorIndexEncoder.vertices_, colorIndexEncoder.ops_,
									 m_ifs->GetColorIndex(), m_ifs->GetNColorIndex(), m_ifs->GetNColor(),
									 colorIndexEncoder.GetMaxV2T());
				colorIndexDecoder.Decode();
				//===========================================
			}
			else
			{
				// to be added
			}
		}
		
		TFANConnectivityEncoder texCoordIndexEncoder;
		TFANConnectivityDecoder texCoordIndexDecoder;
		if (m_ifs->GetNTexCoordIndex()>0)
		{
			// reorder the texCoordIndex
			coordIndexEncoder.ReOrderTab(m_ifs->GetTexCoordIndex(), 3, m_ifs->GetNTexCoordIndex());
			coordIndexEncoder.ReOrderPerm(m_ifs->GetTexCoordIndex(), m_ifs->GetNTexCoordIndex());
			// encode the texCoordIndex
			texCoordIndexEncoder.SetTriangleList(m_ifs->GetTexCoordIndex());
			texCoordIndexEncoder.SetNTriangles(m_ifs->GetNTexCoordIndex());
			texCoordIndexEncoder.SetNVertices(m_ifs->GetNTexCoord());
			texCoordIndexEncoder.SetVertexOrderPres(m_params.m_vertexOrderPres);
			texCoordIndexEncoder.SetTriangleOrderPres(1);
			texCoordIndexEncoder.SetTrianglePermPres(1);
			texCoordIndexEncoder.Encode(compressedStream, compressedStreamSize);
			// compute the decoder connectivity
			texCoordIndexDecoder.ReadStream(texCoordIndexEncoder.nbrFans_, texCoordIndexEncoder.degree_, 
								 texCoordIndexEncoder.cases_, texCoordIndexEncoder.vertices_, texCoordIndexEncoder.ops_,
								 m_ifs->GetTexCoordIndex(), m_ifs->GetNTexCoordIndex(), m_ifs->GetNTexCoord(),
								 texCoordIndexEncoder.GetMaxV2T());
			texCoordIndexDecoder.Decode();
			//===========================================
		}


		if (m_ifs->GetNCoord() >0)
		{
			// reorder the coord array
			coordIndexEncoder.ReOrderTab(m_ifs->GetCoord(), 3, m_ifs->GetNCoord());		
			TFANGeometryEncoder coordEncoder;
			coordEncoder.SetGeometryList(m_ifs->GetCoord());
			coordEncoder.SetNQuantizationBits(m_params.m_nQCoordBits);
			coordEncoder.SetNVertices(m_ifs->GetNCoord());
			coordEncoder.SetDim(3);
			coordEncoder.SetTriangleList(m_ifs->GetCoordIndex());
			coordEncoder.SetNTriangles(m_ifs->GetNCoordIndex());
			coordEncoder.SetPredMode(m_params.m_coordPredMode);
			coordEncoder.SetTFANS(coordIndexDecoder.GetTFans());
			coordEncoder.SetV2T(coordIndexDecoder.GetV2T());
			coordEncoder.Encode(compressedStream, compressedStreamSize);
		}
		if (m_ifs->GetNNormal() >0)
		{
			if (m_ifs->GetNormalPerVertex())
			{
				if (m_ifs->GetNNormalIndex()==0)
				{
					// reorder the normal array
					coordIndexEncoder.ReOrderTab(m_ifs->GetNormal(), 3, m_ifs->GetNNormal());		
					TFANGeometryEncoder normalEncoder;
					normalEncoder.SetGeometryList(m_ifs->GetNormal());
					normalEncoder.SetNQuantizationBits(m_params.m_nQNormalBits);
					normalEncoder.SetNVertices(m_ifs->GetNNormal());
					normalEncoder.SetDim(3);
					normalEncoder.SetTriangleList(m_ifs->GetCoordIndex());
					normalEncoder.SetNTriangles(m_ifs->GetNCoordIndex());
					normalEncoder.SetPredMode(m_params.m_normalPredMode);
					normalEncoder.SetTFANS(coordIndexDecoder.GetTFans());
					normalEncoder.SetV2T(coordIndexDecoder.GetV2T());
					normalEncoder.Encode(compressedStream, compressedStreamSize);
				}
				else
				{
					// reorder the normal array
					normalIndexEncoder.ReOrderTab(m_ifs->GetNormal(), 3, m_ifs->GetNNormal());		
					TFANGeometryEncoder normalEncoder;
					normalEncoder.SetGeometryList(m_ifs->GetNormal());
					normalEncoder.SetNQuantizationBits(m_params.m_nQNormalBits);
					normalEncoder.SetNVertices(m_ifs->GetNNormal());
					normalEncoder.SetDim(3);
					normalEncoder.SetTriangleList(m_ifs->GetNormalIndex());
					normalEncoder.SetNTriangles(m_ifs->GetNNormalIndex());
					normalEncoder.SetPredMode(m_params.m_normalPredMode);
					normalEncoder.SetTFANS(normalIndexDecoder.GetTFans());
					normalEncoder.SetV2T(normalIndexDecoder.GetV2T());
					normalEncoder.Encode(compressedStream, compressedStreamSize);
				}
			}
			else
			{
				// to be added
			}
		}

		if (m_ifs->GetNColor() >0)
		{
			if (m_ifs->GetColorPerVertex())
			{
				if (m_ifs->GetNColorIndex()==0)
				{
					// reorder the color array
					coordIndexEncoder.ReOrderTab(m_ifs->GetColor(), 3, m_ifs->GetNColor());		
					TFANGeometryEncoder colorEncoder;
					colorEncoder.SetGeometryList(m_ifs->GetColor());
					colorEncoder.SetNQuantizationBits(m_params.m_nQColorBits);
					colorEncoder.SetNVertices(m_ifs->GetNColor());
					colorEncoder.SetDim(3);
					colorEncoder.SetTriangleList(m_ifs->GetCoordIndex());
					colorEncoder.SetNTriangles(m_ifs->GetNCoordIndex());
					colorEncoder.SetPredMode(m_params.m_colorPredMode);
					colorEncoder.SetTFANS(coordIndexDecoder.GetTFans());
					colorEncoder.SetV2T(coordIndexDecoder.GetV2T());
					colorEncoder.Encode(compressedStream, compressedStreamSize);
				}
				else
				{
					// reorder the color array
					colorIndexEncoder.ReOrderTab(m_ifs->GetColor(), 3, m_ifs->GetNColor());		
					TFANGeometryEncoder colorEncoder;
					colorEncoder.SetGeometryList(m_ifs->GetColor());
					colorEncoder.SetNQuantizationBits(m_params.m_nQColorBits);
					colorEncoder.SetNVertices(m_ifs->GetNColor());
					colorEncoder.SetDim(3);
					colorEncoder.SetTriangleList(m_ifs->GetColorIndex());
					colorEncoder.SetNTriangles(m_ifs->GetNColorIndex());
					colorEncoder.SetPredMode(m_params.m_colorPredMode);
					colorEncoder.SetTFANS(colorIndexDecoder.GetTFans());
					colorEncoder.SetV2T(colorIndexDecoder.GetV2T());
					colorEncoder.Encode(compressedStream, compressedStreamSize);
				}
			}
			else
			{
				// to be added
			}
		}

		if (m_ifs->GetNTexCoord() >0)
		{
			if (m_ifs->GetNTexCoordIndex()==0)
			{
				// reorder the texCoord array
				coordIndexEncoder.ReOrderTab(m_ifs->GetTexCoord(), 3, m_ifs->GetNTexCoord());		
				TFANGeometryEncoder texCoordEncoder;
				texCoordEncoder.SetGeometryList(m_ifs->GetTexCoord());
				texCoordEncoder.SetNQuantizationBits(m_params.m_nQTexCoordBits);
				texCoordEncoder.SetNVertices(m_ifs->GetNTexCoord());
				texCoordEncoder.SetDim(3);
				texCoordEncoder.SetTriangleList(m_ifs->GetCoordIndex());
				texCoordEncoder.SetNTriangles(m_ifs->GetNCoordIndex());
				texCoordEncoder.SetPredMode(m_params.m_texCoordPredMode);
				texCoordEncoder.SetTFANS(coordIndexDecoder.GetTFans());
				texCoordEncoder.SetV2T(coordIndexDecoder.GetV2T());
				texCoordEncoder.Encode(compressedStream, compressedStreamSize);
			}
			else
			{
				// reorder the texCoord array
				texCoordIndexEncoder.ReOrderTab(m_ifs->GetTexCoord(), 2, m_ifs->GetNTexCoord());		
				TFANGeometryEncoder texCoordEncoder;
				texCoordEncoder.SetGeometryList(m_ifs->GetTexCoord());
				texCoordEncoder.SetNQuantizationBits(m_params.m_nQTexCoordBits);
				texCoordEncoder.SetNVertices(m_ifs->GetNTexCoord());
				texCoordEncoder.SetDim(2);
				texCoordEncoder.SetTriangleList(m_ifs->GetTexCoordIndex());
				texCoordEncoder.SetNTriangles(m_ifs->GetNTexCoordIndex());
				texCoordEncoder.SetPredMode(m_params.m_texCoordPredMode);
				texCoordEncoder.SetTFANS(texCoordIndexDecoder.GetTFans());
				texCoordEncoder.SetV2T(texCoordIndexDecoder.GetV2T());
				texCoordEncoder.Encode(compressedStream, compressedStreamSize);
			}
		}
	}
}