//-------------------------------------------------------------------------
// 
// Copyright ETRI, Hanyang University 2008, 2009
// Licensed Materials, Program Property of ETRI & Hanyang University
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.
// ETRI and Hanyang University retain full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//    AUTHORS:
//    Seung Wook Lee (ETRI)
//    Daiyong Kim (Hanyang University)
//    KyoungSoo Son (Hanyang University)
//
//    http://www.etri.re.kr/
//    http://dmlab.hanyang.ac.kr/
//  
//    Seung Wook Lee     (tajinet@etri.re.kr)  
//    Daiyong Kim        (dykim@dmlab.hanyang.ac.kr)    
//    KyoungSoo Son      (bonokensin@dmlab.hanyang.ac.kr)
//
//-------------------------------------------------------------------------



#ifndef _QBCRDECODER_H_
#define _QBCRDECODER_H_


#include "IndexedFaceSet.h"
#include "arithmetic_codec.h"
#include "EncodingParameter.h"
#include "SC3DMCHeader.h"




#define QUANT_SIZE 10

class QBCRDecoder{

public :
	
//	std::vector <int> quantCoord;

	IndexedFaceSet* outIFS_;
	SC3DMCHeader *m_Header; 

	unsigned char *readBitstream;
	
	
	unsigned int numberOfCoord;
	unsigned int numberOfNormal;
	unsigned int numberOfColor;
	unsigned int numberOfTexCoord;
	unsigned int numberOfOtherAttributes;
	unsigned int dimensionOfOtherAttributes;
	
	unsigned int numberOfCoordIndex;
	unsigned int numberOfNormalIndex;
	unsigned int numberOfColorIndex;
	unsigned int numberOfTexCoordIndex;
	unsigned int numberOfOtherAttributesIndex;

	unsigned int m_streamSize;
	unsigned char compressionMask;
	unsigned char m_meshType;
	float creaseAngle;
	unsigned char encodingType;

	int codingtype;
	int entropytype;

	float quantMinForGeometry[3];					// The min value in x, y, z
	float quantRangeForGeometry[3];					// the max range between min and max in x, y, z
	float quantMinForNormal[3];					// The min value in x, y, z
	float quantRangeForNormal[3];						// the max range between min and max in x, y, z
	float quantMinForColor[3];					// The min value in x, y, z
	float quantRangeForColor[3];						// the max range between min and max in x, y, z
	float quantMinForTexCoord[2];					// The min value in x, y, z
	float quantRangeForTexCoord[3];						// the max range between min and max in x, y, z
	float quantMinForOtherAttribute[3];					// The min value in x, y, z
	float quantRangeForOtherAttribute;						// the max range between min and max in x, y, z

	int *BPLTable;

	unsigned int QPforGeometry;
	float quantValueforGeometry;
	unsigned int QPforNormal;
	float quantValueforNormal;
	unsigned int QPforTexCoord;
	float quantValueforTexCoord;
	unsigned int QPforColor;
	float quantValueforColor;

	unsigned char *m_pReadStream;
	unsigned int unCurrentBitPosition;
	unsigned int unCurrentBufferPosition;

	
public:
	int ReadData(char * filename);
	float InverseQuantization(unsigned quantCoord, float quantRange, float quantMin, float quantValue);
	int DecoderQBCRData(unsigned char * &compressedStream);
	inline void SetSC3DMCHeader(SC3DMCHeader* pHeader){ m_Header = pHeader;}
	inline void SetIFS(IndexedFaceSet* ifs){ outIFS_ = ifs;}
	void SetHeader() ;

	void DecoderHeader();
	void DecoderGeometry();
	void DecoderConnectivity();
	void DecoderNormal();
	void DecoderNormalIndex();
	void DecoderTexCoord();
	void DecoderTexCoordIndex();
	void DecoderColor();
	void DecoderColorIndex();
	void DumpCompressedVB(char* nameVB);	
	void ComposeIndexedFaceSet(IndexedFaceSet &_myIFS);
	unsigned int ReadBitsFromStorage(unsigned int unReadLength);

	void BPLInit();


	QBCRDecoder(void);
	~QBCRDecoder(void);

private:

};

#endif //_QBCRDECODER_H_