//-------------------------------------------------------------------------
// 
// Copyright ETRI, Hanyang University 2008, 2009
// Licensed Materials, Program Property of ETRI & Hanyang University
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.
// ETRI and Hanyang University retain full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//    AUTHORS:
//    Seung Wook Lee (ETRI)
//    Daiyong Kim (Hanyang University)
//    KyoungSoo Son (Hanyang University)
//
//    http://www.etri.re.kr/
//    http://dmlab.hanyang.ac.kr/
//  
//    Seung Wook Lee     (tajinet@etri.re.kr)  
//    Daiyong Kim        (dykim@dmlab.hanyang.ac.kr)    
//    KyoungSoo Son      (bonokensin@dmlab.hanyang.ac.kr)
//
//-------------------------------------------------------------------------



#ifndef _QBCRENCODER_H
#define _QBCRENCODER_H

#include <stdio.h>

#include "EncodingParameter.h"

#define QUANT_SIZE 10


class QBCREncoder{

	//Add here your features and functions
	//...

public :

	// Encoding results
	std::vector <unsigned char> dumpBitstream;  // Bitstream to save encoded buffer
	std::vector <int> quantCoord;				// save the quantized data
	std::vector <int> quantNormal;
	std::vector <float> normalized;
	std::vector <int> quantColor;
	std::vector <int> quantTexCoord;

	// 3D data
	float *Coord;								// Coordinate information
	int *CoordIndex;							// Coordinate Index information
	float *Normal;								// Normal information
	int *NormalIndex;							// Normal Index information
	float *TexCoord;							// Texture Coordinate information
	int *TexCoordIndex;							// Coordinate Index information
	float *Color;							// Coordinate information
	int *ColorIndex;							// Coordinate Index information

	int QPforGeomerty;
	int QPforNormal;
	int QPforColor;
	int QPforTexCoord;


	unsigned int nCoord;						// The number of coordinate
	unsigned int nCoordIndex;					// The number of coordinate Index

	unsigned int nNormal;						// The number of Normal 
	unsigned int nNormalIndex;					// The number of Normal Index

	unsigned int nColor;						// The number of Color
	unsigned int nColorIndex;					// The number of Color Index

	unsigned int nTexCoord;						// The number of texture 
	unsigned int nTexCoordIndex;					// The number of Color

	unsigned int nOtherAtrribute;						
	unsigned int nOtherAtrributeIndex;		

	unsigned char *m_compressedStream;
	unsigned int m_compressedStreamSize;

	int codingtype;
	int entropytype;

	int MaxDifferentialForGeometry;
	int MaxDifferentialForNormal;
	int MaxDifferentialForColor;
	int MaxDifferentialForTexCoord;
	
	int* BPLTable;
	int* payloadTable;



	// quantization data
	int	quantmodeGeometry;
	float quantMinGeometry[3];					// The min value in x, y, z
	float quantRangeGeometry[3];					// the max range between min and max in x, y, z
	float quantMinNormal[3];					// The min value in x, y, z
	float quantRangeNormal[3];						// the max range between min and max in x, y, z
	float quantMinColor[3];					// The min value in x, y, z
	float quantRangeColor[3];						// the max range between min and max in x, y, z
	float quantMinTexCoord[2];					// The min value in x, y, z
	float quantRangeTexCoord[3];						// the max range between min and max in x, y, z
	float quantMinOtherAttribute[3];					// The min value in x, y, z
	float quantRangeOtherAttribute;						// the max range between min and max in x, y, z

	float quantMax[3];							// The max value in x, y, z
	

	unsigned char dumpBit;						// byte size data to represent the bit before dump
	int bitPostion;								// indicate present bit position in dumpBit
	IndexedFaceSet *m_ifs;
	EncodingParams m_params;
	
	void ParsingConfigure(char* config);
	int Encode_QBCR(IndexedFaceSet &myIFS, EncodingParams &m_params);		// To process QBCR algorithm
	inline void SetIFS(IndexedFaceSet * ifs){ m_ifs = ifs;}
	inline void SetParams(EncodingParams params){ m_params = params;}
	void EncodeDataBuffer(unsigned char * compressedStream, unsigned int & compressedStreamSize);
	inline void SetDumpBit(int nDumpBit) { dumpBit = nDumpBit; }
	inline void SetBitPostion(int nBitPostion) { bitPostion = nBitPostion; }



	int QuantizationGeometry(int QP);			// To process Quantization algorithm 
	int QuantizationNormal(int QP);			// To process Quantization algorithm 
	int QuantizationColor(int QP);			// To process Quantization algorithm 
	int QuantizationTexCoord(int QP);			// To process Quantization algorithm 
	void Normalization();

	void Initialization();						// To initialize the valuable
	void Encode_Header();						// To process QBCR Header encoding
	void Encode_Geometry();						// To process QBCR Geometry encoding
	void Encode_Connectivity();					// To process QBCR Connectivity encoding
	void Encode_Normal();						
	
	void Encode_NormalIndex();					
	void Encode_Texture();						
	void Encode_TextureIndex();					
	void Encode_Color();						
	void Encode_ColorIndex();					

	int DumpBitstream(char* filename);			// encoding result save the buffer
	void Insert_Data(int symbol, int length);	// to insert data in buffer limited in length bit.
	void Flush();

	void Encode_CommonHeader(IndexedFaceSet &m_ifs, EncodingParams &m_params);
	void VectorToBuffer(unsigned char * compressedStream, unsigned int & compressedStreamSize);

	void BPCInit();


	QBCREncoder(void);
	~QBCREncoder(void);

	
private:

};


#endif //_QBCRENCODER

