//-------------------------------------------------------------------------
// 
// Copyright ETRI, Hanyang University 2008, 2009
// Licensed Materials, Program Property of ETRI & Hanyang University
// 
// This software module is an implementation of a part of one or more
// MPEG-4 tools as specified by the MPEG-4 standard (ISO/IEC 14496).
// ISO/IEC gives users of the MPEG-4 standard (ISO/IEC 14496) free
// license to this software module or modifications thereof for use in
// hardware or software products claiming conformance to the MPEG-4
// standard (ISO/IEC 14496).  Those intending to use this software module
// in hardware or software products are advised that its use may infringe
// existing patents.  The original developer of this software module and
// his/her company, the subsequent editors and their companies, and
// ISO/IEC have no liability for use of this software module or
// modifications thereof in an implementation.  Copyright is not released
// for non MPEG-4 standard (ISO/IEC 14496) conforming products.
// ETRI and Hanyang University retain full right to use the code 
// for his/her own purpose, assign or donate the code to a third party and
// to inhibit third parties from using the code for non MPEG-4 standard 
// (ISO/IEC 14496) conforming products. This copyright notice must be included
// in all copies or derivative works.
// 
//    AUTHORS:
//    Seung Wook Lee (ETRI)
//    Daiyong Kim (Hanyang University)
//    KyoungSoo Son (Hanyang University)
//
//    http://www.etri.re.kr/
//    http://dmlab.hanyang.ac.kr/
//  
//    Seung Wook Lee     (tajinet@etri.re.kr)  
//    Daiyong Kim        (dykim@dmlab.hanyang.ac.kr)    
//    KyoungSoo Son      (bonokensin@dmlab.hanyang.ac.kr)
//
//-------------------------------------------------------------------------


#ifndef _SVAENCODER_H
#define _SVAENCODER_H

#include <stdio.h>
#include "EncodingParameter.h"
#define QUANT_SIZE 10


class SVAEncoder{

	//Add here your features and functions
	//...

public :

	// Encoding results
	std::vector <unsigned char> dumpBitstream;
	std::vector <int> quantCoord;
	std::vector <int> quantNormal;
	std::vector <int> quantColor;
	std::vector <int> quantTexCoord;
	std::vector <int> type;
	std::vector <int> difCoordIndex; 
	std::vector <int> difIndex; 
	std::vector <int> position;
	std::vector <int> faceDirection;
	std::vector <int> rotate;

	// 3D data
	float *Coord;
	int *CoordIndex;
	float *Normal;								// Normal information
	int *NormalIndex;							// Normal Index information
	float *TexCoord;							// Texture Coordinate information
	int *TexCoordIndex;							// Coordinate Index information
	float *Color;							// Coordinate information
	int *ColorIndex;							// Coordinate Index information

	unsigned int nCoord;
	unsigned int nCoordIndex;

	unsigned int nNormal;						// The number of Normal 
	unsigned int nNormalIndex;					// The number of Normal Index

	unsigned int nColor;						// The number of Color
	unsigned int nColorIndex;					// The number of Color Index

	unsigned int nTexCoord;						// The number of texture 
	unsigned int nTexCoordIndex;					// The number of Color

	unsigned int nOtherAtrribute;						
	unsigned int nOtherAtrributeIndex;		

	
	unsigned char *m_compressedStream;
	unsigned int m_compressedStreamSize;


	// quantization data
	int	quantmodeGeometry;
	float quantMinGeometry[3];					// The min value in x, y, z
	float quantRangeGeometry[3];					// the max range between min and max in x, y, z
	float quantMinNormal[3];					// The min value in x, y, z
	float quantRangeNormal[3];						// the max range between min and max in x, y, z
	float quantMinColor[3];					// The min value in x, y, z
	float quantRangeColor[3];						// the max range between min and max in x, y, z
	float quantMinTexCoord[2];					// The min value in x, y, z
	float quantRangeTexCoord[2];						// the max range between min and max in x, y, z
	float quantMinOtherAttribute[3];					// The min value in x, y, z
	float quantRangeOtherAttribute;						// the max range between min and max in x, y, z

	float quantMax[3];							// The max value in x, y, z

	unsigned char dumpBit;
	int bitPostion;

	int* BPLTable;
	int* payloadTable;
	int posTable[3];

	int MaxDifferentialForGeometry;
	int MaxDifferentialForConnectivity;
	int MaxDifferentialForNormalIndex;
	int MaxDifferentialForColorIndex;
	int MaxDifferentialForTexCoordIndex;
	int MaxDifferentialForTexCoord;
	int MaxDifferentialForColor;


	int QPforGeomerty;
	int QPforNormal;
	int QPforColor;
	int QPforTexCoord;
	int u, v;

	unsigned int quantValueForGeometry;
	unsigned int quantValueForColor;
	unsigned int quantValueForTexCoord_u;
	unsigned int quantValueForTexCoord_v;

	int FDmode;
	int FDvalue;
	IndexedFaceSet* m_ifs;
	EncodingParams m_params;

	int codingtype;
	int entropytype;
	

	inline void SetIFS(IndexedFaceSet* ifs){ m_ifs = ifs;}
	inline void SetParams(EncodingParams params){ m_params = params;}
	void EncodeDataBuffer(unsigned char * compressedStream, unsigned int & compressedStreamSize);

	void Initialize();
	void ParsingConfigure(char* config);
	int Encode_SVA(IndexedFaceSet &myIFS, EncodingParams &m_params);
	int QuantizationForGeometry(int QP);
	int QuantizationTexCoord(int u, int v);
	int QuantizationColor(int QP);
	int DumpBitstream(char* filename);

	void ConnectivityAnalyze();
	void BPCInit();
	void RunLengthCodingForFaceDirection();

	
	void Insert_Data(int symbol, int length);
	void Flush();
	
	inline void SetDumpBit(int nDumpBit) { dumpBit = nDumpBit; }
	inline void SetBitPostion(int nBitPostion) { bitPostion = nBitPostion; }

	void Encode_CommonHeader(IndexedFaceSet &m_ifs, EncodingParams &m_params);
	void Encode_Header();
	void Encode_Geometry();
	void Encode_Connectivity();
	void Encode_Normal();						
	void Analyze_NormalIndex();
	void Encode_NormalIndex();					
	void Encode_TexCoord();						
	void Analyze_TexCoordIndex();
	void Encode_TexCoordIndex();					
	void Encode_Color();	
	void Analyze_ColorIndex();
	void Encode_ColorIndex();

	
	unsigned long _normal2code(const float *normal,unsigned int subdivision);
	void InverseRotation(int* data, int size);
	int Rotation(int* data, int move);
	int CircularDifferenceForConnectivity(int prev, int cur);
	int CircularDifferenceForNormalIndex(int prev, int cur);
	int CircularDifferenceForColorIndex(int prev, int cur);
	int CircularDifferenceForTexCoordIndex(int prev, int cur);
	int CircularDifferenceForGeometry(int prev, int cur);
	int CircularDifferenceForColor(int prev, int cur);
	int CircularDifferenceForTexCoord(int prev, int cur, int flag);

	void VectorToBuffer(unsigned char * compressedStream, unsigned int & compressedStreamSize);


	SVAEncoder(void);
	~SVAEncoder(void);

	
private:

};

#endif //_SVAENCODER

