#pragma once
#include <vector>
#include <map>
#include "MultiVectOpt.h"
#include "BaseGeometryDecoder.h"
typedef std::vector<int> IntVect;
typedef std::map<int, IntVect> IntMultiVect;
typedef std::vector<IntVect> VectIntVect;

class TFANGeometryDecoder : public BaseGeometryDecoder
{
	int * m_triangles;
	int m_nTriangles;
	int m_predMode;

	int * qTable_;
	int * errTable_;
	float * min_;
	float * max_;
	int * tagVertex_;
	int * tagTriangle_;

	IntVect vertexPredMode_;

	IntMultiVect vertex2Vertex_;
	IntMultiVect vertex2Triangle_;
	IntMultiVect triangle2Triangle_;
	
	MultiVectOpt  * tfans_;
	MultiVectOpt * V2T_;

	inline int * GetTagVertexTable() { return tagVertex_;};
	inline int * GetErrTable() { return errTable_;};
	inline int GetTableV() { return m_nVertices;};
	inline int GetTableD() { return m_dim;};

	inline int GetNCoordIndex(){ return m_nTriangles;};
	bool GetCoordIndex(int pos, int * coordIndex);

	void AddNeighborVertex2Vertex(int v1, int v2);
	void ComputeVertex2Vertex();
	void AddNeighborVertex2Triangle(int v, int t);
	void ComputeVertex2Triangle();
	void AddNeighborTriangle2Triangle(int t1, int t2);
	void ComputeTriangle2Triangle();

	int BaryPred();
	int DeltaPred();
	int ParaPred();
	int AdapPred();
	void BaryPredVertex(int v);
	int FastParaPred();


	int Compute();


	int Round(double a);
	void GetVisitedNeighbors(int v, IntVect & vN);
	void Unquantize();
	void Init();
	
public:
	inline void SetNQuantizationBits(int nQBits){ m_nQBits = nQBits;}
	inline void SetTriangleList(int * triangles){ m_triangles = triangles;}
	inline void SetNTriangles(int nTriangles){ m_nTriangles = nTriangles;}
	inline void SetPredMode(int predMode){ m_predMode = predMode;}
	inline void SetTFANS(MultiVectOpt * tfans){ tfans_ = tfans;}
	inline void SetV2T(MultiVectOpt * V2T){ V2T_ = V2T;} 

	int Decode(unsigned char * &compressedStream);

	TFANGeometryDecoder(void){
		qTable_ = NULL;
		errTable_ = NULL;
		tagVertex_ = NULL;
		tagTriangle_ = NULL;
		min_ = NULL;
		max_ = NULL;
		m_triangles = NULL;
		m_nTriangles = 0;
		m_predMode = 4;
		tfans_ = NULL;
		V2T_ = NULL;
	};
	~TFANGeometryDecoder(void);
};
