#pragma once
#include <vector>
#include <map>
#include "BinaryIO.h"
#include "MultiVectOpt.h"
#include "BaseGeometryEncoder.h"
typedef std::vector<int> IntVect;
typedef std::map<int, IntVect> IntMultiVect;
typedef std::vector<IntVect> VectIntVect;

class TFANGeometryEncoder : public BaseGeometryEncoder
{

	int * m_triangles;
	int m_nTriangles;
	int m_predMode;


	int * qTable_;
	int * errTable_;
	float * min_;
	float * max_;
	int * tagVertex_;
	int * tagTriangle_;

	IntVect vertexPredMode_;

	IntMultiVect vertex2Vertex_;
	IntMultiVect vertex2Triangle_;
	IntMultiVect triangle2Triangle_;

	MultiVectOpt  * tfans_;
	MultiVectOpt * V2T_;

	inline int * GetTagVertexTable() { return tagVertex_;};
	inline int * GetErrTable() { return errTable_;};
	inline int GetTableV() { return m_nVertices;};
	inline int GetTableD() { return m_dim;};

	inline int GetNCoordIndex(){ return m_nTriangles;};
	bool GetCoordIndex(int pos, int * coordIndex);

	void AddNeighborVertex2Vertex(int v1, int v2);
	void ComputeVertex2Vertex();
	void AddNeighborVertex2Triangle(int v, int t);
	void ComputeVertex2Triangle();
	void AddNeighborTriangle2Triangle(int t1, int t2);
	void ComputeTriangle2Triangle();

	int Round(double a);
	void GetVisitedNeighbors(int v, IntVect & vN);

	int BaryPred();
	int DeltaPred();
	int ParaPred();
	int AdapPred();
	void BaryPredVertex(int v);
	int AdapPredAut();
	int LPararPred(int a, int b, int c, int d);
	int FastParaPred();

	void Quantize();
	int Compute();
	int EncodeACM(unsigned char M, unsigned char K);
	void GetEncodeACMParameters(unsigned char & M, unsigned char & K);

	int EncodeRLE(unsigned char M, unsigned char K);
	int EncodeNoRLE(unsigned char M, unsigned char K);

	void Init();

public:

	inline void SetTriangleList(int * triangles){ m_triangles = triangles;}
	inline void SetNTriangles(int nTriangles){ m_nTriangles = nTriangles;}
	inline void SetPredMode(int predMode){ m_predMode = predMode;}
	inline void SetTFANS(MultiVectOpt * tfans){ tfans_ = tfans;}
	inline void SetV2T(MultiVectOpt * V2T){ V2T_ = V2T;} 
	

	// This function is called before encode
	// It should return the nombre of bytes to be allocated for the compressed stream
	unsigned int GetEstimatedCompressedStreamSize() { return 0;};

	// This function encodes the geometry represented as m_dim times m_nVertices array 
	// It returns the compressed stream as a char * and the size of the memory space used to store the compressed stream
	// Rq.: no memory allocation is done inside this function. The user should allocate and free the memory. 
	int Encode(unsigned char * &compressedStream, unsigned int & compressedStreamSize);

	TFANGeometryEncoder(void)
	{
		qTable_ = NULL;
		errTable_ = NULL;
		tagVertex_ = NULL;
		tagTriangle_ = NULL;
		min_ = NULL;
		max_ = NULL;
		m_triangles = NULL;
		m_nTriangles = 0;
		m_predMode = 4;
		tfans_ = NULL;
		V2T_ = NULL;
	}
	~TFANGeometryEncoder(void)
	{
		if (qTable_ != NULL) delete [] qTable_;
		if (errTable_ != NULL) delete [] errTable_;
		if (tagVertex_ != NULL) delete [] tagVertex_;
		if (tagTriangle_ != NULL) delete [] tagTriangle_;
		if (min_ != NULL) delete [] min_;
		if (max_ != NULL) delete [] max_;
	}
};
