/***********************************************************************
This software module was originally developed by Khaled MAMOU
(INT/ARTEMIS) in the course of development of the MPEG-4 Systems
(ISO/IEC 14496-1) standard. This software module is an implementation of
a part of one or more MPEG-4 Systems (ISO/IEC 14496-1) tools as
specified by the MPEG-4 Systems (ISO/IEC 14496-1) standard. ISO/IEC
gives users of the MPEG-4 Systems (ISO/IEC 14496-1) free license to
this software module or modifications thereof for use in hardware or
software products claiming conformance to the MPEG-4 Systems (ISO/IEC
14496-1). Those intending to use this software module in hardware or
software products are advised that its use may infringe existing patents.
The original developer of this software module and his/her company, the
subsequent editors and their companies, and ISO/IEC have no liability
for use of this software module or modifications thereof in an
implementation. Copyright is not released for non MPEG-4 Systems
(ISO/IEC 14496-1) conforming products. The author retains full right
to use the code for his/her own purpose, assign or donate the code to a
third party and to inhibit third parties from using the code for non
MPEG-4 Systems (ISO/IEC 14496-1) conforming products. This copyright
notice must be included in all copies or derivative works.
Copyright  2009.
************************************************************************

For more information or to receive an up to date version of this module,
you may mail Khaled MAMOU at khaled.mamou@int-edu.du.
----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
Project Im1Player
INTERNATIONAL ORGANIZATION FOR STANDARDIZATION
ISO/IEC JTC1/SC29/WG11

SUBSYSTEM:  Multiplexer
FILE:       SC3DMCStream.cpp
AUTHOR:     Khaled MAMOU (INT)
OVERVIEW:   Implementation of SC3DMC Stream
-------------------------------------------------------------------------*/
#include "defs.h"
#include "SC3DMCStream.h"


SC3DMCStream :: SC3DMCStream (FILE *fStream, ESDesc *pDescriptor) :
	Stream (fStream, pDescriptor), m_bRAP (true),
	m_bNextRAP (true),
	m_bEOF(false),
	m_bIsSessionHeader(true),
	m_pAUBuffer(NULL),
	m_lAULength(0),
	m_lIndex(0)
{
	unsigned int streamSize = 0;
	fread(&streamSize, sizeof(unsigned int), 1, fStream);
	m_lAULength = (long) streamSize;
	m_pAUBuffer = new unsigned char [m_lAULength];
	fread(m_pAUBuffer, sizeof(unsigned char), m_lAULength, fStream);
	m_lCTS = m_lDTS = 0;
}

SC3DMCStream :: ~SC3DMCStream ()
{
	if (m_pAUBuffer != NULL) delete [] m_pAUBuffer;
}

/*-----------------------------------------------------------------
Time stamped BIFS file are made of access units, each preceded by
4 bytes of time stamp and 1 byte of AU length.
-----------------------------------------------------------------*/
long SC3DMCStream :: GetDTS (long)
{
	if (m_bEOF) return -1; 
	return m_lDTS;
}

long SC3DMCStream :: GetCTS ()
{
	if (m_bEOF) return -1; // we reached the end of the current file
	m_lCTS = m_lDTS;
	return m_lCTS;
}

/*-------------------------------------------------------------
In this version only the first AU is RAP
-------------------------------------------------------------*/
bool SC3DMCStream :: IsRAP ()
{
	bool bResult = m_bRAP;
	m_bRAP = m_bNextRAP;	
	return bResult;
}


/*-----------------------------------------------------------------
GetChar returns a character and read one char ahead
-----------------------------------------------------------------*/
int SC3DMCStream :: GetChar ()
{
	if (m_bEOF) return -1; 


	if(m_lIndex == m_lAULength) {
		m_bEOF = true;
		return -1;
	}
	int c = m_pAUBuffer[m_lIndex++];
	return c;
}


static StreamFactory<SC3DMCStream> SC3DMCFactory("SC3DMC");