// -*- c++ -*-

// sad2009

#include "OpenGl/OpenGl.h"
#include "IndexedRegionSetProxy.h"  ////////
#include "opengl/base/OpenGLVisualRenderer.h"
#include "../../IM1Decoders/AFX/jwafxstreamdecoder.h"
#include "Stream.h"
#include "opengl/appearance/MultiTextureProxy.h"

// Textures
#include "../appearance/TextureNodeProxy.h"
#include "../appearance/TextureTransformProxy.h"

// Constructor
IndexedRegionSetProxy::IndexedRegionSetProxy(IndexedRegionSet *pIndexedRegionSet): 
AfxStreamProxy(pIndexedRegionSet),
m_nTimeStamp(-1),
m_pAutoNormals(NULL),
cases_found(false),
num_display_lists(0)
{}


// traverse scene graph, checking for intersection
/* virtual */ BOOL
IndexedRegionSetProxy::Select(  ///////////
							OpenGLVisualRenderer* pRenderer, 
							DWORD dwTime, 
							unsigned int& depth)
{
	GLuint selectBuffer[4];

	::glSelectBuffer(4, selectBuffer);
	::glRenderMode(GL_SELECT);

	Render(pRenderer, dwTime);

//#ifdef _DEBUG
//	if (pRenderer->CheckGLError("IndexedRegionSetProxy:"))  
//		TRACE("IndexedRegionSetProxy::Select() [%08x]: ", this); 
//#endif

	if (::glRenderMode(GL_RENDER) > 0)
		if (depth > selectBuffer[1]) {
			depth = selectBuffer[1];

			return TRUE;
		}

		return FALSE;
}

// Render object
BOOL IndexedRegionSetProxy::RenderFromStream(MediaStream* poStream, int id, DWORD dwTime)  
{
	BYTE * pData = NULL;
	int nSize = 0;
	BOOL bEOU;
	int nStatus;
	BOOL bSkip = TRUE;
	BOOL bWait = FALSE;
	long lDelta = 0;
	BOOL bDontClose = TRUE;
	DWORD rdvTime;

	pData = (unsigned char *) poStream->Fetch(
		nSize, rdvTime, bEOU, nStatus, bSkip, bWait, lDelta, bDontClose);

	TRACE("[BitWrapperProxy::Enable] Fetch returned Size %4d \n", nSize);

	if (pData)
	{
		//ory DMLAB 2006 12
		if(type==0)
		{
			JWAFX3DMCDecoder* pDecoder = new JWAFX3DMCDecoder;

			pDecoder->SetValues(pData);

			pDecoder->Write("AFX-IndexedRegionSet.debug.vrml", id);   
		}
		else if(type==1)
		{	
			JWAFX3DMCExtensionDecoder* pDecoder = new JWAFX3DMCExtensionDecoder;

			pDecoder->SetValues(pData);

			pDecoder->Write("AFX-IndexedRegionSet.debug.vrml", id); 
		}
		//~ory DMLAB
		Enable(pData);

		return TRUE;
	}
	return FALSE;
}

BOOL IndexedRegionSetProxy::RenderFromBuffer(SFString *pBuffer, int id)  /////////
{
	BYTE * pData = NULL;
	int nSize = 0;

	//ory DMLAB 2006 12
	if(type==0)
	{
		JWAFX3DMCDecoder* pDecoder = new JWAFX3DMCDecoder;

		// Decode the bitstream
		pDecoder->Decode((LPBYTE) pBuffer->GetValue(), pBuffer->GetLength(), (LPBYTE) pData, nSize);

		pDecoder->Write("AFX-IndexedRegionSet.debug.vrml", id);  /////////
	}
	else if(type==1)
	{
		JWAFX3DMCExtensionDecoder* pDecoder = new JWAFX3DMCExtensionDecoder;

		// Decode the bitstream
		pDecoder->Decode((LPBYTE) pBuffer->GetValue(), pBuffer->GetLength(), (LPBYTE) pData, nSize);

		pDecoder->Write("AFX-IndexedRegionSet.debug.vrml", id);  ///////////
	}
	//~ory DMLAB
	Enable(pData);

	return TRUE;
}

BOOL IndexedRegionSetProxy::Render(DWORD dwTime) ////////
{
	return TRUE;
}

/* virtual */ BOOL
IndexedRegionSetProxy::Render(OpenGLVisualRenderer* pRenderer, DWORD dwTime) 
{
	IndexedRegionSet* pIndexedRegionSet = (IndexedRegionSet*) GetMediaObject();
	// *_case variables if they haven't been calculated before
	if (!cases_found) {
		//     N : 0
		//    PO_V : 3    PI_V : 4   RPI_V : 5                                         RPI_F : 9
		//  RO_V : 6              RI_V : 8                      RO_F : 10           RI_F : 12
		// ARTIF : 13
		cases_found = true;

		int color_case_int = 0;
		if (GetColor())
			color_case_int = 3;
		else if (GetColor(0))
			color_case_int = 6;
		if (GetColorIndex().GetLength() != 0)
			color_case_int += 1;
		else if (GetColorIndex(0).GetLength() != 0)
			color_case_int += 2;
		if (color_case_int > 4)
			color_case_int += ((BOOL) (Bool)pIndexedRegionSet->colorPerVertex) ? 0 : 4;
		color_this_vtx = false;
		color_this_face = false;
		if (color_case_int>8)
			color_this_face = true;
		else if (color_case_int != 0)
			color_this_vtx = true;
		color_case = (Attrib_case) color_case_int;

		int normal_case_int = 0;
		if (GetNormal())
			normal_case_int = 3;
		else if (GetNormal(0))
			normal_case_int = 6;
		if (GetNormalIndex().GetLength() != 0)
			normal_case_int += 1;
		else if (GetNormalIndex(0).GetLength() != 0)
			normal_case_int += 2;
		if (normal_case_int > 4)
			normal_case_int += ((BOOL) (Bool)pIndexedRegionSet->normalPerVertex) ? 0 : 4;
		if (normal_case_int == 0)
			normal_case_int = 13; 
		normal_this_vtx = false;
		normal_this_face = false;
		// "normal_case" can never be "N" => normal_this_vtx or normal_this_face will be "true"
		if ((normal_case_int>8) && (normal_case_int<13))
			normal_this_face = true;
		else
			normal_this_vtx = true;
		normal_case = (Attrib_case) normal_case_int;
		
		int tex_case_int = 0;
		if (GetTexCoord())
			tex_case_int = 3;
		else if (GetTexCoord(0))
			tex_case_int = 6;
		if (GetTexCoordIndex().GetLength() != 0)
			tex_case_int += 1;
		else if (GetTexCoordIndex(0).GetLength() != 0)
			tex_case_int += 2;
		tex_this_vtx = tex_case != 0;
		tex_case = (Attrib_case) tex_case_int;
	}

	// Last update
	bool newrender;
	{
		Coordinate * pCoord = GetCoord();
		Color * pColor = GetColor();
		Normal * pNormal = GetNormal();
		TextureCoordinate * pTexCoord = GetTexCoord();
		long max_t = pIndexedRegionSet->GetUpdateTime();
		if ((pCoord)&&(long(pCoord->GetUpdateTime())>max_t)) max_t = long(pCoord->GetUpdateTime());
		if ((pColor)&&(long(pColor->GetUpdateTime())>max_t)) max_t = long(pColor->GetUpdateTime());
		if ((pNormal)&&(long(pNormal->GetUpdateTime())>max_t)) max_t = long(pNormal->GetUpdateTime());
		if ((pTexCoord)&&(long(pTexCoord->GetUpdateTime())>max_t)) max_t = long(pTexCoord->GetUpdateTime());
		for (int i = 0; i < pIndexedRegionSet->regions.GetLength(); i++) {
			if (long(GetRegion(i)->GetUpdateTime())>max_t) max_t = long(GetRegion(i)->GetUpdateTime());
			pColor = GetColor(i);
			pNormal = GetNormal(i);
			pTexCoord = GetTexCoord(i);
			if ((pColor)&&(long(pColor->GetUpdateTime())>max_t)) max_t = long(pColor->GetUpdateTime());
			if ((pNormal)&&(long(pNormal->GetUpdateTime())>max_t)) max_t = long(pNormal->GetUpdateTime());
			if ((pTexCoord)&&(long(pTexCoord->GetUpdateTime())>max_t)) max_t = long(pTexCoord->GetUpdateTime());
		}
		newrender = max_t > m_nTimeStamp;
	}

	if (newrender) {

		ComputeNormals();

		// solid, ccw, ...
		::glPushAttrib(GL_POLYGON_BIT | GL_CURRENT_BIT); 
		if ((BOOL) (Bool) (pIndexedRegionSet->solid))
			::glEnable(GL_CULL_FACE);
		if (! (BOOL) (Bool) (pIndexedRegionSet->ccw) )
			::glFrontFace(GL_CW);
		if (!( GetNormal()||GetNormal(0) ))
			::glDisable(GL_LIGHTING);

		m_nTimeStamp  = (long) dwTime;

		// Create a list with all the arrays that are going to be used
		Coordinate * pCoord = pIndexedRegionSet->coord;
		const MF<Vec3f>& coordarray = pCoord->point;
		MF<int> coordIndex;

		Color * pColor;
		MF<Color3f> colorarray;
		MF<int> colorIndex;

		Normal * pNormal;
		MF<Vec3f> normalarray;
		MF<int> normalIndex;

		TextureCoordinate * pTexCoord;
		MF<Vec2f> texCoordarray;
		MF<int> texCoordIndex;

		// Initialize arrays that are the same for all regions
		switch (color_case) {
			case PI_V:
				colorIndex = pIndexedRegionSet->colorIndex;
			case PO_V:
			case RPI_V:
			case RPI_F:
				pColor = pIndexedRegionSet->color;
				colorarray = pColor->color;
		}

		switch (normal_case) {
			case PI_V:
				normalIndex = pIndexedRegionSet->normalIndex;
			case PO_V:
			case RPI_V:
			case RPI_F:
				pNormal = GetNormal();
				normalarray = pNormal->vector;
				break;
			case ARTIF:					
				pNormal = GetAutoNormal();
				normalarray = pNormal->vector;
		}

		switch (tex_case) {
			case PI_V:
				texCoordIndex = pIndexedRegionSet->texCoordIndex;
			case PO_V:
			case RPI_V:
				pTexCoord = GetTexCoord();
				texCoordarray = pTexCoord->point;
		}

		// Create position_reorder if needed
		bool reorder = false;
		int* position_reorder;
		int* last_known_region;
		if (( color_case == RPI_V) || ( color_case == RO_V) || ( color_case == RI_V) || 
			(normal_case == RPI_V) || (normal_case == RO_V) || (normal_case == RI_V) || 
			(   tex_case == RPI_V) || (   tex_case == RO_V) || (   tex_case == RI_V)) {
				position_reorder = new int[coordarray.GetLength()];
				last_known_region = new int[coordarray.GetLength()];
				for (int l=0; l<coordarray.GetLength(); l++)
					last_known_region[l] = -1;
				reorder = true;
		}

		// In case we are using artificial normals ...
		int artif_normal_count = 0;

		// DISPLAY LISTS
		// Erase previous display lists
		glDeleteLists(m_nDisplayList, num_display_lists);
		// Create as many new display lists as regions
		num_display_lists = pIndexedRegionSet->regions.GetLength();
		m_nDisplayList = ::glGenLists(num_display_lists);	
		
		

		// FOR EACH REGION
		for (int i = 0; i < pIndexedRegionSet->regions.GetLength(); i++) {

			// Start "recording"
			::glNewList(m_nDisplayList + i, GL_COMPILE_AND_EXECUTE);

			// Textures and texture transformations
			if (GetTexture(i)) {
				glPushAttrib(GL_TEXTURE_BIT);
				((TextureNodeProxy*) (GetRegion(i)->texture->GetRenderingObject()))->Enable(pRenderer, dwTime);
			}
			if (GetRegion(i)->textureTransform)
				((TextureTransformProxy*) (GetRegion(i)->textureTransform->GetRenderingObject()))->Enable(dwTime);

			int face = 0;


			// Initialize arrays that are different for each region

			coordIndex = GetCoordIndex(i);

			switch (color_case) {
				case RI_V:
				case RI_F:
					colorIndex = GetColorIndex(i);
				case RO_V:
				case RO_F:
					pColor = GetColor(i);
					colorarray = pColor->color;
					break;
				case RPI_V:
				case RPI_F:
					colorIndex = GetColorIndex(i);
			}
			switch (normal_case) {
				case RI_V:
				case RI_F:
					normalIndex = GetNormalIndex(i);
				case RO_V:
				case RO_F:
					pNormal = GetNormal(i);
					normalarray = pNormal->vector;
					break;
				case RPI_V:
				case RPI_F:
					normalIndex = GetNormalIndex(i);
			}
			switch (tex_case) {
				case RI_V:
					texCoordIndex = GetTexCoordIndex(i);
				case RO_V:
					pTexCoord = GetTexCoord(i);
					texCoordarray = pTexCoord->point;
					break;
				case RPI_V:
					texCoordIndex = GetTexCoordIndex(i);
			}

			// Reorder vertices if needed
			if (reorder) {
				int vtx_in_this_region = 0;
				for (int j=0; j<coordIndex.GetLength(); j++){
					if ((coordIndex[j] != -1)&&(last_known_region[coordIndex[j]] != i)) {
						last_known_region[coordIndex[j]] = i;
						vtx_in_this_region++;
					}
				}
				int* vtx_in_order = new int [vtx_in_this_region];
				int count = 0;
				for (j=0; j<coordarray.GetLength(); j++) {
					if (last_known_region[j] == i)
						vtx_in_order[count++] = j;
				}						
				// Update position_reorder
				for (j=0; j<vtx_in_this_region; j++)
					position_reorder[vtx_in_order[j]] = j;
			}
			BOOL begin = false;
			int coords = coordIndex.GetLength();
			Color3f col;
			Vec3f norm;
			Vec2f texCoord;
			for (int k=0; k<coords; k++) {
				int vtx = coordIndex[k];
				if (vtx == -1) {
					if (begin) {
						::glEnd();
						begin = FALSE;
						face++;
					}
				} else {
					// Get coordinates of current vertex
					const Vec3f& vertex = coordarray[vtx];
					// ____ Attributes per vertex ____ (casos PO_V, PI_V, RPI_V, RO_V, RI_V, ARTIF)
					
					switch (color_case) {
						case PO_V:
							col = colorarray[vtx];
							break;
						case PI_V:
							col = colorarray[colorIndex[vtx]];
							break;
						case RPI_V:
						case RI_V:
							col = colorarray[colorIndex[position_reorder[vtx]]];
							break;
						case RO_V:
							col = colorarray[position_reorder[vtx]];
							break;
					}		
					switch (normal_case) {
						case PO_V:
							norm = normalarray[vtx];
							break;
						case PI_V:
							norm = normalarray[normalIndex[vtx]];
							break;
						case RPI_V:
						case RI_V:
							norm = normalarray[normalIndex[position_reorder[vtx]]];
							break;
						case RO_V:
							norm = normalarray[position_reorder[vtx]];
							break;
						case ARTIF:
							norm = normalarray[artif_normal_count++];
							break;
					}			
					switch(tex_case) {
						case PO_V:
							texCoord = texCoordarray[vtx];
							break;
						case PI_V:
							texCoord = texCoordarray[texCoordIndex[vtx]];
							break;
						case RPI_V:
						case RI_V:
							texCoord = texCoordarray[texCoordIndex[position_reorder[vtx]]];
							break;
						case RO_V:
							texCoord = texCoordarray[position_reorder[vtx]];
							break;
					}
					// Colors, normals and textures in OpenGL
					if(color_this_vtx)
						glColor3f(col.r, col.g, col.b);
					if (normal_this_vtx)
						glNormal3f(norm.x, norm.y, norm.z);
					if (tex_this_vtx)
						glTexCoord2f(texCoord.x, texCoord.y);


					if (!begin) {
						::glBegin(GL_POLYGON);
						// Attributes specified per face
						switch (color_case) {
							case RPI_F:
							case RI_F:
								col = colorarray[colorIndex[face]];
								break;
							case RO_F:
								col = colorarray[face];
								break;
						}
						switch (normal_case) {
							case RPI_F:
							case RI_F:
								norm = normalarray[normalIndex[face]];
								break;
							case RO_F:
								norm = normalarray[face];
								break;
						}
						if (color_this_face)
							glColor3f(col.r, col.g, col.b);
						if (normal_this_face)
							glNormal3f(norm.x, norm.y, norm.z);


						begin = TRUE;
					}
					glVertex3f(vertex.x, vertex.y, vertex.z);
				}
			}
			if (begin)
				::glEnd();

			// Restore textures and texture transformations
			if (GetRegion(i)->texture) {
				((TextureNodeProxy*) (GetRegion(i)->texture->GetRenderingObject()))->Disable(pRenderer, dwTime);
				glPopAttrib();
			}
			if (GetRegion(i)->textureTransform)
				((TextureTransformProxy*) (GetRegion(i)->textureTransform->GetRenderingObject()))->Disable(dwTime);

			::glEndList();
		}
		
		if (reorder) {
			delete [] position_reorder;
			delete [] last_known_region;
		}

		// Undo ::glPushAttrib(GL_POLYGON_BIT | GL_CURRENT_BIT);
		::glPopAttrib();

	} else {
		// Same for all regions
		::glPushAttrib(GL_POLYGON_BIT | GL_CURRENT_BIT); 
		if ((BOOL) (Bool) (pIndexedRegionSet->solid))
			::glEnable(GL_CULL_FACE);
		if (! (BOOL) (Bool) (pIndexedRegionSet->ccw) )
			::glFrontFace(GL_CW);
		if (!( GetNormal()||GetNormal(0) ))
			::glDisable(GL_LIGHTING);

		// Different for each region 
		for (int i=0; i<pIndexedRegionSet->regions.GetLength(); i++) {
			// Textures and transforms
			if (GetTexture(i)) {
				glPushAttrib(GL_TEXTURE_BIT);
				((TextureNodeProxy*) (GetRegion(i)->texture->GetRenderingObject()))->Enable(pRenderer, dwTime);
			}			
			if (GetRegion(i)->textureTransform)
				((TextureTransformProxy*) (GetRegion(i)->textureTransform->GetRenderingObject()))->Enable(dwTime);


			::glCallList(m_nDisplayList+i);

			// Restore textures and transforms
			if (GetRegion(i)->texture) {
				((TextureNodeProxy*) (GetRegion(i)->texture->GetRenderingObject()))->Disable(pRenderer, dwTime);
				glPopAttrib();
			}
			if (GetRegion(i)->textureTransform)
				((TextureTransformProxy*) (GetRegion(i)->textureTransform->GetRenderingObject()))->Disable(dwTime);
		}

		// For all regions
		::glPopAttrib();
	}
	return TRUE;
}


/* virtual */ void
IndexedRegionSetProxy::RenderDescription(OpenGLVisualRenderer* pRenderer, DWORD /* dwTime */)
{
}


// Update bounding box

void IndexedRegionSetProxy::UpdateBbox()
{
	// Node::UpdateBbox();
	for (int i = m_CollisionParents.GetLength() - 1; i >= 0; --i)
		( (MediaObjectProxy*) ( m_CollisionParents.ElementAt(i) ->
		GetRenderingObject() ) ) ->
		UpdateBbox();
}


void IndexedRegionSetProxy::ComputeNormals()
{
	if (   (!GetNormal())   &&   (!GetNormal(0))   ) {
		delete m_pAutoNormals;
		m_pAutoNormals = new Normal;
		m_pAutoNormals->SetExecutive(GetIndexedRegionSet()->GetExecutive());
		MediaObjectProxy * poNormalProxy = new MediaObjectProxy(m_pAutoNormals);
		m_pAutoNormals->SetRenderingObject(poNormalProxy);				
		GenerateNormals();
	}
}


void IndexedRegionSetProxy::GenerateNormals() {

	// autoNormals stores one normal for each CORNER, following their order in
	// the different coordIndex: first the corner of the first region,...

	MF<Vec3f> normals;

	const MF<Vec3f>& coordarray = GetCoord()->point;
	ZArray<int> vtx_last_reg;
	ZArray<int> vtx_map;
	vtx_last_reg.Expand(coordarray.GetLength());
	vtx_map.Expand(coordarray.GetLength());
	for (int i=0; i<coordarray.GetLength(); i++)
		vtx_last_reg[i] = -1;

	for (int k=0; k<GetIndexedRegionSet()->regions.GetLength(); k++) {
		// Step 1: calculate one normal for each face using its first three vertices
		const MF<int>& coordIndex = GetCoordIndex(k);
		ZArray<ZArray<int> > vertexDependencies;
		ZArray<Vec3f> faceNormals;	
		int point = 0;
		int face = 0;
		int numberOfUniqueVertices = 0;
		for (int i=0; i < coordIndex.GetLength(); i++) {
			int coord = coordIndex[i];
			if (coord != -1) {
				if (vtx_last_reg[coord] < k) {
					vtx_last_reg[coord] = k;
					++vertexDependencies;
					vtx_map[coord] = numberOfUniqueVertices++;
				}
				vertexDependencies[vtx_map[coord]] += face;
				// If it's the third vertex of the current face ...
				if (i - point == 2) {
					Vec3f normal = ComputeNormal(
						coordarray[coordIndex[point]],
						coordarray[coordIndex[point + 1]],
						coordarray[coordIndex[point + 2]]);
					normal.Normalize();
					faceNormals += normal;
				}
			} else {
				face++;
				point = i + 1;
			}
		}
		// Taking creaseAngle into account
		face = 0;
		int counter = 0;		
		for (int j = 0; j<coordIndex.GetLength(); ++j) {
			if (coordIndex[j] != -1) {
				// one normal for each corner
				const ZArray<int>& faces = vertexDependencies[vtx_map[coordIndex[j]]];
				Vec3f normal = faceNormals[face];
				Vec3f vertexNormal;
				for (int k = 0; k < faces.GetLength(); ++k) {
					if (normal == faceNormals[faces[k]] ||  ::acos(normal * faceNormals[faces[k]]) <= GetIndexedRegionSet()->creaseAngle)
						vertexNormal += faceNormals[faces[k]];
				}
				normals += vertexNormal;
			} else {
				++face;
			}			
		}		
	}
	m_pAutoNormals->vector = (normals);
	m_pAutoNormals->SetUpdateTime();
}


// Compute normal

Vec3f IndexedRegionSetProxy::ComputeNormal(
	const Vec3f& point_a,
	const Vec3f& point_b,
	const Vec3f& point_c)
{
	Vec3f vector_a(point_a);
	Vec3f vector_b(point_a);

	vector_a -= point_b;
	vector_b -= point_c;

	vector_a.Cross(vector_b);
	vector_a.Normalize();

	return vector_a;
}


void IndexedRegionSetProxy::Enable (unsigned char *pData)
{
}